#ifndef _NEMO_BASE_TYPES_H_
#define _NEMO_BASE_TYPES_H_

#include "Mapping.h"
#include "detail/nemo-errors.h"

namespace nemo {


/**
 * A Collection represents a set of items of type T that can be accessed and manipulated through this interface.
 * In particular this interface allows all-at-once manipulation by using Mappings.
 * @see Mapping.h 
 * @see Vector.h Matrix.h 
 */
template <class T >
class Collection {
public:

	/**
	 * @return Number of elements in a collection
	 */
	virtual unsigned int numElements() const = 0;

	/**
	 * @param elementIndex between zero and numElements()-1
	 * @return Respective element in collection
	 */
	virtual T getElement(unsigned int elementIndex) const = 0;

	/**
	 * Change the value of an element
	 * @param elementIndex between zero and numElements()-1
	 * @param value New value for respective element
	 */
	virtual void setElement(unsigned int elementIndex, T value) = 0;

	/**
	 * Apply an element-wise in-place transformation.
	 * Each element e is replaced by map(e).
	 */
	virtual void transform(Mapping<T,T> map) = 0;

	/**
	 * Apply an element-wise transformation.
	 * The collection itself is not changed, but a new collection is returned that contains the results.
	 * Return values of implementing classes should be covariant: for instance
	 * MathVector<T>::map() returns a MathVector<OutType>.
	 */
	template <typename OutType>
	Collection<OutType> map(Mapping<T,OutType>) const {_NEMO_LOGIC_ERROR("")}

	virtual ~Collection(){}
protected:
	Collection(){}
};


} // end namespace

#endif

