#ifndef _NEMO_FILTERS_H_
#define _NEMO_FILTERS_H_

#include "Mapping.h"


namespace nemo {


/**
 * This Mapping implements a temporal filter that clamps differences
 * between successively presented values. A norm over differences must 
 * be specified at construction time. Thereby a norm is a Mapping from
 * the ValueType to a scalar type, for instance the Euclidean Norm RealVector->double.
 * In general, this could be anything, but it must guarantee that for any value x:
 * - norm( (1/norm(x)) * x) == 1.
 * 
 * An example of the input-output logic for ValueType=double, norm=abs() and a max. difference of 0.4 is 
 * - filter(0.0) == 0.0
 * - filter(1.0) == 0.4
 * - filter(1.0) == 0.8
 * - filter(1.0) == 1.0
 * - filter(0.0) == 0.6
 * 
 * This filter is tested for scalar as well as Vector types:
 * - Mapping<double,double> d = FirstOrderFilter<double,double>::create(0.25, new AbsNorm(), 0.0);
 * - Mapping<RealVector,RealVector> d = FirstOrderFilter<RealVector,double>::create(
			0.25, new EuclideanNorm(), RealVector(0.0,0.0)
	);
 */
template <class ValueType, class NormType>
class FirstOrderFilter : public MappingDefinition<ValueType,ValueType> {
public:
	// FIXME (Value-Value) might have different type
	// e.g. smoothing timestamps requires norm<duration,NormType>

	static Mapping<ValueType,ValueType> create(
			NormType maxDiffNormP, Mapping<ValueType,NormType> normP, ValueType initValue
	){
		return Mapping<ValueType,ValueType>(
			new FirstOrderFilter<ValueType,NormType>(maxDiffNormP, normP, initValue)
		);
	}

	ValueType evaluate(ValueType value) const {
		NormType diffNorm = norm( value-lastValue );
		if(diffNorm <= maxDiffNorm){
			lastValue = value;
			return value;
		}
		ValueType newValue = lastValue + (maxDiffNorm/diffNorm) * (value-lastValue);
		lastValue = newValue;
		return newValue;
	}

private:
	FirstOrderFilter(NormType maxDiffNormP, Mapping<ValueType,NormType> normP, ValueType initValue)
	 : maxDiffNorm(maxDiffNormP), norm(normP), lastValue(initValue)
	{}

	NormType maxDiffNorm;
	Mapping<ValueType,NormType> norm; 
	mutable ValueType lastValue;
};

/**
 * This Mapping implements a temporal filter that clamps differences
 * between successively presented values, as well as second order differences (accelaration). 
 * 
 * It follows the same logic as the FirstOrderFilter.
 * 
 * If the input pattern is a step function, then the output will be an s-shaped curve that reaches the
 * target value without overshoot. For other input-patterns, the slow-down might be suboptimal, i.e. slower than
 * it might be possible within the specified ranges.
 * Nevertheless the output pattern will never exceed the maximum first- and second-order differences.
 * 
 * This filter is tested for scalar as well as Vector types:
 * - Mapping<double,double> d = SecondOrderFilter<double,double>::create(0.25, 0.125 new AbsNorm(), 0.0, 0.0);
 * - Mapping<RealVector,RealVector> d = SecondOrderFilter<RealVector,double>::create(
			0.25, 0.125, new EuclideanNorm(), RealVector(0.0,0.0), RealVector(0.0,0.0)
	);
 */
template <class ValueType, class NormType>
class SecondOrderFilter : public MappingDefinition<ValueType,ValueType> {
public:
	static Mapping<ValueType,ValueType> create(
			NormType maxDiffNormP, NormType maxDiffDiffNormP, 
			Mapping<ValueType,NormType> normP, ValueType initValue, ValueType initDiffValue
	){
		return Mapping<ValueType,ValueType>(
			new SecondOrderFilter<ValueType,NormType>(maxDiffNormP, maxDiffDiffNormP, normP, initValue, initDiffValue)
		);
	}

	ValueType evaluate(ValueType value) const {
		// how many steps from max. velocity to 0.0 velocity?
		NormType slowDownSteps = maxDiffNorm / maxDiffDiffNorm;
		// what is the path length necessary for such 'breaking' maneuvers?
		NormType slowDownLength = slowDownSteps*(slowDownSteps+1)/2 * maxDiffDiffNorm;

		ValueType targetDiff = value-lastValue;
		NormType targetDiffNorm = norm( targetDiff );

		// this scaling causes a 'breaking', soon enough to arrive without overshoot
		// TODO: this works only optimally, if the input is a step function, 
		//    otherwise it might slow down too much
		if(targetDiffNorm < slowDownLength){
			targetDiff *= maxDiffNorm / slowDownLength;
			targetDiffNorm *= maxDiffNorm / slowDownLength;
		}

		// first order clamping
		if(maxDiffNorm < targetDiffNorm){
			targetDiff = (maxDiffNorm/targetDiffNorm) * targetDiff;
			//targetDiffNorm = (maxDiffNorm/targetDiffNorm) * targetDiffNorm;
		}
		
		// second order clamping
		ValueType targetDiffDiff = targetDiff-lastDiff;
		NormType targetDiffDiffNorm = norm( targetDiffDiff );
		if(maxDiffDiffNorm < targetDiffDiffNorm){
			targetDiffDiff = (maxDiffDiffNorm/targetDiffDiffNorm) * targetDiffDiff;
			targetDiff = lastDiff + targetDiffDiff;
		}

		this->lastDiff = targetDiff;
		this->lastValue = lastValue + targetDiff;

		return lastValue;
	}

private:
	SecondOrderFilter(NormType maxDiffNormP, NormType maxDiffDiffNormP, 
			Mapping<ValueType,NormType> normP, ValueType initValue, ValueType initDiffValue)
	 : maxDiffNorm(maxDiffNormP), maxDiffDiffNorm(maxDiffDiffNormP), norm(normP), lastValue(initValue), lastDiff(initDiffValue)
	{}

	NormType maxDiffNorm;
	NormType maxDiffDiffNorm;
	Mapping<ValueType,NormType> norm; 
	mutable ValueType lastValue;
	mutable ValueType lastDiff;
};


} // end namespace

#endif

