#ifndef _INTRUSIVE_COUNTED_H_
#define _INTRUSIVE_COUNTED_H_


#include <boost/interprocess/detail/atomic.hpp>
#include <boost/version.hpp>

/* 
 * TODO test usage of boost/smart_ptr/detail/atomic_count.hpp
 * Might be less efficient due to extra object?
 */

namespace nemo {

/**
 * Standard deleter class to be used by IntrusiveCounted<T>. Delegates to delete(T* p).
 */
template <class T>
class StandardDeleter {
public:
	StandardDeleter(const T *ptr){
		delete(ptr);
	}
};

/**
 * Generic "mix-in" class for intrusive reference counting.
 * If you want to use intrusive pointer (e.g. boost::intrusive_ptr<T>)
 * for your class T, you need to provide an interal counter plus
 * methods intrusive_ptr_add_ref(T*) and intrusive_ptr_release(T*)
 * in the global namespace.
 * 
 * This class automizes this task. Just inherit your class 
 * T from IntrusiveCounted<T> and T can be used within intrusive pointers:
 * \code
class Foo : IntrusiveCounted<Foo> {
public:
	int someFoo;
}

boost::intrusive_ptr<Foo> fooPtr(new Foo());
\endcode
 * 
 * The reference counting is atomic, and thus safe to use in multithreaded environments.
 *  
 * The class D determines the way the handled object is deleted. On deletion, the IntrusiveCounted instance
 * calls D(T* p). By default, the class StandardDeleter<T> is used, 
 * in which D(T* p) simply delegates to delete(p).
 * This template parameter is meant to handle classes or objects with non-standard 
 * allocation-methods, that need to be complemented with non-standard deletion-methods. 
 * If you do not know exactly what (and why) you are doing, do not touch this parameter.
 * 
 */
template <class T, class D=StandardDeleter<T> >
class IntrusiveCounted {
private:

#ifndef NEMO_PARSED_BY_DOXYGEN
#if (BOOST_VERSION >= 104800)
		#define _NEMO_USE_BOOST_IP_DETAIL using namespace boost::interprocess::ipcdetail;
#else
		#define _NEMO_USE_BOOST_IP_DETAIL using namespace boost::interprocess::detail;
#endif
#endif

	static void impl_intrusive_ptr_add_ref(const T* p){
		_NEMO_USE_BOOST_IP_DETAIL
		atomic_inc32( &(static_cast<const IntrusiveCounted*>(p)->refCount) );
	}
	static void impl_intrusive_ptr_release(const T* p){
		_NEMO_USE_BOOST_IP_DETAIL
		if( atomic_dec32( &(static_cast<const IntrusiveCounted*>(p)->refCount) ) == 1){
			//delete(p); // FIXME custom disposal?
			D del(p); // custom disposal!
		}
	}
	static bool impl_unique(const T* p){
		_NEMO_USE_BOOST_IP_DETAIL
		return (atomic_read32( &(static_cast<const IntrusiveCounted*>(p)->refCount) ) == 1);
	}
public:
	/**
	 * Increments the reference counter.
	 */
	friend void intrusive_ptr_add_ref(const T* p){
		impl_intrusive_ptr_add_ref(p);
	}

	/**
	 * Decrements the reference counter and calls delete(p) if the counter has become 0.
	 */
	friend void intrusive_ptr_release(const T* p){
		impl_intrusive_ptr_release(p);
	}

	/**
	 * Returns true is and only if the reference counter is 1.
	 */
	friend bool unique(const T* p){
		return impl_unique(p);
	}

	void intrusive_ptr_add_ref(){
		impl_intrusive_ptr_add_ref( static_cast<T*>(this) );
	}
	void intrusive_ptr_release(){
		impl_intrusive_ptr_release( static_cast<T*>(this) );
	}
	bool unique(){
		return impl_unique( static_cast<T*>(this) );
	}

protected:
	IntrusiveCounted(): refCount(0) {}
	IntrusiveCounted(const IntrusiveCounted&) : refCount(0) {}

	// cppcheck-suppress operatorEqVarError
	IntrusiveCounted& operator=(const IntrusiveCounted&) { 
		// note: it is correct that refCount is NOT changed!
		return *this;
	}

	virtual ~IntrusiveCounted() {};
	//void swap(ref_counter&) {};

	void initRefCount(){
		refCount = 0;
	}

private:
	
	mutable boost::uint32_t refCount;
};


} // end namespace


#endif

