#ifndef _INTRUSIVE_COW_SUPPPORT_H_
#define _INTRUSIVE_COW_SUPPPORT_H_

#ifndef NO_GCC_ADL_WORKAROUND
#include "IntrusiveCounted.h"
#endif

namespace nemo {

/**
 * An interface for generalized copy-construction of objects of type T inside an IntrusiveCowSupport<T>.
 * The default behavior is the 'standard' copy-construction of pointer-contents: @newPtr = new T(*ptr)@.
 * If (and only if) your class T has a memory management that does not permit such an implementation,
 * You can modify the behavior by template-specialization (effective for all instances of T), or by
 * inheritance (in order to use different allocators for different instances of T).
 */
template <class T>
class CopyAllocator {
public:
	/**
	 * Allocates a new object with same content, such that @*copyAllocate(ptr) == *ptr@
	 */
	static T* copyAllocate(const T* value) {
		return new T(*value);
	}
};


/**
 * Generic utility class for "copy-on-write" (COW) data types.
 * Altough it has pointer semantics, this class is NOT a full blown, automatic COW pointer.
 * Rather it can be used in some class C, that supports COW, internally.
 * 
 * The class C is responsible for calling the detach() function before
 * C makes any change in the data T. detach() will then automatically 
 * create a data copy of T is more than one reference point to that data.
 * 
 * There are several requirements on the type T:
 * - T must be copy-constructable.
 * - There must be an operation void intrusive_ptr_add_ref(T*)
 * - There must be an operation void intrusive_ptr_release(T*)
 * - There must be an operation bool unique(T*)
 * The easiest way to achieve the last three requirments is to let 
 * T inherit from nemo::IntrusiveCounted<T>.
 * 
 * C is furthermore responsible for not assigning null-pointers to IntrusiveCowSupport.
 * 
 * The second type argument CA can be used to customize the data-copy procedure
 * that is used in the detach routine. See CopyAllocator.
 */
template <class T, class CA=CopyAllocator<T> >
class IntrusiveCowSupport {

private:
	T *ptr;
 
public:
	IntrusiveCowSupport(T* t) : ptr(t) {
#ifdef NO_GCC_ADL_WORKAROUND
		intrusive_ptr_add_ref(ptr);
#else
		ptr->intrusive_ptr_add_ref();
#endif
	}

	IntrusiveCowSupport(const IntrusiveCowSupport<T,CA>& p) : ptr(p.ptr){
#ifdef NO_GCC_ADL_WORKAROUND
		intrusive_ptr_add_ref(ptr);
#else
		ptr->intrusive_ptr_add_ref();
#endif
	}

	IntrusiveCowSupport& operator=(const IntrusiveCowSupport<T,CA>& rhs){
		if(rhs.ptr != ptr){ // exclude self-assignment
#ifdef NO_GCC_ADL_WORKAROUND
			intrusive_ptr_add_ref(rhs.ptr);
			intrusive_ptr_release(ptr);
#else
			rhs.ptr->intrusive_ptr_add_ref();
			ptr->intrusive_ptr_release();
#endif

			ptr = rhs.ptr;
		}
		return *this;
	}
	virtual ~IntrusiveCowSupport(){
#ifdef NO_GCC_ADL_WORKAROUND
		intrusive_ptr_release(ptr);
#else
		ptr->intrusive_ptr_release();
#endif
	}

	//this operator does NOT detach! the is client responsibility
	// TODO return reference here?
	inline const T* operator->() const{
		return ptr;
	}	

	void detach(){
		T* tmp = ptr;
#ifdef NO_GCC_ADL_WORKAROUND
		if( !( unique(ptr) ) ) {
			//ptr = new T(*tmp);
			ptr = CA::copyAllocate(tmp);
			intrusive_ptr_add_ref(ptr);
			intrusive_ptr_release(tmp);
		}
#else
		if( !( ptr->unique() ) ) {
			//ptr = new T(*tmp);
			ptr = CA::copyAllocate(tmp);
			ptr->intrusive_ptr_add_ref();
			tmp->intrusive_ptr_release();
		}
#endif
	}
};



} // end namespace


#endif

