#ifndef _INTRUSIVE_PTR_H_
#define _INTRUSIVE_PTR_H_


#include "IntrusiveCounted.h"

#include <cassert>


namespace nemo {


/**
 * A tiny utility class for intrusive pointering. This class is almost identical to boost::intrusive_ptr,
 * except that it does not rely on global methods for add_ref and release, but on member functions as
 * defined in the nemo::IntrusiveCounted class. This pointer implementation can be useful is you want to
 * use the IntrusiveCounted implementation without running into problems with GCC bugs, sometimes occuring when
 * 'argument depended lookup' of the add_ref/release methods is used inside template.
 */
template <class T>
class IntrusivePtr {
public:

	IntrusivePtr()
	 : ptr(NULL) 
	{}
	
	IntrusivePtr(const IntrusivePtr<T>& rhs)
	 : ptr(NULL) 
	{
		if(rhs.ptr){
			rhs.ptr->intrusive_ptr_add_ref();
			ptr = rhs.ptr;
		}
	}
	
	IntrusivePtr(T* rhsP)
	 : ptr(NULL) 
	{
		if(rhsP){
			rhsP->intrusive_ptr_add_ref();
			ptr = rhsP;
		}
	}

	IntrusivePtr& operator=(const IntrusivePtr<T>& rhs){ 
		if(rhs.ptr != ptr){ // exclude self-assignment
			if(rhs.ptr) rhs.ptr->intrusive_ptr_add_ref();
			if(ptr) ptr->intrusive_ptr_release();
			ptr = rhs.ptr;
		}
		return *this;		
	}

	virtual ~IntrusivePtr(){
		if(ptr) ptr->intrusive_ptr_release();
	};

	void reset(){
		if(ptr) ptr->intrusive_ptr_release();
		ptr = NULL;
	}

    	inline T& operator*() const{
		assert(ptr!=NULL);
		return *ptr;
	}

    	inline T* operator->() const{
		assert(ptr!=NULL);
		return ptr;
	}

	void swap(IntrusivePtr<T> & rhs){
		T * tmp = ptr;
		ptr = rhs.ptr;
		rhs.ptr = tmp;
	}

private:
	
	T *ptr;
};


} // end namespace


#endif

