#ifndef _NEMO_MAPPING_H_
#define _NEMO_MAPPING_H_


//#include <boost/smart_ptr/intrusive_ptr.hpp>
#include <boost/bind.hpp>


#include "IntrusiveCounted.h"
#include "IntrusivePtr.h"
#include "detail/nemo-cpp11.h"
#include "detail/nemo-errors.h"

#include <stdexcept>




namespace nemo {



/**
 * A MappingDefinition defines the behavior of a function object.
 * The input-output behavior is defined in the evaluate() method
 * that receives an object of type InType and returns an object 
 * of type OutType. Implement this function to create a custom mapping.
 * 
 * A mapping does not necessarily need to a mathematical function.
 * It can have an internal state that influences the input-output behavior.
 * 
 * This class can be used in isolation, just as it is, but is not meant to.
 * Rather, assign a MappingDefinition to a Mapping Object and get some
 * nice syntactic sugar and automatic memory management.
 *
 * @see Mapping
 */
template <class InType, class OutType>
class MappingDefinition : public IntrusiveCounted<MappingDefinition<InType,OutType> >{
public:
	
	/**
	 * Compute the outut value for a given input value.
	 */
	virtual OutType evaluate(InType value) const = 0;

	/**
	 * Creates an inverse function if applicable. This functionality is optional!
	 * If the MappingDefinition implements a mathematical function that is invertible,
	 * it should implement this method, but does not have to.
	 * For non-bijective functions it is sound to return a generalized (left or right) 
	 * inverse function. For a function f this means that it must return a function g
	 * such that  f(g(f(x))) = f(x)  and  g(f(g(x))) = g(x).
	 * If the Mapping is not invertible in this sense, it MUST throw an exception
	 * when getInverse() is called.
	 */
	virtual MappingDefinition<OutType,InType>* getInverse() const {
		_NEMO_LOGIC_ERROR("Inversion not supported for this mapping type");
	}
	
	virtual ~MappingDefinition(){}
};

template <class InType, class OutType>
class Mapping;

#ifndef NEMO_PARSED_BY_DOXYGEN
/**
 * A MappingDefinition for f(x)=x (identity)
 */
template <class T>
class Identity : public MappingDefinition<T,T> {
public:
	static Mapping<T,T> create(){
		return Mapping<T,T>(new Identity<T>());
	}

	MappingDefinition<T,T>* getInverse() const {
		return new Identity<T>();
	}

	T evaluate(T value) const {
		return value;
	}
};
#endif

/**
 * Returns an identity mapping such that x==arg<X>()(x).
 */
template <class T>
static Mapping<T,T> arg(){ return Identity<T>::create(); }


#ifndef NEMO_PARSED_BY_DOXYGEN
namespace detail {
/**
 * A MappingDefinition for f(x)=g(h(x)), i.e. for composing functions in a sequential manner.
 * These definitions are automatically created when you use Mapping::operator()(Mapping) .
 */
template <class InType, class Intermediate, class OutType>
class SequentialMapping : public MappingDefinition<InType,OutType> {
public:
	OutType evaluate(InType value) const{
		return second->evaluate(first->evaluate(value));
	}
	MappingDefinition<OutType,InType>* getInverse() const {
		IntrusivePtr<MappingDefinition<Intermediate,InType> > firstInv( first->getInverse() );
		IntrusivePtr<MappingDefinition<OutType,Intermediate> > secondInv( second->getInverse() );
		return new SequentialMapping<OutType,Intermediate,InType>( secondInv, firstInv );
	}
	/**
	 * Takes two Mappings, where as the first argument is the function that is first evaluated (the inner expression).
	 * evaluate(x) will compute second(first(x)).
	 */
	SequentialMapping(
			IntrusivePtr<MappingDefinition<InType,Intermediate> > firstP, 
			IntrusivePtr<MappingDefinition<Intermediate,OutType> > secondP)
	 : first(firstP), second(secondP)
	{}
private:
	friend class Mapping<Intermediate,OutType>;

	IntrusivePtr<MappingDefinition<InType,Intermediate> > first;
	IntrusivePtr<MappingDefinition<Intermediate,OutType> > second;
};
template <class InType, class Intermediate, class OutType>
class CastMappingDef : public MappingDefinition<InType,OutType> {
public:

	OutType evaluate(InType input) const {
		return (OutType)(map->evaluate(input));
	}

	MappingDefinition<OutType,InType>* getInverse() const {
		IntrusivePtr<MappingDefinition<Intermediate,InType> > innerInv( map->getInverse() );
		IntrusivePtr<MappingDefinition<OutType,Intermediate> > castInv(
			new CastMappingDef<OutType,OutType,Intermediate>(
				IntrusivePtr<MappingDefinition<OutType,OutType> >(new Identity<OutType>())
			)
		);
		return new SequentialMapping<OutType,Intermediate,InType>( castInv, innerInv );
	}

	CastMappingDef(IntrusivePtr<MappingDefinition<InType,Intermediate> > mapP) : map(mapP) {}
private:
	IntrusivePtr<MappingDefinition<InType,Intermediate> > map;
};
}// end namespace detail
#endif

/**
 * A Mapping represents an input-output behavior (for instance a mathematical function)
 * that is defined be a MappingDefinition. The Mapping object adds a lot of convenience
 * to the MappingDefinition object. It acts as a "smart pointer" on the MappingDefinition
 * and adds intuitive operator support.
 * A mapping can be evaluated with the round-bracket operator() which corresponds to evaluate() of
 * the mapping definition. Also, you can use mathematical operators like + on mappings (see ExampleMappingArithmetics.cpp).
 */
template <class InType, class OutType>
class Mapping {
public:
	/**
	 * Create a mapping from a mapping definition. The MappingDefinition object is afterwards owned by the Mapping
	 */
	Mapping(MappingDefinition<InType,OutType> *defP)
	 : def(defP)
	{
		if(!defP){
			_NEMO_ARGUMENT_ERROR("Received null pointer as MappingDefinition");
		}
	}

	/**
	 * Create a mapping With constant value OutType().
	 */
	Mapping()
	 : def(new ConstMappingDef(OutType()))
	{}

	/**
	 * Create a mapping With constant value.
	 */
	Mapping(const OutType &value)
	 : def(new ConstMappingDef(value))
	{}

	/** 
	 * Creates a mapping that evaluates the function that is passed as first argument by a function pointer.
	 * The second argument is optional an can be used to specify the corresponding inverse function.
	 */
	Mapping(OutType (*const func)(InType), InType (*const invFunc)(OutType) = NULL)
	 : def(new CFuncMappingDef<InType,OutType>(func,invFunc))
	{
		if(!func){
			_NEMO_ARGUMENT_ERROR("Received null function pointer");
		}
	}

	/** 
	 * Creates a mapping form a boost::bind expression.
	 */
	template <typename F, typename BoostArgDescription>
	Mapping(boost::_bi::bind_t<OutType, F, BoostArgDescription > func)
	 : def(new BoostBindMappingDef<InType,OutType,F,BoostArgDescription>(func))
	{}

	/**
	 * Evaluate the outcome of the mapping for a given input value.
	 */
	OutType operator()(InType value) const{
		return def->evaluate(value);
	}

	/**
	 * Chain two mappings. h = f(g) means that h(x)=f(g(x)).
	 */
	template <typename OtherInType>
	Mapping<OtherInType,OutType> operator()(Mapping<OtherInType,InType> m) const{
		return Mapping<OtherInType,OutType>(new detail::SequentialMapping<OtherInType,InType,OutType>(m.def, this->def));
	}

	/**
	 * Return the definition object belonging to this Mapping
	 */
	inline IntrusivePtr<MappingDefinition<InType,OutType> > getDefinition() const{
		return def;
	}

	Mapping<OutType,InType> inverse() const{
		MappingDefinition<OutType,InType> *invDef = def->getInverse();
		return Mapping<OutType,InType>(invDef);
	}

private:
	// declare mappings with arbitrary other type-parameters as friend
	template <class I, class O>
	friend class Mapping;

	IntrusivePtr<MappingDefinition<InType,OutType> > def;

#ifndef NEMO_PARSED_BY_DOXYGEN
	/**
	 * A MappingDefinition for f(x)=const
	 */
	//template <class InType, class OutType>
	class ConstMappingDef : public MappingDefinition<InType,OutType> {
	public:
		static Mapping<InType,OutType> create(OutType value){
			return Mapping<InType,OutType>(new ConstMappingDef(value));
		}

		OutType evaluate(InType) const {
			return value;
		}

		ConstMappingDef(OutType valueP) : value(valueP) {}
	private:
		OutType value;
	};

	/**
	 * A MappingDefinition for f(x)=func(x)
	 */
	template <typename I, typename O>
	class CFuncMappingDef : public MappingDefinition<I,O> {
	public:
		O evaluate(I value) const {
			return func(value);
		}

		MappingDefinition<O,I>* getInverse() const {
			if(!inverse){
				_NEMO_LOGIC_ERROR("No inverse function specified for this function pointer");
			}
			return new CFuncMappingDef<O,I>(inverse,func);
		}

		CFuncMappingDef(O (*const funcP)(I), I (*const inverseP)(O)) : func(funcP), inverse(inverseP) {}
	private:
		O (*const func)(I);
		I (*const inverse)(O);
	};

	/**
	 * A MappingDefinition for f(x) = boost::bind(f())
	 */
	template <typename I, typename O, typename F, typename BoostArgDescription>
	class BoostBindMappingDef : public MappingDefinition<I,O> {
	public:
		O evaluate(I value) const {
			return func(value);
		}

		BoostBindMappingDef(boost::_bi::bind_t<OutType, F, BoostArgDescription> funcP) : func(funcP) {}
	private:
		boost::_bi::bind_t<O, F, BoostArgDescription > func;
	};
#endif

};

/**
 * Returns a mapping that casts the output (of type Intermediate) of another mapping into type OutType.
 * Example: cast a double to an int:
 * Mapping<double,int> round = reinterpret<int>(arg<double>());
 */
template <class OutType, typename InType, typename Intermediate>
Mapping<InType,OutType> 
reinterpret(const Mapping<InType,Intermediate> &map){
	return Mapping<InType,OutType>(new detail::CastMappingDef<InType,Intermediate,OutType>(map.getDefinition()));
}


/// for any expression c*x, with c of type T, leftMultInverse<T>(c) must either return a value such that leftMultInverse(c)*(c*x)=x, or throw an exception if that is not possible
template <typename Type> Type leftMultInverse(const Type &)
	{_NEMO_LOGIC_ERROR("leftMultInverse not defined for this data type");}
/// for any expression x*c, with c of type T, rightMultInverse<T>(c) must either return a value such that (x*c)*rightMultInverse(c)=x, or throw an exception if that is not possible
template <typename Type> Type rightMultInverse(const Type &)
	{_NEMO_LOGIC_ERROR("rightMultInverse not defined for this data type");}
/// for any expression c/x, with c of type T, leftDivInverse<T>(c) must either return a value such that leftDivInverse(c)/(c/x)=x, or throw an exception if that is not possible
template <typename Type> Type leftDivideInverse(const Type &)
	{_NEMO_LOGIC_ERROR("leftDivideInverse not defined for this data type");}
/// for any expression x/c, with c of type T, rightDivideInverse<T>(c) must either return a value such that (x/c)/rightDivideInverse(c)=x, or throw an exception if that is not possible
template <typename Type> Type rightDivideInverse(const Type &)
	{_NEMO_LOGIC_ERROR("rightDivideInverse not defined for this data type");}
/// for any expression c+x, with c of type T, leftAddInverse<T>(c) must either return a value such that leftAddInverse(c)+(c+x)=x, or throw an exception if that is not possible
template <typename Type> Type leftAddInverse(const Type &)
	{_NEMO_LOGIC_ERROR("leftAddInverse not defined for this data type");}
/// for any expression x+c, with c of type T, rightAddInverse<T>(c) must either return a value such that (x+c)+rightAddInverse(c)=x, or throw an exception if that is not possible
template <typename Type> Type rightAddInverse(const Type &)
	{_NEMO_LOGIC_ERROR("rightAddInverse not defined for this data type");}
/// for any expression c-x, with c of type T, leftSubtractInverse<T>(c) must either return a value such that leftSubtractInverse(c)-(c-x)=x, or throw an exception if that is not possible
template <typename Type> Type leftSubtractInverse(const Type &)
	{_NEMO_LOGIC_ERROR("leftSubtractInverse not defined for this data type");}
/// for any expression x-c, with c of type T, rightSubtractInverse<T>(c) must either return a value such that (x-c)-rightSubtractInverse(c)=x, or throw an exception if that is not possible
template <typename Type> Type rightSubtractInverse(const Type &)
	{_NEMO_LOGIC_ERROR("rightSubtractInverse not defined for this data type");}



#ifndef NEMO_PARSED_BY_DOXYGEN
#define _nemo_mapping_binary_operators_(Name, OP) \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##Mapping : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return first->evaluate(value) OP second->evaluate(value); \
		} \
		Name##Mapping( \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType1> > &firstP,  \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType2> > &secondP) \
		 : first(firstP), second(secondP) {} \
	private: \
		IntrusivePtr<MappingDefinition<InType,InnerOutType1> > first; \
		IntrusivePtr<MappingDefinition<InType,InnerOutType2> > second; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##MappingConst : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return first->evaluate(value) OP constValue; \
		} \
		Name##MappingConst( \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType1> > &firstP,  \
			const InnerOutType2 &constValueP) \
		 : first(firstP), constValue(constValueP) \
		{} \
	private: \
		IntrusivePtr<MappingDefinition<InType,InnerOutType1> > first; \
		InnerOutType2 constValue; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##ConstMapping : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return constValue OP second->evaluate(value); \
		} \
		MappingDefinition<OutType,InType>* getInverse() const { \
			_NEMO_LOGIC_ERROR("Inversion not supported for 'OP' mapping type"); \
		} \
		Name##ConstMapping( const InnerOutType1 &constValueP, const IntrusivePtr<MappingDefinition<InType,InnerOutType2> > &secondP) \
		 : constValue(constValueP), second(secondP) {} \
	private: \
		InnerOutType1 constValue; \
		IntrusivePtr<MappingDefinition<InType,InnerOutType2> > second; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (Mapping<InType,InnerOutType1> l, Mapping<InType,InnerOutType2> r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##Mapping<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l.getDefinition(), r.getDefinition())); \
	} \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (Mapping<InType,InnerOutType1> l, const InnerOutType2 &r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##MappingConst<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l.getDefinition(), r)); \
	} \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (const InnerOutType1 &l, Mapping<InType,InnerOutType2> r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##ConstMapping<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l, r.getDefinition())); \
	}



#define _nemo_mapping_binary_operators_invertible_(Name, OP) \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##Mapping : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return first->evaluate(value) OP second->evaluate(value); \
		} \
		Name##Mapping( \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType1> > &firstP,  \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType2> > &secondP) \
		 : first(firstP), second(secondP) {} \
	private: \
		IntrusivePtr<MappingDefinition<InType,InnerOutType1> > first; \
		IntrusivePtr<MappingDefinition<InType,InnerOutType2> > second; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##MappingConst : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return first->evaluate(value) OP constValue; \
		} \
		MappingDefinition<OutType,InType>* getInverse() const { \
			IntrusivePtr<MappingDefinition<InnerOutType1,InType> > mapInv(first->getInverse()); \
			IntrusivePtr<MappingDefinition<OutType,InnerOutType1> > innerInv( \
				new detail::CastMappingDef<OutType,_nemo_get_type_(OutType() OP constValue),InnerOutType1>( \
					IntrusivePtr<MappingDefinition<OutType,_nemo_get_type_(OutType() OP constValue)> >( \
						new Name##MappingConst<OutType,OutType,InnerOutType2,_nemo_get_type_(OutType() OP constValue)>(  \
							IntrusivePtr<MappingDefinition<OutType,OutType> >(new Identity<OutType>()), \
							right##Name##Inverse<InnerOutType2>( constValue )  \
						) \
					) \
				) \
			); \
			return new detail::SequentialMapping<OutType,InnerOutType1,InType>(innerInv, mapInv); \
		} \
		Name##MappingConst( \
			const IntrusivePtr<MappingDefinition<InType,InnerOutType1> > &firstP,  \
			const InnerOutType2 &constValueP) \
		 : first(firstP), constValue(constValueP) \
		{} \
	private: \
		IntrusivePtr<MappingDefinition<InType,InnerOutType1> > first; \
		InnerOutType2 constValue; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2, class OutType> \
	class Name##ConstMapping : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return constValue OP second->evaluate(value); \
		} \
		MappingDefinition<OutType,InType>* getInverse() const { \
			IntrusivePtr<MappingDefinition<InnerOutType2,InType> > mapInv(second->getInverse()); \
			IntrusivePtr<MappingDefinition<OutType,InnerOutType2> > innerInv( \
				new detail::CastMappingDef<OutType,_nemo_get_type_(constValue OP OutType()),InnerOutType2>( \
					IntrusivePtr<MappingDefinition<OutType,_nemo_get_type_(constValue OP OutType())> >( \
						new Name##ConstMapping<OutType,InnerOutType1,OutType,_nemo_get_type_(constValue OP OutType())>(  \
							left##Name##Inverse<InnerOutType1>( constValue ),  \
							IntrusivePtr<MappingDefinition<OutType,OutType> >(new Identity<OutType>()) \
						) \
					) \
				) \
			); \
			return new detail::SequentialMapping<OutType,InnerOutType2,InType>(innerInv, mapInv); \
		} \
		Name##ConstMapping( const InnerOutType1 &constValueP, const IntrusivePtr<MappingDefinition<InType,InnerOutType2> > &secondP) \
		 : constValue(constValueP), second(secondP) {} \
	private: \
		InnerOutType1 constValue; \
		IntrusivePtr<MappingDefinition<InType,InnerOutType2> > second; \
	}; \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (Mapping<InType,InnerOutType1> l, Mapping<InType,InnerOutType2> r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##Mapping<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l.getDefinition(), r.getDefinition())); \
	} \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (Mapping<InType,InnerOutType1> l, const InnerOutType2 &r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##MappingConst<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l.getDefinition(), r)); \
	} \
	template <class InType, class InnerOutType1, class InnerOutType2> \
	inline Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))> operator OP (const InnerOutType1 &l, Mapping<InType,InnerOutType2> r){ \
		return Mapping<InType,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(new Name##ConstMapping<InType,InnerOutType1,InnerOutType2,_nemo_get_type_((InnerOutType1() OP InnerOutType2()))>(l, r.getDefinition())); \
	}
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
_nemo_mapping_binary_operators_invertible_(Add, +)
_nemo_mapping_binary_operators_invertible_(Subtract, -)
_nemo_mapping_binary_operators_invertible_(Mult, *)
_nemo_mapping_binary_operators_invertible_(Divide, /)

_nemo_mapping_binary_operators_(Modulo, %)
_nemo_mapping_binary_operators_(BitAnd, &)
_nemo_mapping_binary_operators_(BitOr, |)
_nemo_mapping_binary_operators_(BitXor, ^)
_nemo_mapping_binary_operators_(LeftShift, <<)
_nemo_mapping_binary_operators_(RightShift, >>)

_nemo_mapping_binary_operators_(IsEqual, ==)
_nemo_mapping_binary_operators_(IsNotEqual, !=)
_nemo_mapping_binary_operators_(IsLower, <)
_nemo_mapping_binary_operators_(IsHigher, >)
_nemo_mapping_binary_operators_(IsLowerEqual, <=)
_nemo_mapping_binary_operators_(IsHigherEqual, >=)

_nemo_mapping_binary_operators_(LogicAnd, &&)
_nemo_mapping_binary_operators_(LogicOr, ||)
#endif



#ifndef NEMO_PARSED_BY_DOXYGEN
	// Non-type-changing unary operators
	#define _nemo_mapping_unary_operator_(Name, OP) \
	template <class InType, class OutType> \
	class Name##Mapping : public MappingDefinition<InType,OutType> { \
	public: \
		OutType evaluate(InType value) const{ \
			return OP map->evaluate(value); \
		} \
		MappingDefinition<OutType,InType>* getInverse() const { \
			IntrusivePtr<MappingDefinition<OutType,InType> > mapInv(map->getInverse()); \
			IntrusivePtr<MappingDefinition<OutType,OutType> > innerInv( \
				new Name##Mapping<OutType,OutType> (  \
					IntrusivePtr<MappingDefinition<OutType,OutType> >(new Identity<OutType>()) \
				) \
			); \
			return new detail::SequentialMapping<OutType,OutType,InType>(innerInv, mapInv); \
		} \
		Name##Mapping( const IntrusivePtr<MappingDefinition<InType,OutType> > &mapP )  \
		 : map(mapP) {} \
	private: \
		IntrusivePtr<MappingDefinition<InType,OutType> > map; \
	}; \
	template <class InType, class OutType> \
	inline Mapping<InType,OutType> operator OP (const Mapping<InType,OutType> &m){ \
		return Mapping<InType,OutType>(new Name##Mapping<InType,OutType>(m.getDefinition())); \
	}
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
_nemo_mapping_unary_operator_(UnaryPlus, +)
_nemo_mapping_unary_operator_(UnaryMinus, -)
_nemo_mapping_unary_operator_(LogicNot, !)
_nemo_mapping_unary_operator_(BitwiseNot, ~)
#endif


// inversion rules

// inlining helps against multiple definitions

#define _nemo_scalar_inversion_(Type) \
template <> inline Type leftMultInverse<>(const Type &value){return ((Type)1.0)/value;} \
template <> inline Type rightMultInverse<>(const Type &value){return ((Type)1.0)/value;} \
template <> inline Type leftDivideInverse<>(const Type &value){return value;} \
template <> inline Type rightDivideInverse<>(const Type &value){return ((Type)1.0)/value;} \
template <> inline Type leftAddInverse<>(const Type &value){return -value;} \
template <> inline Type rightAddInverse<>(const Type &value){return -value;} \
template <> inline Type leftSubtractInverse<>(const Type &value){return value;} \
template <> inline Type rightSubtractInverse<>(const Type &value){return -value;} 

#ifndef NEMO_PARSED_BY_DOXYGEN
_nemo_scalar_inversion_(float)
_nemo_scalar_inversion_(double)
_nemo_scalar_inversion_(long double)
#endif

#define _nemo_signed_integer_inversion_(Type) \
template <> inline Type leftAddInverse<>(const Type &value){return -value;} \
template <> inline Type rightAddInverse<>(const Type &value){return -value;} \
template <> inline Type leftSubtractInverse<>(const Type &value){return value;} \
template <> inline Type rightSubtractInverse<>(const Type &value){return -value;} 

#ifndef NEMO_PARSED_BY_DOXYGEN
_nemo_signed_integer_inversion_(short)
_nemo_signed_integer_inversion_(int)
_nemo_signed_integer_inversion_(long int)
#endif


} // end namespace

#endif

