#ifndef _NEMO_MATRIX_H_
#define _NEMO_MATRIX_H_


#include "IntrusiveCowSupport.h"
#include "IntrusiveCounted.h"
#include "Vector.h"
#include "Collection.h"
#include "detail/nemo-cpp11.h"
#include "detail/nemo-errors.h"

#include <Eigen/Core>

#include <stdexcept>
#include <sstream>


namespace nemo {


class MatrixDimension;
static MatrixDimension dim(unsigned int rows, unsigned int cols);

/**
 * Represents the dimension of a Matrix. It is actually nothing else than an pair of unsigned ints representing row and col dimension,
 * but allows typesafe distinction between values and dimensions in Matrices.
 */
class MatrixDimension {
public:
	/**
	 * @return Number of rows
	 */
	inline const unsigned int& rows() const {return numRows;}
	/**
	 * @return Number of columns
	 */
	inline const unsigned int& cols() const {return numCols;}

	friend MatrixDimension dim(unsigned int rows, unsigned int cols);
	MatrixDimension(const MatrixDimension &o) : numRows(o.numRows), numCols(o.numCols) {}
	MatrixDimension& operator=(const MatrixDimension& o){numRows=o.numRows; numCols=o.numCols; return *this;}
	bool operator==(const MatrixDimension &o)const{return numCols==o.numCols&&numRows==o.numRows;}
	bool operator!=(const MatrixDimension &o)const{return !(*this==o);}
	friend std::ostream& operator<<(std::ostream &o, const MatrixDimension &d){
		o<<std::string("dim(")<<(d.numRows)<<std::string(",")<<(d.numRows)<<std::string(")");
		return o;
	}
private:
	MatrixDimension(unsigned int r, unsigned int c) : numRows(r), numCols(c) {}
	unsigned int numRows;
	unsigned int numCols;
};

/**
 * Neat factory method for MatrixDimension objects to be put into Matrix objects.
 */
static inline MatrixDimension dim(unsigned int rows, unsigned int cols){
	return MatrixDimension(rows, cols);
}


// forward declaration for friends in MatrixMath.h
template <class T>
	class Matrix;
#ifndef NEMO_PARSED_BY_DOXYGEN
template <typename T>
	Matrix<T> matMatT(const Matrix<T>&);
template <typename T>
	void _nemo_impl_matMatTRecursive(const Matrix<T>&,Matrix<T> &,unsigned int,
		unsigned int,unsigned int, unsigned int,unsigned int,bool);
template <typename T>
	void _nemo_impl_matMatTBlockRecursive(const Matrix<T>&,Matrix<T> &,unsigned int,
		unsigned int,unsigned int,unsigned int,unsigned int,unsigned int,bool);
template <typename T>
	void _nemo_impl_matMatTBlockIterative(const Matrix<T> &m, Matrix<T> &result);
#endif


#ifndef NEMO_PARSED_BY_DOXYGEN
#ifdef NEMO_ABORT_ON_ERROR
        #define _NEMO_MATRIX_CHECK_INDEX(rowIndex,colIndex) \
                if(rows() <= rowIndex || cols() <= colIndex){ \
                        abort(); \
                }
#else
        #define _NEMO_MATRIX_CHECK_INDEX(rowIndex,colIndex) \
		if(rows() <= rowIndex){ \
			throw std::invalid_argument("Matrix: row-index out of range"); \
		} \
		if(cols() <= colIndex){ \
			throw std::invalid_argument("Matrix: column-index out of range"); \
		}
#endif
#endif

/**
 * This class represents a mathematical matrix of some data type T.
 * This can for instance be T=double, but all types are applicable 
 * that define:
 * - default-constructor, copy-constructor and operator=() 
 * - numerical operators +, -, *, /
 * - equality operators ==, !=
 * 
 * Each matrix has a certain row- and column-dimension, which is returned by the rows() and cols() methods.
 * The matrix contains rows()*cols() values of type T which can be accessed by an index
 * between zero and dimension-1.

TODO
 * The class provides basic access and manipulation of these values with the [] operator:
 * Matrix<int> x(3, 2);
 * int two = x[0];
 * for(unsigned int i=0; i<x.dimension(); i++){
 * 	x[i] = 42;
 * }
 */
template <class T>
class Matrix : public Collection<T> {
private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	//friend class MatrixMath<T>;
	//friend Matrix<T> pseudoInverse(Matrix<T>&);
	
	//template <typename D>
	//friend Matrix<D> matMatT(const Matrix<D>&);
	//template <typename D>
	//friend void _nemo_impl_matMatTRecursive(const Matrix<D>&,Matrix<D> &,unsigned int,
	//	unsigned int,unsigned int, unsigned int,unsigned int,bool);
	friend Matrix<T> matMatT<T>(const Matrix<T>&);
	friend void _nemo_impl_matMatTRecursive<T>(const Matrix<T>&,Matrix<T> &,unsigned int,
		unsigned int,unsigned int, unsigned int,unsigned int,bool);
	friend void _nemo_impl_matMatTBlockRecursive<T>(const Matrix<T>&,Matrix<T> &,unsigned int,
		unsigned int,unsigned int, unsigned int,unsigned int,unsigned int,bool);
	friend void _nemo_impl_matMatTBlockIterative<T>(const Matrix<T> &m, Matrix<T> &result);
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
	class MatrixData : public IntrusiveCounted<MatrixData > {
	public:
		MatrixData(const MatrixDimension &dimP)
		 : IntrusiveCounted<MatrixData >(), dim(dimP), mat()
		{
			// note eigen does only allow 0x0, but not 0xN matrices ^^
			if(dim.rows()>0&&dim.cols()>0){
				mat = Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic>((int)dim.rows(),(int)dim.cols());
			}
		}
		MatrixData(const Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> &matP)
		 : IntrusiveCounted<MatrixData >(), dim(nemo::dim((unsigned int)matP.rows(), (unsigned int)matP.cols())), mat(matP)
		{
		}

		MatrixData(const MatrixData &md)
		 : IntrusiveCounted<MatrixData >(), dim(md.dim), mat(md.mat)
		{}

		MatrixData& operator=(const MatrixData &r){
			if(this != &r){
				dim = r.dim;
				mat = r.mat;
			}
			return *this;
		}

		~MatrixData(){
		}
	public:
		mutable MatrixDimension dim;
		mutable Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> mat;
	};
#endif

public:
// =============================================================================
//    Construction
// =============================================================================

	/**
	 * Creates a matrix with dim(0,0).
	 */
	explicit Matrix()
	 : data(new MatrixData(dim(0,0)))
	{}

	/**
	 * Creates a matrix with specified dimension.
	 * All values inside are initialized with the default constructor,
	 * which, for instance, gives zeros for primitive numbers.
	 * 
	 * @param dim Dimension of the matrix
	 */
	explicit Matrix(const MatrixDimension& dim)
	 : data(new MatrixData(dim))
	{
		for(unsigned int r=0; r<dim.rows(); r++){
			for(unsigned int c=0; c<dim.cols(); c++){
				data->mat((int)r,(int)c) = T();
			}
		}
	}

	/**
	 * Creates a matrix with dimension dim, in which all values
	 * are initialized with value.
	 * 
	 * @param dim Dimension of the matrix
	 * @param value Initial value of all entries 
	 */
	explicit Matrix(const MatrixDimension &dim, const T &val)
	 : data(new MatrixData(dim))
	{
		for(unsigned int r=0; r<dim.rows(); r++){
			for(unsigned int c=0; c<dim.cols(); c++){
				// FIXME eigen is indexed with signed ints
				// => check maxDimension to assure no sign-conversion
				data->mat((int)r,(int)c) = val;
			}
		}
	}
	
	/**
	 * Creates a matrix with dimension dim, in which all values
	 * are initialized individually by the return value of gen().
	 * gen might, for instance be a random number generator from Random.h
	 * 
	 * @see MathVector::MathVector(VectorDimension, generatorType)
	 * 
	 * @param dim Dimension of the vector
	 * @param gen Value generator
	 */
	template <typename generatorType>
	explicit Matrix(const MatrixDimension &dim, generatorType gen)
	 : data(new MatrixData(dim))
	{
		for(unsigned int r=0; r<dim.rows(); r++){
			for(unsigned int c=0; c<dim.cols(); c++){
				data->mat((int)r,(int)c) = gen();
			}
		}
	}

#ifdef _NEMO_HAS_INIT_LISTS_
	Matrix(const std::initializer_list<std::initializer_list<T>> &init)
	 : data(new MatrixData(dim(0,0)))
	{
		unsigned int numRows = init.size();
		unsigned int numCols = 0;
		for(const std::initializer_list<T> *p = init.begin (); p != init.end (); ++p){
			if( (*p).size() > numCols ){
				numCols = (*p).size();
			}
		}
		data = IntrusiveCowSupport<MatrixData>(new MatrixData(dim(numRows,numCols)));
		if(numRows == 0 || numCols == 0){
			return; // no internal Eigen matrix created
		}
		unsigned int rowIndex = 0;
		for(const std::initializer_list<T> *rowP = init.begin(); rowP != init.end(); ++rowP){
			unsigned int colIndex = 0;
			for(const T *val = (*rowP).begin(); val != (*rowP).end(); ++val){
				data->mat((int)rowIndex,(int)(colIndex++)) = *val;
			}
			while(colIndex <= numCols-1){
				data->mat((int)rowIndex,(int)(colIndex++)) = T();
			}
			rowIndex++;
		}
	}
#endif

	/**
	 * Creates a matrix with dimension dim, in which elements on
	 * the main diagonal are inititialized with diagValue. All
	 * other elements are zero (initialized with T())
	 * 
	 * @param dim Dimension of the vector
	 * @param diagValue Value of the matrix elements on the main diagonal
	 */
	static Matrix<T> diagonal(const MatrixDimension &dim, const T &diagValue){
		Matrix<T> matrix(dim);
		unsigned int diagCount = (dim.rows()<dim.cols()) ? dim.rows() : dim.cols();
		
		for(unsigned int i=0; i<diagCount; i++){
			matrix.data->mat((int)i,(int)i) = diagValue;
		}
		return matrix;
	}

	/**
	 * Creates a matrix with dimension dim, in which elements on
	 * the main diagonal are inititialized with the values in diag. All
	 * other elements are zero (initialized with T())
	 * 
	 * @param dim Dimension of the vector
	 * @param diag Vector with diagonal values, must have dimension min(dim.rows(),dim.cols())
	 */
	static Matrix<T> diagonal(const MatrixDimension &dim, const MathVector<T> &diag){
		unsigned int diagCount = (dim.rows()<dim.cols()) ? dim.rows() : dim.cols();
		if(diagCount != diag.dimension()){
			_NEMO_ARGUMENT_ERROR("dimension mismatch in diagonal matrix contruction");
		}
		Matrix<T> matrix(dim);
		for(unsigned int i=0; i<diagCount; i++){
			matrix.data->mat((int)i,(int)i) = diag[i];
		}
		return matrix;
	}
	/**
	 * Creates a square-matrix, in which elements on
	 * the main diagonal are inititialized with the values in diag. All
	 * other elements are zero (initialized with T())
	 * 
	 * @param dim Dimension of the vector
	 * @param diag Vector with diagonal values, must have dimension min(dim.rows(),dim.cols())
	 */
	static Matrix<T> diagonal(const MathVector<T> &diag){
		Matrix<T> matrix(dim(diag.dimension(), diag.dimension()));
		for(unsigned int i=0; i<diag.dimension(); i++){
			matrix.data->mat((int)i,(int)i) = diag[i];
		}
		return matrix;
	}

	/**
	 * Creates a (size)x(size) identity-matrix (1's one the main diagonal, 0 everywhere else)
	 * 
	 * @param size 
	 */
	static Matrix<T> identity(unsigned int size){
		return Matrix<T>::diagonal(dim(size,size), (T)1);
	}

	// cppcheck-suppress operatorEq
	Matrix<T>& operator=(const Matrix<T> &m){
		this->data = m.data;
		return *this;
	}

//private:
	explicit Matrix(const Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> &m)
	 : data(new MatrixData(m))
	{}
	//explicit Matrix(Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> m)
	// : data(new MatrixData(m))
	//{}
// =============================================================================
//    Access
// =============================================================================
public:
	/**
	 * Returns the dimension of the matrix.
	 */
	MatrixDimension dimension() const {
		return data->dim;
	}

	/**
	 * Returns the number of rows.
	 */
	unsigned int rows() const {
		return data->dim.rows();
	}

	/**
	 * Returns the number of columns.
	 */
	unsigned int cols() const {
		return data->dim.cols();
	}

	Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> eigen() const {
		return data->mat;
	}

	/**
	 * Returns a readable reference to the element at position row/col
	 */
	const T& operator()(const unsigned int rowIndex,const unsigned int colIndex) const{
		_NEMO_MATRIX_CHECK_INDEX(rowIndex, colIndex);
		return data->mat((int)rowIndex,(int)colIndex);
	}

#ifndef NEMO_PARSED_BY_DOXYGEN
	class ReadWriteRef {
	friend class Matrix;
	private:
		ReadWriteRef(IntrusiveCowSupport<MatrixData > &d, const unsigned int r, const unsigned int c)
		 : data(d), rowIndex(r), colIndex(c) {}
	public:	
		ReadWriteRef& operator=(const T &value){ 
			// write access
			data.detach();
			data->mat((int)rowIndex,(int)colIndex) = value;
			return *this;
		}
		ReadWriteRef& operator=(ReadWriteRef &value){ 
			// write access
			data.detach();
			data->mat((int)rowIndex,(int)colIndex) = (T)value;
			return *this;
		}

		operator const T() const{return data->mat((int)rowIndex,(int)colIndex);}

		#define _nemo_matrix_ref_define_compound_(OP) \
		ReadWriteRef& operator OP (const T& value){\
			data.detach();\
			data->mat((int)rowIndex,(int)colIndex) OP value;\
			return *this;\
		}
		_nemo_matrix_ref_define_compound_(+=)
		_nemo_matrix_ref_define_compound_(-=)
		_nemo_matrix_ref_define_compound_(*=)
		_nemo_matrix_ref_define_compound_(/=)
		_nemo_matrix_ref_define_compound_(%=)
		_nemo_matrix_ref_define_compound_(&=)
		_nemo_matrix_ref_define_compound_(|=)
		_nemo_matrix_ref_define_compound_(^=)
		_nemo_matrix_ref_define_compound_(<<=)
		_nemo_matrix_ref_define_compound_(>>=)

		#define _nemo_matrix_ref_define_compare_(OP) \
		bool operator OP (const typename Matrix<T>::ReadWriteRef& rhs) const{\
			return data->mat((int)rowIndex,(int)colIndex) OP rhs.data->mat((int)rhs.rowIndex,(int)rhs.colIndex);\
		}
		_nemo_matrix_ref_define_compare_(==)
		_nemo_matrix_ref_define_compare_(!=)
		_nemo_matrix_ref_define_compare_(<=)
		_nemo_matrix_ref_define_compare_(<)
		_nemo_matrix_ref_define_compare_(>=)
		_nemo_matrix_ref_define_compare_(>)

		// this is a 'specialization' of std::swap that can be found by argument-dependent lookup.
		// this one is necessary for STL write operations to work correctly, since the standard swap
		// implementation does not work correctly on ReadWriteRef, since it already has reference semantics
		friend inline void swap( typename nemo::Matrix<T>::ReadWriteRef a, typename nemo::Matrix<T>::ReadWriteRef b ){
			a.data.detach();
			b.data.detach();
			using std::swap;
			swap( 
				a.data->mat((int)a.rowIndex,(int)a.colIndex), 
				b.data->mat((int)b.rowIndex,(int)b.colIndex) 
			);
		}
		friend inline T max( const typename nemo::Matrix<T>::ReadWriteRef& a, const typename nemo::Matrix<T>::ReadWriteRef& b ){
			using std::max;
			return max( 
				a.data->mat((int)a.rowIndex,(int)a.colIndex), 
				b.data->mat((int)b.rowIndex,(int)b.colIndex) 
			);
		}
		friend inline T max( const typename nemo::Matrix<T>::ReadWriteRef& a, T b ){
			using std::max;
			return max( a.data->mat((int)a.rowIndex,(int)a.colIndex), b );
		}
		friend inline T max( T a, const typename nemo::Matrix<T>::ReadWriteRef& b ){
			using std::max;
			return max( a, b.data->mat((int)b.rowIndex,(int)b.colIndex) );
		}
		friend inline T min( const typename nemo::Matrix<T>::ReadWriteRef& a, const typename nemo::Matrix<T>::ReadWriteRef& b ){
			using std::min;
			return min( 
				a.data->mat((int)a.rowIndex,(int)a.colIndex), 
				b.data->mat((int)b.rowIndex,(int)b.colIndex) 
			);
		}
		friend inline T min( const typename nemo::Matrix<T>::ReadWriteRef& a, T b ){
			using std::min;
			return min( a.data->mat((int)a.rowIndex,(int)a.colIndex), b );
		}
		friend inline T min( T a, const typename nemo::Matrix<T>::ReadWriteRef& b ){
			using std::min;
			return min( a, b.data->mat((int)b.rowIndex,(int)b.colIndex) );
		}
		
	private:
		IntrusiveCowSupport<MatrixData > &data;
		unsigned int rowIndex;
		unsigned int colIndex;
	};
#endif

	/**
	 * Returns a reference to the element at position row/col.
	 * Can be used for both read and write access
	 */
	ReadWriteRef operator()(const unsigned int rowIndex, const unsigned int colIndex){
		_NEMO_MATRIX_CHECK_INDEX(rowIndex,colIndex);
		return ReadWriteRef(data, rowIndex, colIndex);
	}


	/**
	 * Returns a sub-matrix of dimension d.
	 */
	Matrix<T> subMatrix(const unsigned int fromRow, const unsigned int fromCol, const MatrixDimension &d) const {
		checkSubsumes(fromRow, fromCol, d);

		// prevent Eigen from going mad about 0-dimensional matrices
		if( d.cols()==0 || d.rows()==0 ){
			return Matrix<T>(d);
		}

		Matrix<T> sub(d);
		sub.data->mat = this->data->mat.block((int)fromRow,(int)fromCol,(int)d.rows(),(int)d.cols());
		return sub;
	}

	/**
	 * Sets the sub-matrix (fromRow...fromRow+m.rows()-1,fromCol...fromCol+m.cols()-1) 
	 * to the values in m.
	 */
	void setSubMatrix(const unsigned int fromRow, const unsigned int fromCol, const Matrix<T> &m){
		MatrixDimension d = m.dimension();
		checkSubsumes(fromRow, fromCol, d);

		// prevent Eigen from going mad about 0-dimensional matrices
		if( d.cols()==0 || d.rows()==0 ){
			return;
		}

		this->data->mat.block((int)fromRow,(int)fromCol,(int)d.rows(),(int)d.cols())
			= m.data->mat;
	}

	/** Returns the colIndex'th column as vector */
	MathVector<T> col(unsigned int colIndex){
		if(cols() <= colIndex){
			_NEMO_ARGUMENT_ERROR("Matrix::col(uint): column index out of range");
		}
		MathVector<T> c(dim(this->rows()));
		for(unsigned int i=0; i<rows(); i++){
			c[i] = data->mat((int)i, (int)colIndex);
		}
		return c;	
	}

	/** Sets the values in the colIndex'th column to those in the vector */
	void setCol(unsigned int colIndex, const MathVector<T> &vec){
		if(cols() <= colIndex){
			_NEMO_ARGUMENT_ERROR("Matrix::setCol(uint,MathVector): column index out of range");
		}
		if(vec.dimension() != this->rows()){
			_NEMO_ARGUMENT_ERROR("Matrix::setCol(uint,MathVector): dimension mismatch");
		}

		for(unsigned int i=0; i<rows(); i++){
			data->mat((int)i, (int)colIndex) = vec[i];
		}
	}

	/** Returns the rowIndex'th row as transpose of a vector */
	VectorTranspose<T> row(unsigned int rowIndex){
		if(rows() <= rowIndex){
			_NEMO_ARGUMENT_ERROR("Matrix::row(uint): row index out of range");
		}
		MathVector<T> r(dim(this->cols()));
		for(unsigned int i=0; i<cols(); i++){
			r[i] = data->mat((int)rowIndex, (int)i);
		}
		return r.transpose();	
	}

	/** Sets the values in the rowIndex'th row to those in the vector-transpose */
	void setRow(unsigned int rowIndex, const VectorTranspose<T> &vec){
		if(rows() <= rowIndex){
			_NEMO_ARGUMENT_ERROR("Matrix::setRow(uint,MathVector): row index out of range");
		}
		if(vec.dimension() != this->cols()){
			_NEMO_ARGUMENT_ERROR("Matrix::setRow(uint,MathVector): dimension mismatch");
		}

		for(unsigned int i=0; i<cols(); i++){
			data->mat((int)rowIndex, (int)i) = vec[i];
		}
	}

// =============================================================================
//    Comparison
// =============================================================================
	/**
	 * Returns true if dimensions and all contained values are equal (i.e. exactly equal 
	 * in terms of T::operator==(T&) ).
	 */
	bool operator==(const Matrix<T> &o) const{
		if(data->dim != o.data->dim){
			return false;
		}
		for(unsigned int r=0; r<rows(); r++){
			for(unsigned int c=0; c<cols(); c++){
				if(data->mat((int)r,(int)c) != o.data->mat((int)r,(int)c)){
					return false;
				}
			}
		}
		// else
		return true;
	}

	/**
	 * ! operator==
	 */
	bool operator!=(const Matrix<T> &r) const{
		return !operator==(r);
	}

// =============================================================================
//    Collection API
// =============================================================================

	unsigned int numElements() const {
		return rows() * cols();
	}
	T getElement(unsigned int elementIndex) const{
		if(numElements() <= elementIndex){
			_NEMO_ARGUMENT_ERROR("element index out of range");
		}
		const unsigned int rowIndex = elementIndex / cols();
		return (*this)(rowIndex,elementIndex-rowIndex*cols());
	}
	void setElement(unsigned int elementIndex, T value){
		if(numElements() <= elementIndex){
			_NEMO_ARGUMENT_ERROR("element index out of range");
		}
		const unsigned int rowIndex = elementIndex / cols();
		(*this)(rowIndex,elementIndex-rowIndex*cols()) = value;
	}

	void transform(Mapping<T,T> m){
		for(unsigned int i=0; i<rows(); i++){
			for(unsigned int k=0; k<cols(); k++){
				(*this)(i,k) = m((*this)(i,k));
			}
		}
	}

	template <typename OutType>
	Matrix<OutType> map(Mapping<T,OutType> m){
		Matrix<OutType> out(this->dimension());
		for(unsigned int i=0; i<rows(); i++){
			for(unsigned int k=0; k<cols(); k++){
				out(i,k) = m((*this)(i,k));
			}
		}
		return out;
	}

// =============================================================================
//    STL compliant iterator interface
// =============================================================================

#ifndef NEMO_PARSED_BY_DOXYGEN
	class Iterator{
	private:
		Iterator(const Matrix<T> *matP, int indexP) : mat(matP), index(indexP), rwRef(const_cast<Matrix<T>*>(matP)->data,0,0) {}
	public:
		Iterator(const Iterator& it)
		 : mat(it.mat), index(it.index), 
		   rwRef(const_cast<Matrix<T>*>(it.mat)->data,(unsigned int)it.rwRef.rowIndex,(unsigned int)it.rwRef.colIndex) 
		{}
		Iterator& operator=(const Iterator &it){
			if(this==&it){return *this;}
			this->mat = it.mat;
			this->index = it.index;
			this->rwRef.data = it.rwRef.data;
			this->rwRef.rowIndex = it.rwRef.rowIndex;
			this->rwRef.colIndex = it.rwRef.colIndex;
			return *this;
		}
	
		typedef int difference_type;
		typedef T value_type;
		typedef ReadWriteRef* pointer;
		typedef ReadWriteRef reference;
		typedef std::random_access_iterator_tag iterator_category;

		ReadWriteRef& operator*(){
			if(index<0 || (int)mat->rows()*(int)mat->cols()<=index){_NEMO_LOGIC_ERROR("Iterator out of range");}
			rwRef.colIndex = (unsigned int)(index%(int)mat->cols());
			rwRef.rowIndex = (unsigned int)(index/(int)mat->cols());
			return rwRef;
			//return (*const_cast<MathVector<T>*>(vec))[(unsigned int)index];
		}
		T operator*() const{
			if(index<0 || (int)mat->rows()*(int)mat->cols()<=index){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return mat->getElement((unsigned int)index);
		}
		ReadWriteRef& operator[](difference_type offset){
			if(index+offset<0 || (int)mat->rows()*(int)mat->cols()<=index+offset){_NEMO_LOGIC_ERROR("Iterator out of range");}
			rwRef.colIndex = (unsigned int)(index%(int)mat->cols());
			rwRef.rowIndex = (unsigned int)(index/(int)mat->cols());
			rwRef.index = (unsigned int)index+offset;
			return rwRef;
			//return (*const_cast<MathVector<T>*>(vec))[(unsigned int)index+offset];
		}
		T operator[](difference_type offset) const{
			if(index+offset<0 || (int)mat->rows()*(int)mat->cols()<=index+offset){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return mat->getElement((unsigned int)index+offset);
		}
		Iterator& operator++() const{
			index++;
			return const_cast<Iterator&>(*this);
		}
		Iterator operator++(int) const{
			Iterator copy = *this;
			index++;
			return copy;
		}
		Iterator& operator--() const{
			index--;
			return const_cast<Iterator&>(*this);
		}
		Iterator operator--(int) const{
			Iterator copy = *this;
			index--;
			return copy;
		}
		Iterator operator+(difference_type off) const {
			Iterator it = *this;
			it.index += off;
			return it;
		}
		Iterator &operator+=(difference_type off) {
			index += off;
			return *this;
		}
		Iterator &operator-=(difference_type off) {
			index -= off;
			return *this;
		}
		Iterator operator-(difference_type off) const {
			Iterator it = *this;
			it.index -= off;
			return it;
		}
		difference_type operator-(const Iterator& rhs) const {
			return index - rhs.index;
		}
		bool operator==(const Iterator &it) const{
			return (mat==it.mat) && (index==it.index);
		}
		bool operator!=(const Iterator &it) const{
			return !(*this==it);
		}
		bool operator<=(const Iterator &it) const{
			return index<=it.index;
		}
		bool operator<(const Iterator &it) const{
			return index<it.index;
		}
		bool operator>=(const Iterator &it) const{
			return index>=it.index;
		}
		bool operator>(const Iterator &it) const{
			return index>it.index;
		}
	private:
		friend class Matrix<T>;
		const Matrix<T> *mat;
		mutable int index;
		mutable ReadWriteRef rwRef;
	};
#endif

	/**
	 * Returns an STL-compliant (writable) iterator pointing to the first value, going through the matrix row by row.
	 */
	Iterator begin(){
		return Iterator(this,0);
	}
	/**
	 * Returns an STL-compliant (writable) iterator pointing one position after the last element.
	 */
	Iterator end(){
		return Iterator(this,(int)rows()*(int)cols());
	}

	const Iterator begin() const{
		return Iterator(this,0);
	}
	const Iterator end() const{
		return Iterator(this,(int)rows()*(int)cols());
	}


// =============================================================================
//    Unary Matrix operations
// =============================================================================

	Matrix<T> transpose() const {
		Matrix<T> t(dim(this->cols(), this->rows()));
		t.data->mat = this->data->mat.transpose();
		return t;		
	}


// =============================================================================
//    Matrix-Matrix operations
// =============================================================================

	/**
	 * Returns a new matrix that contains the concatenated rows of this matrix and the argument.
	 */
	const Matrix<T> attachRows(const Matrix<T> &o) const {
		checkColumnDimensionMatch(o);

		Matrix<T> newMat(dim(rows()+o.rows(), cols()));
		if(rows()>0 && cols()>0){
			newMat.data->mat.block(0,0,(int)rows(),(int)cols()) = this->data->mat;
		}
		if(o.rows()>0 && o.cols()>0){
			newMat.data->mat.block((int)rows(),0,(int)o.rows(),(int)o.cols()) = o.data->mat;
		}
		return newMat;
	}

	/**
	 * Returns a new matrix that contains the concatenated columns of this matrix and the argument.
	 */
	const Matrix<T> attachCols(const Matrix<T> &o) const {
		checkRowDimensionMatch(o);

		Matrix<T> newMat(dim(rows(), cols()+o.cols()));
		if(rows()>0 && cols()>0){
			newMat.data->mat.block(0,0,(int)rows(),(int)cols()) = this->data->mat;
		}
		if(o.rows()>0 && o.cols()>0){
			newMat.data->mat.block(0,(int)cols(),(int)o.rows(),(int)o.cols()) = o.data->mat;
		}
		return newMat;
	}

private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	class MatrixMult{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			//if(m1.rows()!=0 && m2.cols()!=0 &&  m1.cols()==0){
			//	result.data->mat = Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>::Zero(m1.rows(), m2.cols());
			//}
			if(m1.rows()!=0 && m2.cols()!=0 &&  m1.cols()==0){
				result.data->mat = Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>::Zero(m1.rows(), m2.cols());
			}
			else if( m1.rows()==0 || m2.cols()==0 ){
				// zero-dim
				result.data->mat = Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>();
			}
			else{
				result.data->mat = m1.data->mat * m2.data->mat;
			}
		}
	};
	class MatrixAdd{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat + m2.data->mat;
		}
	};
	class MatrixDiff{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat - m2.data->mat;
		}
	};
	class MatrixCWiseMult{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat.cwise() * m2.data->mat;
		}
	};
	class MatrixCWiseDiv{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat.cwise() / m2.data->mat;
		}
	};
	class MatrixCWiseMax{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat.cwise().max( m2.data->mat );
		}
	};
	class MatrixCWiseMin{
	public:
		static void evaluate(const Matrix<T> &m1, const Matrix<T> &m2, Matrix<T> &result){
			result.data->mat = m1.data->mat.cwise().min( m2.data->mat );
		}
	};
#endif

public:
	/**
	 * Multiplies this matrix with another matrix and returns the result.
	 */
	Matrix<T> operator*(const Matrix<T> &r) const{
		if(this->cols() != r.rows()){
			_NEMO_ARGUMENT_ERROR("Matrix multiplication: dimension mismatch");
		}
		// FIXME [n x 0]*[0 x m] must not evaluate to Eigen::mat [0 x 0]
		Matrix<T> result(dim(this->rows(), r.cols()));
		MatrixMult::evaluate(*this, r, result);
		return result;
	}
	/**
	 * Multiplies this matrix with another matrix in place.
	 * If the argument matrix is not square, this operation will change the dimension of this matrix.
	 */
	Matrix<T>& operator*=(const Matrix<T> &r){
		if(this->cols() != r.rows()){
			_NEMO_ARGUMENT_ERROR("Matrix multiplication: dimension mismatch");
		}
		data.detach();
		data->dim = dim(this->rows(), r.cols());
		//MatrixDimension newDim = dim(this->rows(), r.cols());
		//data->dim = newDim;
		MatrixMult::evaluate(*this, r, *this);
		return *this;
	}
	/**
	 * Adds this matrix to another matrix with same dimension and returns the result.
	 */
	Matrix<T> operator+(const Matrix<T> &r) const{
		checkDimensionMatch(r);
		Matrix<T> result(dimension());
		MatrixAdd::evaluate(*this, r, result);
		return result;
	}
	/**
	 * Adds the argument to this matrix, in place.
	 */
	Matrix<T>& operator+=(const Matrix<T> &r){
		checkDimensionMatch(r);
		data.detach();
		MatrixAdd::evaluate(*this, r, *this);
		return *this;
	}
	/**
	 * Subtracts another matrix with same dimension from this matrix and returns the result.
	 */
	Matrix<T> operator-(const Matrix<T> &r) const{
		checkDimensionMatch(r);
		Matrix<T> result(dimension());
		MatrixDiff::evaluate(*this, r, result);
		return result;
	}
	/**
	 * Subtracts another matrix with same dimension from this matrix, in place.
	 */
	Matrix<T>& operator-=(const Matrix<T> &r){
		checkDimensionMatch(r);
		data.detach();
		MatrixDiff::evaluate(*this, r, *this);
		return *this;
	}

// =============================================================================
//    Matrix-Matrix component-wise operations
// =============================================================================

	/**
	 * This class provides component-wise mathematical operations on matrices
	 * that are not covered by standard operators/methods in the Matrix class.
	 */
	class CWise {
	friend class Matrix;
	private:
		CWise(Matrix<T> &m) : mat(m) {}
		CWise(const CWise &o) : mat(o.mat) {}
		CWise& operator=(const CWise &o){this->mat=o.mat; return *this;}

		Matrix<T> &mat;
	public:
		/** Returns result of component-wise multiplication.
		 */
		Matrix<T> operator*(const Matrix<T> &o) const {
			mat.checkDimensionMatch(o);
			Matrix<T> result(o.data->dim);
			MatrixCWiseMult::evaluate(mat, o, result);
			return result;
		}
		/** Computes component-wise multiplication in place.
		 */
		typename Matrix<T>::CWise& operator*=(const Matrix<T> &o){
			mat.checkDimensionMatch(o);
			mat.data.detach();
			MatrixCWiseMult::evaluate(mat, o, mat);
			return *this;
		}
		/** Returns result of component-wise Division.
		 */
		Matrix<T> operator/(const Matrix<T> &o) const {
			mat.checkDimensionMatch(o);
			Matrix<T> result(o.data->dim);
			MatrixCWiseDiv::evaluate(mat, o, result);
			return result;
		}
		/** Computes component-wise division in place.
		 */
		typename Matrix<T>::CWise& operator/=(const Matrix<T> &o){
			mat.checkDimensionMatch(o);
			mat.data.detach();
			MatrixCWiseDiv::evaluate(mat, o, mat);
			return *this;
		}
		/** Returns component-wise maximum of this vector and the argument.
		 */
		Matrix<T> max(const Matrix<T> &o) const {
			mat.checkDimensionMatch(o);
			Matrix<T> result(mat.data->dim);
			MatrixCWiseMax::evaluate(mat, o, result);
			return result;
		}
		/** Returns component-wise minimum of this vector and the argument.
		 */
		Matrix<T> min(const Matrix<T> &o) const {
			mat.checkDimensionMatch(o);
			Matrix<T> result(mat.data->dim);
			MatrixCWiseMin::evaluate(mat, o, result);
			return result;
		}
		operator const Matrix<T>(){return mat;}
	};

	/**
	 * Get a CWise object to perform component-wise operations on this vector.
	 */
	CWise cwise(){
		return CWise(*this);
	}
	const CWise cwise() const {
		return CWise(*this);
	}

// =============================================================================
//    Matrix unary operations
// =============================================================================
private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	class MatrixMinus{
	public:
		static void evaluate(const Matrix<T> &v, Matrix<T> &result){
			result.data->mat = - v.data->mat;
		}
	};
#endif

public:
	Matrix<T> operator-() const {
		Matrix<T> result(dimension());
		MatrixMinus::evaluate(*this, result);
		return result;
	}
	Matrix<T> operator+() const{
		return *this;
	}

// =============================================================================
//    Matrix-Vector operations
// =============================================================================
private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	class MatrixMultVector{
	public:
		static void evaluate(const Matrix<T> &m, const MathVector<T> &v, MathVector<T> &result){
			// FIXME performance: no index checking!!!
			for(unsigned int row = 0; row < m.rows(); row++){
				result[row] = 0.0;
				for(unsigned int col = 0; col < m.cols(); col++){
					result[row] += m(row,col) * v[col];
				}
			}
		}
	};
#endif

public:
	/**
	 * Multiply this matrix with a vector
	 */
	MathVector<T> operator*(const MathVector<T> &r) const{
		checkColumnDimensionMatch(r);
		MathVector<T> result(dim(rows()));
		MatrixMultVector::evaluate(*this, r, result);
		return result;
	}

// =============================================================================
//    Matrix-Scalar operations
// =============================================================================
private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	class MatrixMultScalar{
	public:
		static void evaluate(const Matrix<T> &m, const T &s, Matrix<T> &result){
			for(unsigned int row = 0; row < m.rows(); row++){
				for(unsigned int col = 0; col < m.cols(); col++){
					// FIXME performance: no index checking!!!
					result(row,col) = m(row,col) * s;
				}
			}
		}
	};
#endif

public:
	/**
	 * Multiply this matrix with a scalar and return result
	 */
	Matrix<T> operator*(const T &r) const{
		Matrix<T> result(this->dimension());
		MatrixMultScalar::evaluate(*this, r, result);
		return result;
	}
	/**
	 * Multiply a scalar and a matrix
	 */
	friend Matrix<T> operator*(const T &s, const Matrix<T> &m){
		Matrix<T> result(m.dimension());
		MatrixMultScalar::evaluate(m, s, result);
		return result;
	}
	/**
	 * Multiply this matrix in place with a scalar 
	 */
	Matrix<T>& operator*=(const T &r){
		data.detach();
		MatrixMultScalar::evaluate(*this, r, *this);
		return *this;
	}

private:
	
	inline void checkColumnDimensionMatch(const Matrix<T> &r) const{
		if(cols() != r.cols()){
			_NEMO_ARGUMENT_ERROR("Matrix: dimension mismatch");
		}
	}
	inline void checkColumnDimensionMatch(const MathVector<T> &v) const{
		if(cols() != v.dimension()){
			_NEMO_ARGUMENT_ERROR("Matrix/Vector: dimension mismatch");
		}
	}
	inline void checkRowDimensionMatch(const Matrix<T> &r) const{
		if(rows() != r.rows()){
			_NEMO_ARGUMENT_ERROR("Matrix: dimension mismatch");
		}
	}

	/*inline void checkIndex(const unsigned int rowIndex, const unsigned int colIndex) const{
		if(rows() <= rowIndex){
			_NEMO_ARGUMENT_ERROR("Matrix: row-index out of range");
		}
		if(cols() <= colIndex){
			_NEMO_ARGUMENT_ERROR("Matrix: column-index out of range");
		}
	}*/
	template <class D>
	inline void checkDimensionMatch(const Matrix<D> &r) const{
		if(dimension() != r.dimension()){
			std::stringstream msg;
			msg << "Matrix: dimension mismatch: argument should have "
			    << "dimension (" << dimension().rows() << "x" << dimension().cols() << ") "
			    << "but has (" << r.dimension().rows() << "x" << r.dimension().cols() << ")";
			_NEMO_ARGUMENT_ERROR(msg.str());
		}
	}
	inline void checkSubsumes(const unsigned int fromRow, const unsigned int fromCol, MatrixDimension d) const {
		// check index (allowing index 0 for 0-dimension)
		if( this->rows() < fromRow+d.rows() ){
			_NEMO_ARGUMENT_ERROR("");
		}
		if( this->cols() < fromCol+d.cols() ){
			_NEMO_ARGUMENT_ERROR("");
		}
	}
	
	IntrusiveCowSupport<MatrixData > data;
	
};

// =============================================================================
//    Inversion rules
// =============================================================================

#define _nemo_matrix_inversion_(TYPE) \
template <> inline Matrix<TYPE> leftAddInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return -value;}  \
template <> inline Matrix<TYPE> rightAddInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return -value;}  \
template <> inline Matrix<TYPE> leftSubtractInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return value;}  \
template <> inline Matrix<TYPE> rightSubtractInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return -value;} 

_nemo_matrix_inversion_(float)
_nemo_matrix_inversion_(double)
_nemo_matrix_inversion_(long double)
//_nemo_matrix_inversion_(short)
_nemo_matrix_inversion_(int)
//_nemo_matrix_inversion_(long int)



// =============================================================================
//    Vector outer-product
// =============================================================================


template <class T>
Matrix<T> operator*(const MathVector<T> &l, const VectorTranspose<T> &r){
	Matrix<T> m(dim(l.dimension(), r.dimension()));
	for(unsigned int i=0; i<l.dimension(); i++){
		for(unsigned int k=0; k<r.dimension(); k++){
			m(i,k) = l[i] * r[k];
		}
	}
	return m;
}



// =============================================================================
//    Streaming
// =============================================================================

template <class T>
std::ostream& operator<<(std::ostream& out, const Matrix<T> &mat){

	MatrixDimension dim = mat.dimension();

	if(dim.rows() == 0 || dim.cols() == 0){
		out << std::string("[]");
	}
	/*else if(dim.rows() == 0){
		out << std::string("[]");
	}
	else if(dim.cols() == 0){
		out << std::string("[");
		for(unsigned int r = 0; r < dim.rows()-1; r++){
			out << std::string(";") << std::endl;
		}
		out << std::string(";]");
	}*/
	else{
		out << std::string("["); 
		for(unsigned int r=0; r<dim.rows(); r++){
			if(r!=0){
				out << std::string(" ");
			}
		
			for(unsigned int c=0; c<dim.cols()-1; c++){
				out << mat(r,c) << ",\t";
			}
			out << mat(r, dim.cols()-1);
		
			if(r < dim.rows()-1){
				out << ";" << std::endl;
			}
		}
		out << std::string("]"); 
	}

	return out;
}


typedef Matrix<double> RealMatrix;


} // end namespace


#endif

