#ifndef _NEMO_MATRIX_MATH_H_
#define _NEMO_MATRIX_MATH_H_

#include "Matrix.h"

#include "Norms.h"

#include <Eigen/Array> 
#include <Eigen/SVD>
#include <Eigen/Core>

namespace nemo {


/**
 * Implementation of Singular Value Decomposition.
 * Decomposes a matrix M into U, S, and V such that
 * 
 * M = U S V^T
 * 
 * U and V are unitary matrices, i.e. they have orthonormal rows.
 * S is a square diagonal matrix with the singular values on its main diagonal.
 * 
 * For an input matrix M with dimension (n)x(m) and l=min(n,m) the dimensions of
 * the decomposition are  U in (n)x(l), S in (l)x(l) and V in (m)x(l).
 */
template <typename T>
class SVD {
public:
	SVD(Matrix<T> mat) {

		// lets first work around some Eigen "special cases"
		// Error message 1: "In Eigen 2.0, SVD doesn't work on 1x1 matrices"
		// Error message 2: "In Eigen 2.0, SVD only works for MxN matrices with M>=N. Sorry!"

		if(mat.rows() < mat.cols()){
			SVD<T> svdT(mat.transpose());
			singVals = svdT.singVals;
			matU = svdT.matV;
			matV = svdT.matU;
		}
		else if(mat.cols() == 0){
			singVals = MathVector<T>();
			//matU = Matrix<T>::identity(mat.rows());
			matU = Matrix<T>(dim(mat.rows(), 0));
			matV = Matrix<T>();
		}
		else if(mat.cols() == 1){
			Mapping<MathVector<T>,T> normV = euclideanNorm<MathVector<T>,T>();
			Mapping<Matrix<T>,T> normM = euclideanNorm<Matrix<T>,T>();

			const unsigned int n = mat.rows();
			const unsigned int m = mat.cols();
			const unsigned int l = (m<n)?m:n;
			
			singVals = MathVector<T>(dim(l), T());
			singVals[0] = normM(mat);

			if(singVals[0] == T()){
				//matU = Matrix<T>::identity(mat.cols());
				matU = Matrix<T>::diagonal(dim(n,l), (T)1);
			}
			else{
				matU = Matrix<T>(dim(n,l));
				matU.setCol(0, mat.col(0) * ( ((T)1)/normV(mat.col(0))) ) ;

				//Matrix<T> matUT = Matrix<T>::identity(mat.rows());
				//matUT.setCol(0, mat.col(0) * (1.0/normV(mat.col(0))) ) ;
				
				// Gram-Schmidt orthonormalization
				/*for(unsigned int col = 1; col < matUT.cols(); col++){
					while(true){
						for(int c = col-1; 0 <= c; c--){
							T scalarProd = matUT.col(c).transpose() * matUT.col(col);
							matUT.setCol(col, matUT.col(col) - scalarProd*matUT.col(c));
						}
						if( T() < normV(matUT.col(col))){
							matUT.setCol(col, matUT.col(col) * ((T)1 / normV(matUT.col(col))) );
							break;
						}
						else{
							// zero-norm: this column was already contained in space of prev. cols
							matUT.setCol(col, MathVector<T>(dim(matUT.rows()), uniform<T>(T(),(T)1)) );
						}
					}
				}
				matU = matUT.transpose();*/
				
			}
			
			matV = Matrix<T>(dim(1,1), (T)1);
		}
		else{
			// FIXED: Eigen2's SVD may loop forever, if it get NaN of infinity values (for generic T, only NaN testable)
			for(unsigned int i=0; i<mat.rows(); i++){
				for(unsigned int k=0; k<mat.cols(); k++){
					T value = mat(i,k);
					//if(value != value){ // this one generically tests for NaN
					// this one generically tests for both NaN and Infty (pretty cool, actually)
					if( (value * (T)0) != (T)0 ){
						_NEMO_ARGUMENT_ERROR("SVD: matrix to be inverted contains NaN or Infty values");
					}
				}
			}

			Eigen::SVD<Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic> > svdEigen( mat.eigen() );

			Eigen::Matrix<T, Eigen::Dynamic, 1> singValsEigen = svdEigen.singularValues();
			singVals = MathVector<T>(dim(singValsEigen.size()));
			for(unsigned int i=0; i<singVals.dimension(); i++){
				singVals[i] = singValsEigen[i];
			}
			matU = Matrix<T>(svdEigen.matrixU());
			matV = Matrix<T>(svdEigen.matrixV());
		}
			
	}

	/**
	 * Returns the singular values in a vector.
	 */
	MathVector<T> singularValues() const{
		return singVals;
	}

	/**
	 * Returns the singular values in the decompositions's S matrix
	 */
	Matrix<T> matrixS() const {
		//return Matrix<T>::diagonal(dim(matU.cols(),matV.cols()), singVals);
		return Matrix<T>::diagonal(dim(singVals.dimension(),singVals.dimension()), singVals);
	}

	/**
	 * Returns the decomposition's U matrix.
	 */
	Matrix<T> matrixU() const {
		return matU;
	}

	/**
	 * Returns the decomposition's V matrix.
	 */
	Matrix<T> matrixV() const {
		return matV;
	}

private:
	MathVector<T> singVals;
	Matrix<T> matU;
	Matrix<T> matV;
};


/**
 * Eigen decomposition of a symmetric matrix M into Q D Q^T, where
 * the orthonormal columns of the square matrix Q are the Eigen-vectors
 * of M, and the entries of the diagonal matrix D are the corresponding Eigen-values.
 */
template <typename T>
class SymmetricEigenDecomposition {
public:
	SymmetricEigenDecomposition(Matrix<T> mat) {
		// check symmetry
		for(unsigned int i=0; i<mat.rows(); i++){
			for(unsigned int j=i; j<mat.cols(); j++){
				if(mat(i,j) != mat(j,i)){
					_NEMO_ARGUMENT_ERROR("SymmetricEigenDecomposition: input matrix is not symmetric");
				}
			}
		}
		// SVD based implementation based on 
		//http://math.stackexchange.com/questions/28036/relationship-between-eigendecomposition-and-singular-value-decomposition
		SVD<T> svd(mat);
		eigVecs = svd.matrixU();
		eigVals = svd.singularValues();
		for(unsigned int i=0; i<eigVals.dimension(); i++){
			// sign of Eigen values determined by scalarProd(matrixU_i, matrixV_i) = {1;-1}
			T prod = eigVecs.col(i).transpose() * svd.matrixV().col(i);
			assert( ((T)-1.00001<=prod && prod<=-(T)0.99999) || ((T)0.99999<=prod && prod<=(T)1.00001) );
			if( eigVecs.col(i).transpose() * svd.matrixV().col(i) < (T)0 ){
				eigVals[i] = -eigVals[i];
			}
		}
	}

	/**
	 * Returns the singular values in a vector.
	 */
	MathVector<T> eigenValues() const{
		return eigVals;
	}

	/**
	 * Returns the matrix D with the Eigen values on the main diagonal.
	 */
	Matrix<T> matrixD() const{
		return Matrix<T>::diagonal(eigVals);
	}

	/**
	 * Returns the square matrix Q with the Eigen vectors in its columns
	 */
	Matrix<T> matrixQ() const{
		return eigVecs;
	}

private:
	MathVector<T> eigVals;
	Matrix<T> eigVecs;
};


/**
 * An SVD based implementation of the Moore-Penrose pseudo-inverse
 */
template <typename T>
static const Matrix<T> pseudoInverse(const Matrix<T> &m, T epsilon = (T)1E-9) {

	SVD<T> svd(m);
	Matrix<T> matU = svd.matrixU();
	MathVector<T> singVals = svd.singularValues();
	Matrix<T> matV = svd.matrixV();
	
	T maxSingVal = ((T)0.0);
	for(unsigned int i=0; i<singVals.dimension(); i++){
		if(maxSingVal < singVals[i]){
			maxSingVal = singVals[i];
		}
	}
	for(unsigned int i=0; i<singVals.dimension(); i++){
		if(singVals[i] <= epsilon*maxSingVal){
			singVals[i] = ((T)0.0); // FIXED can not be safely inverted
		}
		else{
			singVals[i] = ((T)1.0) / singVals[i];
		}
	}

	return matV * Matrix<T>::diagonal(singVals) * matU.transpose();
}

#define _nemo_matrix_mult_inversion_(TYPE) \
template <> inline Matrix<TYPE> leftMultInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return pseudoInverse(value);}  \
template <> inline Matrix<TYPE> rightMultInverse<Matrix<TYPE> >(const Matrix<TYPE> &value){return pseudoInverse(value);} 

#ifndef NEMO_PARSED_BY_DOXYGEN
_nemo_matrix_mult_inversion_(float)
_nemo_matrix_mult_inversion_(double)
_nemo_matrix_mult_inversion_(long double)
#endif


#ifndef NEMO_PARSED_BY_DOXYGEN
namespace detail {
// compute integer log2
static inline unsigned int ilog2(register unsigned int x)
{
    register unsigned int l=0;
    if(x >= 1<<16) { x>>=16; l|=16; }
    if(x >= 1<<8) { x>>=8; l|=8; }
    if(x >= 1<<4) { x>>=4; l|=4; }
    if(x >= 1<<2) { x>>=2; l|=2; }
    if(x >= 1<<1) l|=1;
    return l;
}
} // end namespace
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
template <typename T>
inline void _nemo_impl_matMatTRecursive( 
		const Matrix<T> &m, Matrix<T> &result, unsigned int numCols,
		unsigned int fromRowLeft, unsigned int toRowLeft, 
		unsigned int fromRowRight, unsigned int toRowRight,
		bool computeLowerTriangle
){
	/*if(toRowLeft-fromRowLeft <= 2){
		for(unsigned int leftRowIndex = fromRowLeft; leftRowIndex<toRowLeft; leftRowIndex++){
			for(unsigned int rightColIndex = fromRowRight; rightColIndex<toRowRight; rightColIndex++){
				double value = 0.0;
				for(unsigned int i=0; i<numCols; i++){
					value += m.data->mat(leftRowIndex,i) * m.data->mat(rightColIndex,i);
				}
				result.data->mat(leftRowIndex,rightColIndex) = value;
				result.data->mat(rightColIndex,leftRowIndex) = value;
			}
		}
	}*/
	if(toRowLeft-fromRowLeft == 1){
		for(unsigned int rightColIndex = fromRowRight; rightColIndex<toRowRight; rightColIndex++){
			double value = 0.0;
			for(unsigned int i=0; i<numCols; i++){
				value += m.data->mat(fromRowLeft,i) * m.data->mat(rightColIndex,i);
			}
			result.data->mat(fromRowLeft,rightColIndex) = value;
			result.data->mat(rightColIndex,fromRowLeft) = value;
		}
	}
	else if(toRowLeft-fromRowLeft == 2){
		for(unsigned int rightColIndex = fromRowRight; rightColIndex<toRowRight; rightColIndex++){
			double value1 = 0.0;
			double value2 = 0.0;
			for(unsigned int i=0; i<numCols; i++){
				value1 += m.data->mat(fromRowLeft,i) * m.data->mat(rightColIndex,i);
				value2 += m.data->mat(fromRowLeft+1,i) * m.data->mat(rightColIndex,i);
			}
			result.data->mat(fromRowLeft,rightColIndex) = value1;
			result.data->mat(rightColIndex,fromRowLeft) = value1;
			result.data->mat(fromRowLeft+1,rightColIndex) = value2;
			result.data->mat(rightColIndex,fromRowLeft+1) = value2;
		}
	}
	else if(toRowLeft-fromRowLeft == 4){
		for(unsigned int rightColIndex = fromRowRight; rightColIndex<toRowRight; rightColIndex++){
			double value1 = 0.0;
			double value2 = 0.0;
			double value3 = 0.0;
			double value4 = 0.0;
			for(unsigned int i=0; i<numCols; i++){
				value1 += m.data->mat(fromRowLeft,i) * m.data->mat(rightColIndex,i);
				value2 += m.data->mat(fromRowLeft+1,i) * m.data->mat(rightColIndex,i);
				value3 += m.data->mat(fromRowLeft+2,i) * m.data->mat(rightColIndex,i);
				value4 += m.data->mat(fromRowLeft+3,i) * m.data->mat(rightColIndex,i);
			}
			result.data->mat(fromRowLeft,rightColIndex) = value1;
			result.data->mat(rightColIndex,fromRowLeft) = value1;
			result.data->mat(fromRowLeft+1,rightColIndex) = value2;
			result.data->mat(rightColIndex,fromRowLeft+1) = value2;
			result.data->mat(fromRowLeft+2,rightColIndex) = value3;
			result.data->mat(rightColIndex,fromRowLeft+2) = value3;
			result.data->mat(fromRowLeft+3,rightColIndex) = value4;
			result.data->mat(rightColIndex,fromRowLeft+3) = value4;
		}
	}
	else{
		//unsigned int leftDiv = fromRowLeft + (toRowLeft-fromRowLeft)/2;
		//unsigned int rightDiv = fromRowRight + (toRowRight-fromRowRight)/2;
		unsigned int leftDiv = fromRowLeft + (1 << (detail::ilog2(toRowLeft-fromRowLeft)-1));
		unsigned int rightDiv = fromRowRight + (1 << (detail::ilog2(toRowRight-fromRowRight)-1));
		// top left
		_nemo_impl_matMatTRecursive(m,result,numCols,fromRowLeft,leftDiv,fromRowRight,rightDiv,computeLowerTriangle);
		if(computeLowerTriangle){
			// lower left
			_nemo_impl_matMatTRecursive(m,result,numCols,leftDiv,toRowLeft,fromRowRight,rightDiv, true);
		}
		// top right
		_nemo_impl_matMatTRecursive(m,result,numCols,fromRowLeft,leftDiv,rightDiv,toRowRight,true);
		// lower right
		_nemo_impl_matMatTRecursive(m,result,numCols,leftDiv,toRowLeft,rightDiv,toRowRight,computeLowerTriangle);
	}
}
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
template <typename T>
inline void _nemo_impl_matMatTBlockRecursive( 
		const Matrix<T> &m, Matrix<T> &result,
		unsigned int fromRowLeft, unsigned int toRowLeft, 
		unsigned int fromRowRight, unsigned int toRowRight, 
		unsigned int fromColLeft, unsigned int toColLeft,
		bool computeLowerTriangle
){
	assert( fromRowLeft < toRowLeft );
	assert( fromRowRight < toRowRight );
	assert( fromColLeft < toColLeft );
	unsigned int rowDiffLeft = toRowLeft - fromRowLeft;
	unsigned int rowDiffRight = toRowRight - fromRowRight;
	//assert( rowDiffLeft == rowDiffRight );
	//unsigned int rowDiff = rowDiffLeft;
	unsigned int colDiff = toColLeft - fromColLeft;

	/*if(rowDiffLeft==1 && rowDiffRight==1 && colDiff==1){
		result.data->mat(fromRowLeft,fromRowRight)
			 += m.data->mat(fromRowLeft,fromColLeft) * m.data->mat(fromRowRight,fromColLeft);
	}*/
	/*if(rowDiffLeft<=8 && rowDiffRight<=8 && colDiff==8){
		for(unsigned int rowLeft=fromRowLeft; rowLeft<toRowLeft; rowLeft++){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
			T v0 = m.data->mat(rowLeft,fromColLeft) * m.data->mat(rowRight,fromColLeft);
			T v1 = m.data->mat(rowLeft,fromColLeft+1) * m.data->mat(rowRight,fromColLeft+1);
			T v2 = m.data->mat(rowLeft,fromColLeft+2) * m.data->mat(rowRight,fromColLeft+2);
			T v3 = m.data->mat(rowLeft,fromColLeft+3) * m.data->mat(rowRight,fromColLeft+3);
			T v4 = m.data->mat(rowLeft,fromColLeft+4) * m.data->mat(rowRight,fromColLeft+4);
			T v5 = m.data->mat(rowLeft,fromColLeft+5) * m.data->mat(rowRight,fromColLeft+5);
			T v6 = m.data->mat(rowLeft,fromColLeft+6) * m.data->mat(rowRight,fromColLeft+6);
			T v7 = m.data->mat(rowLeft,fromColLeft+7) * m.data->mat(rowRight,fromColLeft+7);
			result.data->mat(rowLeft,rowRight) += 
				( (v0+v1)+(v2+v3) ) + ( (v4+v5)+(v6+v7) );
		}	
		}
	}*/
	if(rowDiffLeft==1 && rowDiffRight<=8 && colDiff<=8){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
		for(unsigned int col=fromColLeft; col<toColLeft; col++){
			result.data->mat(fromRowLeft,rowRight)
				 += m.data->mat(fromRowLeft,col) * m.data->mat(rowRight,col);
		}	
		}
	}
	else if(rowDiffLeft==2 && rowDiffRight<=8 && colDiff<=8){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
		for(unsigned int col=fromColLeft; col<toColLeft; col++){
			result.data->mat(fromRowLeft,rowRight)
				 += m.data->mat(fromRowLeft,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+1,rowRight)
				 += m.data->mat(fromRowLeft+1,col) * m.data->mat(rowRight,col);
		}	
		}
	}
	else if(rowDiffLeft==4 && rowDiffRight<=8 && colDiff<=8){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
		for(unsigned int col=fromColLeft; col<toColLeft; col++){
			result.data->mat(fromRowLeft,rowRight)
				 += m.data->mat(fromRowLeft,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+1,rowRight)
				 += m.data->mat(fromRowLeft+1,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+2,rowRight)
				 += m.data->mat(fromRowLeft+2,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+3,rowRight)
				 += m.data->mat(fromRowLeft+3,col) * m.data->mat(rowRight,col);
		}	
		}
	}
	else if(rowDiffLeft==8 && rowDiffRight<=8 && colDiff<=8){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
		for(unsigned int col=fromColLeft; col<toColLeft; col++){
			result.data->mat(fromRowLeft,rowRight)
				 += m.data->mat(fromRowLeft,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+1,rowRight)
				 += m.data->mat(fromRowLeft+1,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+2,rowRight)
				 += m.data->mat(fromRowLeft+2,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+3,rowRight)
				 += m.data->mat(fromRowLeft+3,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+4,rowRight)
				 += m.data->mat(fromRowLeft+4,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+5,rowRight)
				 += m.data->mat(fromRowLeft+5,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+6,rowRight)
				 += m.data->mat(fromRowLeft+6,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+7,rowRight)
				 += m.data->mat(fromRowLeft+7,col) * m.data->mat(rowRight,col);
		}	
		}
	}
	else if(rowDiffLeft<=8 && rowDiffRight<=8 && colDiff<=8){
		for(unsigned int rowLeft=fromRowLeft; rowLeft<toRowLeft; rowLeft++){
		for(unsigned int rowRight=fromRowRight; rowRight<toRowRight; rowRight++){
		for(unsigned int col=fromColLeft; col<toColLeft; col++){
			result.data->mat(rowLeft,rowRight)
				 += m.data->mat(rowLeft,col) * m.data->mat(rowRight,col);
		}	
		}	
		}
	}
	else if( rowDiffLeft < colDiff && rowDiffRight < colDiff){
		// split columns
		//unsigned int colDiv = fromColLeft + (toColLeft-fromColLeft)/2;
		unsigned int colDiv = fromColLeft + (1 << (detail::ilog2(toColLeft-fromColLeft)-1));

		_nemo_impl_matMatTBlockRecursive(
			m,result,
			fromRowLeft,toRowLeft,fromRowRight,toRowRight,
			fromColLeft, colDiv, computeLowerTriangle
		);
		_nemo_impl_matMatTBlockRecursive(
			m,result,
			fromRowLeft,toRowLeft,fromRowRight,toRowRight,
			colDiv, toColLeft, computeLowerTriangle
		);
	}
	else if( rowDiffLeft < rowDiffRight){
		// split right rows
		//unsigned int rowDivRight = fromRowRight + (toRowRight-fromRowRight)/2;
		unsigned int rowDivRight = fromRowRight + (1 << (detail::ilog2(toRowRight-fromRowRight)-1));
		
		_nemo_impl_matMatTBlockRecursive(
			m,result,
			fromRowLeft,toRowLeft,fromRowRight,rowDivRight,
			fromColLeft, toColLeft, computeLowerTriangle
		);
		_nemo_impl_matMatTBlockRecursive(
			m,result,
			fromRowLeft,toRowLeft,rowDivRight,toRowRight,
			fromColLeft, toColLeft, computeLowerTriangle
		);
	}
	else{
		// split left rows
		//unsigned int rowDivLeft = fromRowLeft + (toRowLeft-fromRowLeft)/2;
		unsigned int rowDivLeft = fromRowLeft + (1 << (detail::ilog2(toRowLeft-fromRowLeft)-1));
	
		_nemo_impl_matMatTBlockRecursive(
			m,result,
			fromRowLeft,rowDivLeft,fromRowRight,toRowRight,
			fromColLeft, toColLeft, computeLowerTriangle
		);
		_nemo_impl_matMatTBlockRecursive(
			m,result,
			rowDivLeft, toRowLeft,fromRowRight,toRowRight,
			fromColLeft, toColLeft, computeLowerTriangle
		);
	}
}
#endif

#ifndef NEMO_PARSED_BY_DOXYGEN
template <typename T>
inline void _nemo_impl_matMatTBlockIterative(const Matrix<T> &m, Matrix<T> &result){
	const unsigned int blockSize = 4;
	const unsigned int numRows = m.dimension().rows();
	const unsigned int numCols = m.dimension().cols();

	for(unsigned int blockRowLeft = 0; blockRowLeft < numRows/blockSize; blockRowLeft++){
	for(unsigned int blockRowRight = blockRowLeft; blockRowRight < numRows/blockSize; blockRowRight++){
	// TODO first run with blockRowRight==blockRowLeft only evaluates upper triangle
	for(unsigned int blockCol = 0; blockCol < numCols/blockSize; blockCol++){
		const unsigned int fromRowLeft = blockRowLeft*blockSize;
		for(unsigned int rowRight=blockRowRight*blockSize; rowRight<(blockRowRight+1)*blockSize; rowRight++){
		for(unsigned int col=blockCol*blockSize; col<(blockCol+1)*blockSize; col++){

			assert(blockSize==4);
			result.data->mat(fromRowLeft,rowRight)
				 += m.data->mat(fromRowLeft,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+1,rowRight)
				 += m.data->mat(fromRowLeft+1,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+2,rowRight)
				 += m.data->mat(fromRowLeft+2,col) * m.data->mat(rowRight,col);
			result.data->mat(fromRowLeft+3,rowRight)
				 += m.data->mat(fromRowLeft+3,col) * m.data->mat(rowRight,col);
		}	
		}
	}
	}
	}
	// FIXME compute remaining entries, not fitting into blocks exactly

	// copy upper triangle to lower triangle
	for(unsigned int row = 0; row < numRows; row++){
		for(unsigned int col = row; col < numRows; col++){
			result.data->mat(col,row) = result.data->mat(row,col);
		}
	}	
}
#endif

/** 
 * Optimized method to compute m*m.transpose()
 */
template <typename T>
Matrix<T> matMatT(const Matrix<T> &m){
	const unsigned int numRows = m.dimension().rows();
	const unsigned int numCols = m.dimension().cols();

	
	Matrix<T> result( dim(numRows,numRows), (T)0 );
	if(numRows==0 || numCols==0){
		return result;
	}

	// compute upper triangle and mirror it downwards
	/*for(unsigned int leftRowIndex = 0; leftRowIndex<numRows; leftRowIndex++){
		for(unsigned int rightColIndex = leftRowIndex; rightColIndex<numRows; rightColIndex++){
			double value = 0.0;
			for(unsigned int i=0; i<numCols; i++){
				value += m.data->mat(leftRowIndex,i) * m.data->mat(rightColIndex,i);
			}
			result.data->mat(leftRowIndex,rightColIndex) = value;
			result.data->mat(rightColIndex,leftRowIndex) = value;
		}
	}*/
	if((numRows*numCols) <= 5000){
		/* Compared to the above implementation, this cache-oblivious one reduces runtime
		 * by -37% for   100x1000 matrices  (compared to eigen2 still +36%)
		 * by -53% for 1000x10000 matrices  (compared to eigen2 still +139%!!!)
		 * For small to medium matrices (e.g. 10x100) its 3x faster than eigen2
		 */
		_nemo_impl_matMatTRecursive(m,result,numCols, 0, numRows, 0, numRows, false);
	}
	else{
		result = m*m.transpose();
	}

	//_nemo_impl_matMatTBlockRecursive(m,result, 0, numRows, 0, numRows, 0, numCols, false);

	//_nemo_impl_matMatTBlockIterative(m, result);

	return result;

	/* This one, optimized 'within' Eigen2 is freaking slow!
	 * For small matrices it's 50% slower than _nemo_impl_matMatTRecursive
	 * For large matrices it's slow as hell, there does not seem to be any cache optimization */
	/*Eigen::MatrixXd resultEigen((int)numRows, (int)numRows);
	resultEigen.part<Eigen::SelfAdjoint>() = (m.data->mat * m.data->mat.adjoint()).lazy();
	return Matrix<T>(resultEigen);*/
}

/** 
 * Optimized method to compute m.transpose()*m
 */
template <typename T>
Matrix<T> matTMat(const Matrix<T> &m){
	return matMatT(m.transpose());
}


} // end namespace

#endif

