#ifndef _NEMO_RANDOM_H_
#define _NEMO_RANDOM_H_

#include <stdlib.h>
#include <cmath>


namespace nemo {

/**
 * Generic interface for generators of any kind.
 */
template <typename T>
class Generator{
protected:
	/**
	 * Override this method to create a custom generator.
	 */
	virtual T generate(){return T();}

public:
	/**
	 * Returns an object of type T. 
	 * Which one it is depends on the generator you are using
	 */
	T operator()(){
		return generate();
	}

	virtual ~Generator(){}
};

/**
 * A random number generator for uniform distributions of scalar data-types.
 * Will return a random number between min and max.
 */
template <typename T>
class uniform : public Generator<T> {
public:
	uniform(T minP, T maxP)
		 : Generator<T>(), 
		   min( (minP<=maxP)?minP:maxP ), 
		   max( (minP<=maxP)?maxP:minP ),
		   scale( isIntegral ? 
				( (max+(T)1-min) / ((double)RAND_MAX+(double)1) ) : 
				( (max-min) / (double)RAND_MAX )
		   ) 
	{}

protected:
	T generate(){
		// FIXED within scale: rand/RAND_MAX must not evaluate to 1.0 for integral values
		// => div. by (RAND_MAX+1) and evaluate it within double precision to avoid overflow in case of RAND_MAX==INT_max
		return min + (T) ( (double)rand() * scale );
	}
private:
	const T min, max;
	const double scale;
	const static bool isIntegral;
};
#ifndef NEMO_PARSED_BY_DOXYGEN
template <typename T>
const bool uniform<T>::isIntegral = ( ((T)(T()+0.5))==T() );
#endif


/**
 * A random number generator for gaussian distributions of scalar floating data-types.
 * Will return a random numbers with standard deviation and mean specified in the constructor.
 */
template <typename T>
class gaussian : public Generator<T> {
public:
	gaussian(T stdDevP=(T)1, T meanP=(T)0)
	 : Generator<T>(), 
	   stdDev( stdDevP ), 
	   mean( meanP ) ,
	   uniRand(-1.0, 1.0),
	   hasNext(false),
	   next()
	{}

	gaussian(const gaussian& g)
	 : Generator<T>(), 
	   stdDev( g.stdDev ), 
	   mean( g.mean ) ,
	   uniRand(g.uniRand),
	   hasNext(false), // no copy here!
	   next()	   // otherwise they won't be independent
	{}
	

protected:
	T generate(){
		if(hasNext){
			hasNext = false;
			return next;
		}
		else{
			T x1, x2, r;
			do {
				x1 = uniRand();
				x2 = uniRand();
				r = x1*x1 + x2*x2;
			}
			while(r > (T)1 || r == (T)0);

			/* Box-Muller transform */
			hasNext = true;
			// these two number are independent and both gaussian!
			T scale = stdDev * sqrt(-2.0 * log (r) / r);
			next = scale * x1 + mean;
			return scale * x2 + mean;
		}
	}
private:
	T stdDev, mean;

	uniform<T> uniRand;
	bool hasNext;
	T next;
};

} // end namespace


#endif

