#ifndef _NEMO_TIME_SERIES_H_
#define _NEMO_TIME_SERIES_H_

#include "IntrusiveCowSupport.h"
#include "IntrusiveCounted.h"
#include "Collection.h"
#include "detail/nemo-cpp11.h"

#include <stdexcept>
#include <vector>
#include <algorithm>


#include <iostream>



namespace nemo {

/* Note: absolute-time + duration for values possible by using
 * timestamp-objects that combine both???
 */



template <typename T>
class TimeOffset;
template <typename T>
static TimeOffset<T> offset(T o);

/**
 * Utility class for TimeSeries manipulation. Allows typesafe specification 
 * of offsets/durations.
 */
template <typename T>
class TimeOffset {
public:
	inline const T& get() const {return o;}
	friend TimeOffset<T> offset<T>(T o);
	TimeOffset(const TimeOffset &other) : o(other.o) {}
private:
	TimeOffset(T oP) : o(oP) {}
	TimeOffset& operator=(const TimeOffset& other){o = other.o; return *this;}
	T o;
};

/**
 * Neat factory method for TimeOffset objects to be put into TimeSeries objects.
 */
template <typename T>
static inline TimeOffset<T> offset(T o){
	return TimeOffset<T>(o);
}

/**
 * This class represents one entry in a TimeSeries.
 */
template <typename TimestampType, typename ValueType>
class TimeSeriesEntry {
public:
	TimestampType timestamp;
	ValueType value;

	TimeSeriesEntry(TimestampType timestampP, ValueType valueP)
	 : timestamp(timestampP), value(valueP) {}
	bool operator==(const TimeSeriesEntry<TimestampType,ValueType> &other) const {
		return (timestamp==other.timestamp) && (value==other.value);
	}
	bool operator!=(const TimeSeriesEntry<TimestampType,ValueType> &other) const {
		return !(*this==other);
	}
};



/**
 * Represents a temporal series of values of type \c ValueType. Each value is associated
 * to a timestamp (of type \c TimestampType), which can be used to address the value.
 * 
 * Both values and timestamps can also be accessed in an index-based manner using the
 * public members \c values and \c timestamps. The ordering of that indexing corresponds 
 * to an ordered list of the timestamps.
 * 
 * The third template argument \cbool \c strictOrdering allows to require a strict 'smaller than'
 * relation in the ordering of timestamps. If it is set to \c true, the TimeSeries will throw
 * \c std::invalid_argument whenever two timestamps in the series would become equal, such that
 * a strict ordering is impossible. By default \c strictOrdering is set to \c false.
 * 
 * Write operations are generally permitted to change the ordering of timestamps.
 * They will automatically be re-sorted, and the values are re-ordered accordingly such
 * that the original timestamp->value association is maintained.
 */
template <typename TimestampType, typename ValueType, bool strictOrdering=false>
class TimeSeries : public Collection<TimeSeriesEntry<TimestampType,ValueType> > {
public:
	typedef _nemo_get_type_(TimestampType()-TimestampType()) DurationType;
	typedef TimeSeries<TimestampType,ValueType,strictOrdering> ThisType;
	typedef TimeSeriesEntry<TimestampType,ValueType> EntryType;


private:
#ifndef NEMO_PARSED_BY_DOXYGEN
	class TimeSeriesData : public IntrusiveCounted<TimeSeriesData> {
	public:
		mutable std::vector<ValueType> values;
		mutable std::vector<TimestampType> timestamps;
		//DurationType lastEntryDuration;
	};
	IntrusiveCowSupport<TimeSeriesData> data;


		
	template <typename OtherTimestampType,typename OtherValueType>
	class EntryComparator {
	public:
		typedef TimeSeriesEntry<OtherTimestampType,OtherValueType> OtherEntryType;
		bool operator()(OtherEntryType first, OtherEntryType second){ 
			return (first.timestamp < second.timestamp);
		}
	};
#endif


// =============================================================================
//    Content View Abstractions
// =============================================================================
public:
	class TimestampView;
	class ValueView;

#ifndef NEMO_PARSED_BY_DOXYGEN
	template <typename T>
	class ViewIterator{
	private:
		ViewIterator(int indexP, std::vector<T> &viewDataP) : index(indexP), viewData(viewDataP) {}
	public:
		friend class TimestampView;
		friend class ValueView;
		
		ViewIterator(const ViewIterator& it) : index(it.index), viewData(it.viewData) {}
		ViewIterator& operator=(const ViewIterator &it){
			if(this==&it){return *this;}
			this->index = it.index;
			this->viewData = it.viewData;
			return *this;
		}
	
		typedef int difference_type;
		typedef T value_type;
		typedef T* pointer;
		typedef T reference;
		typedef std::random_access_iterator_tag iterator_category;

		T operator*() const{
			if(index<0 || (int)viewData.size()<=index){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return viewData[index];
		}
		T operator[](difference_type offset) const{
			if(index+offset<0 || (int)viewData.size()<=index+offset){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return viewData[index+offset];
		}
		ViewIterator& operator++(){
			index++;
			return const_cast<ViewIterator&>(*this);
		}
		ViewIterator operator++(int){
			ViewIterator copy = *this;
			index++;
			return copy;
		}
		ViewIterator& operator--(){
			index--;
			return const_cast<ViewIterator&>(*this);
		}
		ViewIterator operator--(int){
			ViewIterator copy = *this;
			index--;
			return copy;
		}
		ViewIterator operator+(difference_type off) const {
			ViewIterator it = *this;
			it.index += off;
			return it;
		}
		ViewIterator operator-(difference_type off) const {
			ViewIterator it = *this;
			it.index -= off;
			return it;
		}
		difference_type operator-(const ViewIterator& rhs) const {
			return index - rhs.index;
		}
		bool operator==(const ViewIterator &it) const{
			return (&viewData==&it.viewData) && (index==it.index);
		}
		bool operator!=(const ViewIterator &it) const{
			return !(*this==it);
		}
		bool operator<=(const ViewIterator &it) const{
			return index<=it.index;
		}
		bool operator<(const ViewIterator &it) const{
			return index<it.index;
		}
		bool operator>=(const ViewIterator &it) const{
			return index>=it.index;
		}
		bool operator>(const ViewIterator &it) const{
			return index>it.index;
		}
	private:
		int index;
		std::vector<T> &viewData;
	};
#endif

	/**
	 * This class represents a view on the timestamps within one timeseries.
	 * It allows index-based access and manipulation of the timestamps, as well
	 * as \c Collection::map and \C Collection::transform operations.
	 * This class cannot be created or assigned manually, but can only be used 
	 * via the \c TimeSeries::timestamps instance.
	 */
	class TimestampView : public Collection<TimestampType> {
	private:
		TimestampView(TimeSeries<TimestampType,ValueType,strictOrdering> &seriesP)
		 : series(seriesP) {}
		TimestampView(const TimestampView &other)
		 : series(other.series) {_NEMO_LOGIC_ERROR("");}
		TimestampView& operator=(const TimestampView& other){
			series = other.series;
			return *this;
		}

		void rearrangeSeries(std::vector<EntryType> &newTimestamps){
			// get new timestamps in a sorted order
			std::sort(newTimestamps.begin(), newTimestamps.end(), EntryComparator<TimestampType,ValueType>());

			// insert timestamps and pick up correct values
			series.data.detach();
			for(unsigned int i = 0; i < newTimestamps.size(); i++){
				series.data->timestamps[i] = newTimestamps[i].timestamp;
				series.data->values[i] = newTimestamps[i].value;
			}
		}
	public:

		friend class TimeSeries<TimestampType,ValueType,strictOrdering>;
		
		// ======== General access  ========

		/**
		 * Index-based read access.
		 */
		TimestampType operator[](unsigned int index) const {
			series.checkIndex(index);
			return series.data->timestamps[index]; // return const-ref???
		}
	
		/**
		 * Set the timestamp at specified index to a new timestamp-value.
		 * This operation might lead to a re-ordering, such that the entry
		 * will be available at a different index.
		 * 
		 * Runtime: O(n)  if re-ordering is necessary
		 * Runtime: O(1)  if re-ordering is NOT necessary
		 */
		void set(unsigned int index, TimestampType value){
			series.checkIndex(index);
			series.data.detach();
			// TODO simplify this code
			if(0<index && !series.compareTime(series.data->timestamps[index-1], value)){
				// timestamp smaller than left element
				// right-rotation of content required
				unsigned int leftLimit = 0;
				for(int i = index-1; 0 <= i; i--){
					if(series.compareTime(series.data->timestamps[i], value)){
						leftLimit = i+1;
						break;
					}
				}
				if(strictOrdering && !series.compareTime(value, series.data->timestamps[leftLimit])){
					_NEMO_ARGUMENT_ERROR("Timeseries: timestamp already exists");
				}
				series.data->timestamps[index] = value;
				std::rotate( 
					series.data->timestamps.begin()+leftLimit, 
					series.data->timestamps.begin()+index,
					series.data->timestamps.begin()+index+1 
				);
				std::rotate( 
					series.data->values.begin()+leftLimit, 
					series.data->values.begin()+index,
					series.data->values.begin()+index+1 
				);
			}
			else if(index < series.length()-1  && !series.compareTime(value, series.data->timestamps[index+1])){
				// timestamp bigger than right element
				// left-rotation of content required
				unsigned int rightLimit = series.length();
				for(int i = index+1; i < (int)series.length(); i++){
					if(series.compareTime(value, series.data->timestamps[i])){
						rightLimit = i;
						break;
					}
				}
				if(strictOrdering && !series.compareTime(series.data->timestamps[rightLimit-1], value)){
					_NEMO_ARGUMENT_ERROR("Timeseries: timestamp already exists");
				}
				series.data->timestamps[index] = value;
				std::rotate( 
					series.data->timestamps.begin()+index, 
					series.data->timestamps.begin()+index+1,
					series.data->timestamps.begin()+rightLimit 
				);
				std::rotate( 
					series.data->values.begin()+index,  
					series.data->values.begin()+index+1,
					series.data->values.begin()+rightLimit 
				);
			}
			else{
				series.data->timestamps[index] = value;
			}
		}

		/**
		 * Removes the timeseries entry at position \c index
		 * 
		 * Runtime: O(n) 
		 */
		void remove(unsigned int index){
			series.checkIndex(index);
			series.data.detach();
			series.data->values.erase( series.data->values.begin() + index );
			series.data->timestamps.erase( series.data->timestamps.begin() + index );
		}

		// ======== Collection API  ========
		unsigned int numElements() const{
			return series.length();
		}
		TimestampType getElement(unsigned int elementIndex) const{
			return operator[](elementIndex);
		}
		void setElement(unsigned int elementIndex, TimestampType value){
			set(elementIndex, value);
		}
		void transform(Mapping<TimestampType,TimestampType> m){
			std::vector<EntryType> newTimestamps;
			for(unsigned int i=0; i<series.length(); i++){
				newTimestamps.push_back( 
					EntryType( m(series.data->timestamps[i]), series.data->values[i]) 
				);
			}
			rearrangeSeries(newTimestamps);
		}
		template <typename OutType>
		TimeSeries<OutType,ValueType,strictOrdering> map(Mapping<TimestampType,OutType> m) const {
			// compute new timestamps
			std::vector<TimeSeriesEntry<OutType,ValueType> > newTimestamps;
			for(unsigned int i=0; i<series.length(); i++){
				newTimestamps.push_back( 
					TimeSeriesEntry<OutType,ValueType>( m(series.data->timestamps[i]), series.data->values[i]) 
				);
			}

			// sort and append them
			TimeSeries<OutType,ValueType,strictOrdering> mapped;
			std::sort(newTimestamps.begin(), newTimestamps.end(), EntryComparator<OutType,ValueType>());
			for(unsigned int i = 0; i < newTimestamps.size(); i++){
				mapped.append(newTimestamps[i].timestamp, newTimestamps[i].value);
			}
			return mapped;
		}

		// ======== STL compliant iterator interface  ========
		
		/**
		 * Returns an STL-compliant (read-only) iterator pointing to the first timestamp.
		 */
		ViewIterator<TimestampType> begin() const {
			return ViewIterator<TimestampType>(0, series.data->timestamps);
		}
		/**
		 * Returns an STL-compliant (read-only) iterator pointing one position after the last timestamp.
		 */
		ViewIterator<TimestampType> end() const {
			return ViewIterator<TimestampType>(series.length(), series.data->timestamps);
		}
	private:
		TimeSeries<TimestampType,ValueType,strictOrdering> &series;
	};

	/**
	 * This class represents a view on the values within one timeseries.
	 * It allows index-based access and manipulation of the timestamps, as well
	 * as \c Collection::map and \C Collection::transform operations.
	 * This class cannot be created or assigned manually, but can only be used 
	 * via the \c TimeSeries::values instance.
	 */
	class ValueView : public Collection<ValueType> {
	private:
		ValueView(TimeSeries<TimestampType,ValueType,strictOrdering> &seriesP)
		 : series(seriesP) {}
		ValueView(const ValueView &other)
		 : series(other.series) {_NEMO_LOGIC_ERROR("");}
		ValueView& operator=(const ValueView& other){
			series = other.series;
			return *this;
		}
	public:

		friend class TimeSeries<TimestampType,ValueType,strictOrdering>;
		
		// ======== General access  ========

		/**
		 * Index-based read access.
		 */
		ValueType operator[](unsigned int index) const {
			series.checkIndex(index);
			return series.data->values[index]; // return const-ref???
		}
	
		/**
		 * Sets the value at specified index to a new value
		 */
		void set(unsigned int index, ValueType value){
			series.checkIndex(index);
			series.data.detach();
			series.data->values[index] = value;
		}

		void remove(unsigned int index){
			series.checkIndex(index);
			series.data.detach();
			series.data->values.erase( series.data->values.begin() + index );
			series.data->timestamps.erase( series.data->timestamps.begin() + index );
		}

		// ======== Collection API  ========
		unsigned int numElements() const{
			return series.length();
		}
		ValueType getElement(unsigned int elementIndex) const{
			return operator[](elementIndex);
		}
		void setElement(unsigned int elementIndex, ValueType value){
			set(elementIndex, value);
		}
		void transform(Mapping<ValueType,ValueType> m){
			series.data.detach();
			for(unsigned int i=0; i<series.length(); i++){
				series.data->values[i] = m(series.data->values[i]);
			}
		}
		template <typename OutType>
		TimeSeries<TimestampType,OutType,strictOrdering> map(Mapping<ValueType,OutType> m) const {
			TimeSeries<TimestampType,OutType,strictOrdering> mapped = series;
			mapped.data.detach();
			for(unsigned int i=0; i<series.length(); i++){
				mapped.data->values[i] = m(series.data->values[i]);
			}
			return mapped;
		}

		// ======== STL compliant iterator interface  ========

		/**
		 * Returns an STL-compliant (read-only) iterator pointing to the first value.
		 */
		ViewIterator<ValueType> begin() const {
			return ViewIterator<ValueType>(0, series.data->values);
		}
		/**
		 * Returns an STL-compliant (read-only) iterator pointing one position after the last element.
		 */
		ViewIterator<ValueType> end() const {
			return ViewIterator<ValueType>(series.length(), series.data->values);
		}
	
	private:
		TimeSeries<TimestampType,ValueType,strictOrdering> &series;
	};

	
	/**
	 * View on the timestamps within the series.
	 */
	TimestampView timestamps;
	/**
	 * View on the values within the series.
	 */
	ValueView values;

#ifndef NEMO_PARSED_BY_DOXYGEN
	friend class TimestampView;
	friend class ValueView;
#endif

// =============================================================================
//    Construction
// =============================================================================
public:
	// FIXME use of *this maybe problematic???
	TimeSeries() : data(new TimeSeriesData()), timestamps(*this), values(*this) {}
	TimeSeries(const ThisType &o) : data(o.data), timestamps(*this), values(*this) {}
	TimeSeries<TimestampType,ValueType,strictOrdering>& operator=(const ThisType &o){
		if(this!=&o){
			data = o.data;
			timestamps = TimestampView(*this);
			values = ValueView(*this);
		}
		return *this;
	}

// =============================================================================
//    General Access
// =============================================================================

	/** Returns the number of entries in the TimeSeries.
	 */
	unsigned int length() const {
		return data->values.size();
	}
	/** Returns true iff length()==0.
	 */
	bool isEmpty() const {
		return data->values.empty();
	}
	/**
	 * Returns the value associated to the specified timestamp.
	 * If the timestamp is not (exactly) contained in the TimeSeries,
	 * an invalid_argument exception will be thrown.
	 * 
	 * Runtime: O(log(n)) 
	 */
	ValueType operator[](TimestampType timeP) const {
		unsigned int index = findTimestampIndex(timeP);
		return data->values[index];
	}
	
// =============================================================================
//    General Manipulation
// =============================================================================

	/**
	 * Appends a new entry to the TimeSeries. An invalid_argument exception will be thrown
	 * if the timestamp-ordering does not allow to place this timestamp at the end of the series.
	 * 
	 * Runtime: O(1) 
	 */
	void append(TimestampType timeP, ValueType valueP){
		checkTimestampAppend(timeP);
		data.detach();
		data->values.push_back(valueP);
		data->timestamps.push_back(timeP);
	}
	/**
	 * Appends a new entry to the TimeSeries. The timestamp will be
	 * - \c timestamps(length()-1)+offset   if the series already contains entries
	 * - \c TimestampType()+offset          if the series is empty
	 * 
	 * Usage: @code
series.append(offset(offsetValue), value) @endcode
	 * 
	 * Runtime: O(1) 
	 */
	void append(const TimeOffset<DurationType> &offset, ValueType valueP){
		TimestampType time = isEmpty() ? 
			TimestampType() + offset.get() : 
			data->timestamps.back() + offset.get();
		checkTimestampAppend(time);
		data.detach();
		data->values.push_back(valueP);
		data->timestamps.push_back(time);
	}
	/**
	 * Appends the entries of another series to this one.
	 * An invalid_argument exception will be thrown if the timestamps
	 * are not in a valid ordering.
	 */
	void append(ThisType otherSeries){
		if(!otherSeries.isEmpty()){
			checkTimestampAppend(otherSeries.timestamp(0));
			data.detach();
			for(unsigned int i=0; i<otherSeries.length(); i++){
				data->values.push_back(otherSeries.data->values[i]);
				data->timestamps.push_back(otherSeries.data->timestamps[i]);
			}
		}
	}
	/**
	 * Appends the entries of another series to this one.
	 * Before adding them, this method shifts the timestamps of the other series
	 * such that the difference between the last element of this series and the first element
	 * of the other series will be exactly \c offset.
	 */
	void append(const TimeOffset<DurationType> &offset, ThisType otherSeries){
		if(!otherSeries.isEmpty()){
			DurationType shift = isEmpty() ? 
					offset.get() : 
					(data->timestamps.back() - otherSeries.data->timestamps.front()) + offset.get();
			TimestampType firstTime = data->timestamps.back()+shift;
			checkTimestampAppend(firstTime);
			data.detach();
			for(unsigned int i=0; i<otherSeries.length(); i++){
				data->values.push_back(otherSeries.data->values[i]);
				data->timestamps.push_back(otherSeries.data->timestamps[i]+shift);
			}
		}
	}
	/**
	 * Sets a new value for an existing timestamp.
	 * If the timestamp is not (exactly) contained in the TimeSeries,
	 * an invalid_argument exception will be thrown.
	 * 
	 * Runtime: O(log(n)) 
	 */
	void setValue(TimestampType timeP, ValueType valueP){
		unsigned int index = findTimestampIndex(timeP);
		data.detach();
		data->values[index] = valueP;
	}
	/**
	 * Inserts a new entry into the timeseries. 
	 * If \c strictOrdering==false this operation will never throw.
	 * If \c strictOrdering is enabled it will throw invalid_argument if the timestamp already exists.
	 * 
	 * Runtime:  o(log(n));  O(n)
	 */
	void insert(TimestampType timeP, ValueType valueP){
		typedef _nemo_get_type_(data->timestamps.begin()) Iterator;
		// use STL binary search
		// first element with timestamp greater (non-equal) than timeP
		Iterator pos = upper_bound(data->timestamps.begin(), data->timestamps.end(), timeP);
		
		// check ordering constraint
		if(strictOrdering && pos!=data->timestamps.begin() && *(pos-1)==timeP){
			_NEMO_ARGUMENT_ERROR("Existing timestamp can not be inserted with strict-ordering");
		}

		unsigned int index = pos - data->timestamps.begin();
		data.detach();
		data->timestamps.insert(pos, timeP);
		data->values.insert(data->values.begin()+index, valueP);
	}
	/**
	 * Removes the entry with the specified timestamp.
	 * 
	 * Runtime:  o(log(n));  O(n)
	 */
	void remove(TimestampType timeP){
		unsigned int index = findTimestampIndex(timeP);
		data.detach();
		data->values.erase( data->values.begin() + index );
		data->timestamps.erase( data->timestamps.begin() + index );
	}
	
	// TODO do we need a 'removeAll' method for strictOrdering==false ?

	/**
	 * Removes the entry with the smallest timestamp.
	 * 
	 * Runtime:  O(n)
	 */
	void removeFirst(){
		checkIndex(0);
		data.detach();
		data->values.erase( data->values.begin() );
		data->timestamps.erase( data->timestamps.begin() );
	}
	/**
	 * Removes the \c k entries with the smallest timestamps.
	 * 
	 * Runtime:  O(n)
	 */
	void removeFirst(unsigned int k){
		if(k!=0){ checkIndex(k-1); }
		data.detach();
		data->values.erase( data->values.begin(), data->values.begin()+k );
		data->timestamps.erase( data->timestamps.begin(), data->timestamps.begin()+k );
	}

// =============================================================================
//    Comparison
// =============================================================================
	/**
	 * Returns true if dimensions and all contained values are equal (i.e. exactly equal 
	 * in terms of T::operator==(T&) ).
	 */
	bool operator==(const ThisType &r) const{
		if(length() != r.length()){
			return false;
		}
		for(unsigned int i=0; i<length(); i++){
			if(data->values[i] != r.data->values[i]){
				return false;
			}
			if(data->timestamps[i] != r.data->timestamps[i]){
				return false;
			}
		}
		// else
		return true;
	}

	/**
	 * ! operator==
	 */
	bool operator!=(const ThisType &r) const{
		return !operator==(r);
	}

// =============================================================================
//    Collection API
// =============================================================================
	unsigned int numElements() const{
		return length();
	}

	/**
	 * @param elementIndex between zero and numElements()-1
	 * @return Respective element in collection
	 */
	EntryType getElement(unsigned int elementIndex) const{
		checkIndex(elementIndex);
		return EntryType(data->timestamps[elementIndex], data->values[elementIndex]);
	}

	/**
	 * Change the value of an element
	 * @param elementIndex between zero and numElements()-1
	 * @param value New value for respective element
	 */
	void setElement(unsigned int elementIndex, EntryType entry){\
		checkIndex(elementIndex);
		values.set(elementIndex, entry.value); // value first
		timestamps.set(elementIndex, entry.timestamp); // timestamp might cause re-ordering
	}

	void transform(Mapping<EntryType,EntryType> m){
		data.detach();
		std::vector<EntryType> newEntries;
		for(unsigned int i = 0; i < length(); i++){
			newEntries.push_back( m(EntryType(data->timestamps[i], data->values[i])) );
		}
		// FIXME more efficient ways to do this?
		data->timestamps.clear();
		data->values.clear();
		std::sort( newEntries.begin(), newEntries.end(), EntryComparator<TimestampType,ValueType>() );
		for(unsigned int i = 0; i < newEntries.size(); i++){
			append( newEntries[i].timestamp, newEntries[i].value );
		}
	}

	template <typename OtherTimestampType, typename OtherValueType>
	TimeSeries<OtherTimestampType,OtherValueType,strictOrdering> map(
			Mapping<EntryType,TimeSeriesEntry<OtherTimestampType,OtherValueType> > m)
	const {
		TimeSeries<OtherTimestampType,OtherValueType,strictOrdering> mapped;
		for(unsigned int i=0; i<length(); i++){
			TimeSeriesEntry<OtherTimestampType,OtherValueType> newEntry = m(EntryType(data->timestamps[i], data->values[i]));
			mapped.append( newEntry.timestamp, newEntry.value );
		}
		return mapped;
	}
// =============================================================================
//    STL compliant iterator interface
// =============================================================================
#ifndef NEMO_PARSED_BY_DOXYGEN
	class EntryIterator{
	private:
		EntryIterator(int indexP, ThisType &seriesP) : index(indexP), series(seriesP) {}
	public:
		friend class TimeSeries<TimestampType,ValueType,strictOrdering>;
		
		EntryIterator(const EntryIterator& it) : index(it.index), series(it.series) {}
		EntryIterator& operator=(const EntryIterator &it){
			if(this==&it){return *this;}
			this->index = it.index;
			this->series = it.series;
			return *this;
		}
	
		typedef int difference_type;
		typedef EntryType value_type;
		typedef EntryType* pointer;
		typedef EntryType reference;
		typedef std::random_access_iterator_tag iterator_category;

		EntryType operator*() const{
			if(index<0 || (int)series.length()<=index){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return series.getElement(index);
		}
		EntryType operator[](difference_type offset) const{
			if(index+offset<0 || (int)series.length()<=index+offset){_NEMO_LOGIC_ERROR("Iterator out of range");}
			return series.getElement(index+offset);
		}
		EntryIterator& operator++(){
			index++;
			return const_cast<EntryIterator&>(*this);
		}
		EntryIterator operator++(int){
			EntryIterator copy = *this;
			index++;
			return copy;
		}
		EntryIterator& operator--(){
			index--;
			return const_cast<EntryIterator&>(*this);
		}
		EntryIterator operator--(int){
			EntryIterator copy = *this;
			index--;
			return copy;
		}
		EntryIterator operator+(difference_type off) const {
			EntryIterator it = *this;
			it.index += off;
			return it;
		}
		EntryIterator operator-(difference_type off) const {
			EntryIterator it = *this;
			it.index -= off;
			return it;
		}
		difference_type operator-(const EntryIterator& rhs) const {
			return index - rhs.index;
		}
		bool operator==(const EntryIterator &it) const{
			return (&series==&it.series) && (index==it.index);
		}
		bool operator!=(const EntryIterator &it) const{
			return !(*this==it);
		}
		bool operator<=(const EntryIterator &it) const{
			return index<=it.index;
		}
		bool operator<(const EntryIterator &it) const{
			return index<it.index;
		}
		bool operator>=(const EntryIterator &it) const{
			return index>=it.index;
		}
		bool operator>(const EntryIterator &it) const{
			return index>it.index;
		}
	private:
		int index;
		ThisType &series;
	};
#endif

	/**
	 * Returns an STL-compliant (read-only) iterator pointing to the first value.
	 */
	EntryIterator begin() const {
		return EntryIterator(0, const_cast<ThisType&>(*this));
	}
	/**
	 * Returns an STL-compliant (read-only) iterator pointing one position after the last element.
	 */
	EntryIterator end() const {
		return EntryIterator(length(), const_cast<ThisType&>(*this));
	}

private:
	inline void checkIndex(unsigned int index) const{
		if(data->values.size() <= index){	
			_NEMO_ARGUMENT_ERROR("TimeSeries: index out of range");
		}
	}
	inline void checkTimestampAppend(const TimestampType &ts) const {
		if(!isEmpty()){
			if( !compareTime(data->timestamps.back(), ts) ){	
				_NEMO_ARGUMENT_ERROR("TimeSeries: timestamp not behind latest");
			}
		}
	}
	inline bool compareTime(const TimestampType &first, const TimestampType &second) const{
		if(strictOrdering){
			return first < second;
		}
		else{
			return first <= second;
		}
	}
	unsigned int findTimestampIndex(const TimestampType &ts) const {
		typedef _nemo_get_type_(data->timestamps.begin()) Iterator;
		// use STL binary search
		Iterator pos = lower_bound(data->timestamps.begin(), data->timestamps.end(), ts);
		if(pos==data->timestamps.end() || *pos!=ts){
			_NEMO_ARGUMENT_ERROR("Timestamp could not be found in TimeSeries");
		}
		return pos-data->timestamps.begin();
	}
			
};

template <class V, class T>
std::ostream& operator<<(std::ostream& out, const TimeSeriesEntry<V,T> &entry){
	out << entry.timestamp << "\t" << entry.value;
	return out;
}
template <class V, class T>
std::ostream& operator<<(std::ostream& out, const TimeSeries<V,T> &series){

	for(unsigned int i=0; i<series.length(); i++){
		out << series.timestamps[i] << "\t" << series.values[i] << "\n";
	}
	return out;
}


} // end namespace

#endif

