#ifndef _NEMO_VECTOR_MAPPINGS_H_
#define _NEMO_VECTOR_MAPPINGS_H_

#include "Mapping.h"
#include "Vector.h"

#include <stdexcept>
#include <sstream>
#include <algorithm>

#ifdef _WIN32
	#define _USE_MATH_DEFINES // MSVC 'feature': this define seems necessary for M_PI to be defined
#endif
#include <math.h>


namespace nemo {


/**
 * A Mapping that extracts certain dimensions from an input vector.
 * Which dimensions are extracted can be specified in two ways.
 * Firstly, one can specify a startIndex and an output dimension.
 * Then, the values for indices startIndex ... (startIndex+outDim-1)
 * are put into the output.
 * Secondly, one can specify a boolean mask. All values, whose index
 * has a 'true' in the boolean mask will be set into the output.
 * 
 * This mapping type is not invertible.
 */
template <typename T=double>
class SubspaceMapping : public MappingDefinition<MathVector<T>,MathVector<T> > {
public:
	static Mapping<MathVector<T>,MathVector<T> > create(unsigned int startIndexP, unsigned int outputDimP){
		return Mapping<MathVector<T>,MathVector<T> >(new SubspaceMapping(startIndexP, outputDimP));
	}
	static Mapping<MathVector<T>,MathVector<T> > create(unsigned int inDim, bool *dimSelectionP){
		return Mapping<MathVector<T>,MathVector<T> >(new SubspaceMapping(inDim, dimSelectionP));
	}

	MathVector<T> evaluate(MathVector<T> value) const{
		if(value.dimension() < minInputDim){
			std::stringstream mess;
			mess << "SubspaceMapping::evaluate(): ";
			mess << "Too low input dimension: ";
			mess << "expected is at least" << minInputDim;
			mess << ", but is " << value.dimension() << ".";
			_NEMO_ARGUMENT_ERROR(mess.str());
		}
		if(dimSelection == NULL){
			return value.subVector(start, dim(outDim));
		}
		else{
			MathVector<T> out(dim(outDim));
			unsigned int outDimCount = 0;
			for(unsigned int i=0; i<value.dimension(); i++){
				if(dimSelection[i]){
					out[outDimCount++] = value[i];
				}
			}
			return out;
		}
	}

	~SubspaceMapping(){
		if(this->dimSelection){ delete[](this->dimSelection); }
	}

private:
	SubspaceMapping(unsigned int startIndex, unsigned int outputDim)
	 : start(startIndex), outDim(outputDim), dimSelection(NULL), minInputDim(startIndex+outputDim)
	{}
	SubspaceMapping(unsigned int inDim, bool *dimSelectionP)
	 : start(0), outDim(0), dimSelection(new bool[inDim]), minInputDim(0)
	{
		initFromBoolArray(inDim, dimSelectionP);
	}
	// forbid copies
	SubspaceMapping(SubspaceMapping&m):start(m.start),outDim(m.outDim),dimSelection(NULL),minInputDim(m.minInputDim){}
	SubspaceMapping operator=(const SubspaceMapping&){_NEMO_LOGIC_ERROR("Copying not allowed");}

	void initFromBoolArray(unsigned int inDim, bool *dimSelectionP){
		for(unsigned int i=0; i < inDim; i++){
			dimSelection[i] = dimSelectionP[i];
			if(dimSelection[i]){
				outDim++;
				if(minInputDim <= i){
					minInputDim = i+1;
				}
			}
		}
	}
	
	unsigned int start;
	unsigned int outDim;
	bool *dimSelection;
	unsigned int minInputDim;
};


/**
 * A Mapping computes polynomial terms out of vector-values.
 * Each PolynomExpansion is fully specified by its polynomial order, which it 
 * receives at construction. For instance, specifying a polynomial order 2
 * results in the following input-output behavior:
 * - p2( [x1] ) = [1, x1, x1^2]
 * - p2( [x1, x2] ) = [1, x1, x2, x1^2, x1*x2, x2^2]
 * 
 * This mapping type is not invertible.
 */
template <typename T=double>
class PolynomExpansion : public MappingDefinition<MathVector<T>,MathVector<T> > {
public:
	static Mapping<MathVector<T>,MathVector<T> > create(unsigned int polynomOrderP){
		return Mapping<MathVector<T>,MathVector<T> >(new PolynomExpansion(polynomOrderP));
	}

	MathVector<T> evaluate(MathVector<T> value) const{
		if(value.dimension() == 0){
			return MathVector<T>(dim(1), (T)1);
		}

		if(value.dimension() != inputDim){
			freeBinomials();
			inputDim = value.dimension();
			outputDim = outDimForInDim(inputDim, order);
			createBinomials();
		}

		MathVector<T> expanded(dim(outputDim));
		expanded[0] = 1;
		unsigned int expandedIndex = 1; // index to write next entry in expanded
		unsigned int lastOrderStart = 0;
		unsigned int lastOrderLength = 1;
		for(unsigned int p = 1; p <= order; p++){
			for(unsigned int inputIndex = 0; inputIndex < inputDim; inputIndex++){
				const unsigned int num = binomials[inputDim-inputIndex+p-2][p-1];
				const unsigned int offset = num + expandedIndex - lastOrderStart - lastOrderLength;
				for(unsigned int i = expandedIndex; i < expandedIndex+num; i++){
					expanded[i] = value[inputIndex] * expanded[i-offset];
				}
				expandedIndex += num;
			}
			lastOrderStart += lastOrderLength;
			lastOrderLength = binomials[inputDim+p-1][p];
		}
		return expanded;
	}
	
	~PolynomExpansion(){
		freeBinomials();
	}

private:
	PolynomExpansion(unsigned int polynomOrder)
	 : order(polynomOrder), inputDim(1), outputDim(outDimForInDim(inputDim, polynomOrder)), binomials(NULL) {
		createBinomials();
	}
		
	// forbid copies
	PolynomExpansion(PolynomExpansion&m):order(m.order),inputDim(m.inputDim),outputDim(m.outputDim),binomials(NULL){}
	PolynomExpansion operator=(const PolynomExpansion&){_NEMO_LOGIC_ERROR("Copying not allowed");}

	static unsigned int binomial(unsigned int N, unsigned int k){
		if(k==0){
			return 1;
		}
		if(N < 2*k){
			return binomial(N, N-k);
		}
		unsigned int b = N;
		for(unsigned int i = 2; i <= k; i++){
			b *= (N + 1 - i);
			b /= i;
		}
		return b;
	}

	static unsigned int outDimForInDim(unsigned int inDim, unsigned int order){
		unsigned int outDim = 1;
		if(0 < inDim){
			const unsigned int outDimMax = 1000000;
			for(unsigned int p=1; p<=order; p++){
				outDim += binomial(inDim+p-1, p);
				if(outDimMax < outDim){
					_NEMO_ARGUMENT_ERROR("PolynomExpansionMapping does not allow more than 1000000 output dimensions");
				}
			}
		}
		return outDim;
	}

	void createBinomials() const{
		binomials = new unsigned int*[inputDim+order];
		for(unsigned int i=0; i<inputDim+order; i++){
			binomials[i] = new unsigned int[order+1];
			const unsigned int maxArg = std::min(order, i)+1;
			for(unsigned int k=0; k<maxArg; k++){
				binomials[i][k] = binomial(i, k);// FIXME use addition of prev. values
			}
		}
	}

	void freeBinomials() const{
		for(unsigned int i=0; i<inputDim+order; i++){
			delete[](this->binomials[i]);
		}
		delete[](this->binomials);
	}

	unsigned int order;
	mutable unsigned int inputDim;
	mutable unsigned int outputDim;
	mutable unsigned int **binomials;
};

/**
 * This mapping type cuts off components of input vectors to
 * specified minimum and maximum values. For an input vector x,
 * the output c is computed as
 *  - c[i] = min[i] if x[i] < min[i]
 *  - c[i] = max[i] if x[i] > max[i]
 *  - c[i] = x[i]   else.
 * 
 * This mapping type is not invertible.
 */
template <typename T=double>
class Clamping : public MappingDefinition<MathVector<T>,MathVector<T> > {
public:
	static Mapping<MathVector<T>,MathVector<T> > create(MathVector<T> minP, MathVector<T> maxP){
		if(minP.dimension() != maxP.dimension()){
			_NEMO_ARGUMENT_ERROR("min and max values must have same dimension");
		}
		for(unsigned int i=0; i<minP.dimension(); i++){
			if(maxP[i] < minP[i]){
				_NEMO_ARGUMENT_ERROR("min must be smaller or equal to max");
			}
		}
		return Mapping<MathVector<T>,MathVector<T> >(new Clamping<T>(minP,maxP));
	}
	
	MathVector<T> evaluate(MathVector<T> input) const{
		if(input.dimension() != min.dimension()){
			_NEMO_ARGUMENT_ERROR("dimension mismatch");
		}
		MathVector<T> output = input;
		for(unsigned int i=0; i<input.dimension(); i++){
			if(output[i] < min[i]){
				output[i] = min[i];
			} else if (max[i] < output[i]){
				output[i] = max[i];
			}
		}
		return output;
	}

private:
	Clamping(MathVector<T> minP, MathVector<T> maxP)
	 : min(minP), max(maxP)
	{}
	// forbid copies
	Clamping(Clamping&m):min(m.min),max(m.max){_NEMO_LOGIC_ERROR("Copying not allowed");}
	Clamping operator=(const Clamping&){_NEMO_LOGIC_ERROR("Copying not allowed");}

	MathVector<T> min;
	MathVector<T> max;
};

// forward declaration
template <typename T> class Cartesian2Spherical;

/**
 * This mapping type transforms n-dimensional spherical coordinates to cartesian coordinates.
 * For an input vector s, it follows these conventions:
 * - s[0] represents the radius
 * - s[1] is the polar angle in the range [0:2*PI]
 * - s[i>=2] are azimut angles from the i-th equator to the i'th axis in the range [-PI/2:PI/2]
 * 
 * This mapping type inverts itself with Cartesian2Spherical mappings.
 */
template <typename T=double>
class Spherical2Cartesian : public MappingDefinition<MathVector<T>,MathVector<T> > {
public:
	static Mapping<MathVector<T>,MathVector<T> > create(){
		return Mapping<MathVector<T>,MathVector<T> >(new Spherical2Cartesian<T>());
	}
	
	MathVector<T> evaluate(MathVector<T> spherical) const{
		const unsigned int d = spherical.dimension();
		if(d<2){
			return spherical;
		}
		
		MathVector<T> cartesian(dim(d));

		T radius = spherical[0];

		T innerCosProduct = (T)1;
		for(unsigned int i = d-1; 2 <= i; i--){
			cartesian[i] = radius * innerCosProduct * sin(spherical[i]);
			innerCosProduct *= cos(spherical[i]);
		}
		cartesian[1] = radius * innerCosProduct * sin(spherical[1]);
		cartesian[0] = radius * innerCosProduct * cos(spherical[1]);

		return cartesian;		
	}

	MappingDefinition<MathVector<T>,MathVector<T> >* getInverse() const {
		return new Cartesian2Spherical<T>();
	}

private:
	friend class Cartesian2Spherical<T>;

	Spherical2Cartesian(){}
	// forbid copies
	Spherical2Cartesian(Spherical2Cartesian&){_NEMO_LOGIC_ERROR("Copying not allowed");}
	Spherical2Cartesian operator=(const Spherical2Cartesian&){_NEMO_LOGIC_ERROR("Copying not allowed");}
};

/**
 * This mapping type transforms n-dimensional cartesian coordinates to spherical coordinates.
 * The sperical coordinates are computed accordingly to the conventions in Spherical2Cartesian.
 * 
 * This mapping type inverts itself with Spherical2Cartesian mappings.
 */
template <typename T=double>
class Cartesian2Spherical : public MappingDefinition<MathVector<T>,MathVector<T> > {
public:
	static Mapping<MathVector<T>,MathVector<T> > create(){
		return Mapping<MathVector<T>,MathVector<T> >(new Cartesian2Spherical<T>());
	}
	
	MathVector<T> evaluate(MathVector<T> cartesian) const{
		const unsigned int d = cartesian.dimension();
		if(d<2){
			return cartesian;
		}

		MathVector<T> spherical(dim(d));

		// polar angle
		spherical[1] = atan2( cartesian[1], cartesian[0] );
		if(spherical[1] < (T)0){
			spherical[1] = 2.0*M_PI + spherical[1];
		}

		// azimut angles
		T squareSum = cartesian[0]*cartesian[0] + cartesian[1]*cartesian[1];
		for(unsigned int i=2; i<d; i++){
			if(squareSum==(T)0){
				spherical[i] = M_PI/2.0 * (T)sgn(cartesian[i]);
			}
			else{
				//spherical[i] = arccot( cartesian[i], sqrt(squareSum) );
				spherical[i] = atan( cartesian[i]/sqrt(squareSum) );
			}
			squareSum += cartesian[i]*cartesian[i];
		}

		// special case: all 0
		if(squareSum == (T)0){
			return MathVector<T>(dim(d), (T)0);
		}

		// radius
		spherical[0] = sqrt(squareSum);

		return spherical;
	}

	MappingDefinition<MathVector<T>,MathVector<T> >* getInverse() const {
		return new Spherical2Cartesian<T>();
	}

private:
	friend class Spherical2Cartesian<T>;

	Cartesian2Spherical(){}
	// forbid copies
	Cartesian2Spherical(Cartesian2Spherical&){_NEMO_LOGIC_ERROR("Copying not allowed");}
	Cartesian2Spherical operator=(const Cartesian2Spherical&){_NEMO_LOGIC_ERROR("Copying not allowed");}

	static T arccot(const T first, const T second){
		if(second == (T)0){
			//if(second == (T)0){
			//	return (T)0;
			//}
			return M_PI/2.0 * (1 - sgn(first));
		}
		//else
		return M_PI/2.0 - atan(first/second);
	}

	static int sgn(const T x){
		return (x<0)?(-1):(0<x);
	} 
};


} // end namespace


#endif

