
#define BOOST_TEST_MODULE Filters
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Filters.h"
#include "../src/nemo/Vector.h"
#include "../src/nemo/Norms.h"

#include <iostream>
#include <cmath>

using namespace nemo;
using namespace boost::unit_test;

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

typedef Mapping<RealVector,RealVector > VectorMapping;

/*class AbsNorm : public MappingDefinition<double,double> {
public:
	double evaluate(double value) const {return fabs(value);}
};
class EuclideanNorm : public MappingDefinition<RealVector,double> {
public:
	double evaluate(RealVector value) const {return sqrt(value.transpose()*value);}
};*/


BOOST_AUTO_TEST_SUITE(Filters)


BOOST_AUTO_TEST_CASE(firstOrderSmoothingScalar){
	Mapping<double,double> m1 = FirstOrderFilter<double,double>::create(0.25, absNorm<double>(), 0.0);
	BOOST_CHECK_EQUAL(m1(0.0), 0.0);
	BOOST_CHECK_EQUAL(m1(1.0), 0.25);
	BOOST_CHECK_EQUAL(m1(1.0), 0.5);
	BOOST_CHECK_EQUAL(m1(1.0), 0.75);
	BOOST_CHECK_EQUAL(m1(1.0), 1.0);
	BOOST_CHECK_EQUAL(m1(1.0), 1.0);
	BOOST_CHECK_EQUAL(m1(0.0), 0.75);
	BOOST_CHECK_EQUAL(m1(1.0), 1.0);
}

BOOST_AUTO_TEST_CASE(firstOrderSmoothingVectorial){
	Mapping<RealVector,RealVector> m1 = FirstOrderFilter<RealVector,double>::create(
			0.25, euclideanNorm<RealVector,double>(), RealVector(0.0,0.0)
	);

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m1(RealVector(2.0)), std::invalid_argument);
#endif

	BOOST_CHECK_EQUAL(m1(RealVector(0.0,0.0)), RealVector(0.0,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.25,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.5,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.75,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(1.0,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(1.0,0.0));

	RealVector v;
	// make a diagonal move (total diff sqrt(2)=1.4... , so it should take 6 steps)
	for(unsigned int i=0; i<5; i++){
		v = m1(RealVector(0.0, 1.0));
		BOOST_CHECK( v[0] > 0.0 );
		BOOST_CHECK( v[1] < 1.0 );
		BOOST_CHECK_SMALL( (v[0]+v[1])-1.0, 1e-12 );
	}
	BOOST_CHECK_EQUAL(m1(RealVector(0.0, 1.0)), RealVector(0.0, 1.0));
}


BOOST_AUTO_TEST_CASE(secondOrderSmoothingScalar){
	Mapping<double,double> m1 = SecondOrderFilter<double,double>::create(0.25, 0.125, absNorm<double>(), 0.0, 0.0);
	BOOST_CHECK_EQUAL(m1(0.0), 0.0);
	BOOST_CHECK_EQUAL(m1(1.0), 0.125);
	BOOST_CHECK_EQUAL(m1(1.0), 0.375);
	BOOST_CHECK_EQUAL(m1(1.0), 0.625);
	BOOST_CHECK_EQUAL(m1(1.0), 0.875);
	BOOST_CHECK_EQUAL(m1(1.0), 1.0);
	BOOST_CHECK_EQUAL(m1(1.0), 1.0);
	BOOST_CHECK_EQUAL(m1(0.0), 0.875);
	BOOST_CHECK_EQUAL(m1(1.0), 0.875);
	for(unsigned int i=0; i<10; i++){
		BOOST_CHECK_SMALL(m1(1.0)-1.0, 0.1); // TODO this does not work optimally yet, although hard constraints are fulfilled
	}
	//BOOST_CHECK_EQUAL(m1(1.0), 1.0); // TODO it should eventually reach this value exactly!!!

	Mapping<double,double> m2 = SecondOrderFilter<double,double>::create(1.0, 0.1, absNorm<double>(), 0.0, 0.0);
	BOOST_CHECK_SMALL(m2( 0.0)-0.0, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-0.1, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-0.3, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-0.6, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-1.0, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-1.5, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-2.1, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-2.8, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-3.6, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-4.5, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-5.5, 1e-12);

	BOOST_CHECK_SMALL(m2(10.0)-6.4, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-7.2, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-7.9, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-8.5, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-9.0, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-9.4, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-9.7, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-9.9, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-10.0, 1e-12);
	BOOST_CHECK_SMALL(m2(10.0)-10.0, 1e-12);

	// instantaneous acceleration 
	Mapping<double,double> m3 = SecondOrderFilter<double,double>::create(0.25, 1.0, absNorm<double>(), 0.0, 0.0);
	BOOST_CHECK_EQUAL(m3(0.0), 0.0);
	BOOST_CHECK_EQUAL(m3(1.0), 0.25);
	BOOST_CHECK_EQUAL(m3(1.0), 0.5);
	BOOST_CHECK_EQUAL(m3(1.0), 0.75);
	BOOST_CHECK_EQUAL(m3(1.0), 1.0);
	BOOST_CHECK_EQUAL(m3(1.0), 1.0);
	BOOST_CHECK_EQUAL(m3(0.0), 0.75);
	BOOST_CHECK_EQUAL(m3(1.0), 1.0);
}

BOOST_AUTO_TEST_CASE(secondOrderSmoothingVectorial){
	Mapping<RealVector,double> norm = euclideanNorm<RealVector,double>();

	Mapping<RealVector,RealVector> m1 = SecondOrderFilter<RealVector,double>::create(
			0.25, 0.125, norm, RealVector(0.0,0.0), RealVector(0.0,0.0)
	);

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m1(RealVector(2.0)), std::invalid_argument);
#endif

	BOOST_CHECK_EQUAL(m1(RealVector(0.0,0.0)), RealVector(0.0,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.125,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.375,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.625,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(0.875,0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0,0.0)), RealVector(1.000,0.0));

	Mapping<RealVector,RealVector> m2 = SecondOrderFilter<RealVector,double>::create(
			0.25, 0.125, norm, RealVector(0.0,0.0,0.0), RealVector(0.0,0.0,0.0)
	);
	RealVector lastValue(0.0,0.0,0.0);
	RealVector lastDiff(0.0,0.0,0.0);
	// try breaking it with a totally wild input pattern
	for(unsigned int i=0; i<20; i++){
		RealVector value( 1.0*(i%2), 2.0*(i%3), 3.0*(i%5) );
		RealVector out = m2(value);
		RealVector diff = out-lastValue;

		double f = norm(diff);
		BOOST_CHECK( f <= 0.25+1e-15 );
		BOOST_CHECK( norm(diff-lastDiff) <= 0.125+1e-15 );

		lastValue = out;
		lastDiff = diff;
	}
}


BOOST_AUTO_TEST_SUITE_END()

