
#define BOOST_TEST_MODULE MappingTest
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Mapping.h"
#include "../src/nemo/Vector.h"
#include "../src/nemo/Matrix.h"

#include <cmath>

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MappingTest)


typedef MappingDefinition<double,double> RealMappingDef;
typedef Mapping<double,double> RealMapping;


BOOST_AUTO_TEST_CASE(MappingCreation){
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( RealMapping z = (RealMappingDef*)NULL, std::invalid_argument );
#endif

	RealMapping id = Identity<double>::create();
	BOOST_CHECK_EQUAL( id(42), 42 );

	RealMapping c = 4.2;
	BOOST_CHECK_EQUAL( c(0), 4.2 );
	BOOST_CHECK_EQUAL( c(1), 4.2 );

	RealMapping z;
	BOOST_CHECK_EQUAL( z(0), 0.0 );
	BOOST_CHECK_EQUAL( z(1), 0.0 );
}

class FooMap : public RealMappingDef {
	double evaluate(double value) const{
		return value*2.0;
	}
};

BOOST_AUTO_TEST_CASE(CustomMappingCreation){
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( RealMapping z = (FooMap*)NULL, std::invalid_argument );
#endif

	RealMapping foo = new FooMap();

	BOOST_CHECK_EQUAL( foo(21), 42 );
}

int roundF(double x){return (int)x;}

BOOST_AUTO_TEST_CASE(CFuncMappingCreation){
#ifndef NEMO_ABORT_ON_ERROR
	double (*nil)(double) = NULL;
	BOOST_CHECK_THROW( RealMapping z = nil, std::invalid_argument );
#endif

	RealMapping s = sqrt;
	BOOST_CHECK_EQUAL( s(4), 2 );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( s.inverse(), std::logic_error );
#endif

	Mapping<double,int> r = roundF;
	BOOST_CHECK_EQUAL( r(2.2), 2 );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( r.inverse(), std::logic_error );
#endif

	Mapping<double,double> c(cos, acos);
	Mapping<double,double> ac = c.inverse();

	BOOST_CHECK_EQUAL( c(0), 1 );
	BOOST_CHECK_EQUAL( c(3.14159265358979323846), -1 );
	
	BOOST_CHECK_SMALL( ac(c(0.0))-0.0, 1e-15);
	BOOST_CHECK_SMALL( ac(c(0.5))-0.5, 1e-15);
	BOOST_CHECK_SMALL( ac(c(1.0))-1.0, 1e-15);
	BOOST_CHECK_SMALL( ac(c(1.5))-1.5, 1e-15);
	BOOST_CHECK_SMALL( ac(c(2.0))-2.0, 1e-15);
}

double mult(double x, double y){
	return x*y;
}

double multAdd(double x, double y, double z){
	return x*y + z;
}

BOOST_AUTO_TEST_CASE(BoostBindMappingCreation){
	RealMapping timesTwo = boost::bind(mult, _1, 2.0);
	BOOST_CHECK_EQUAL(timesTwo(0.0), 0.0);
	BOOST_CHECK_EQUAL(timesTwo(1.0), 2.0);

	RealMapping timesTwo2 = boost::bind(mult, 2.0, _1);
	BOOST_CHECK_EQUAL(timesTwo2(0.0), 0.0);
	BOOST_CHECK_EQUAL(timesTwo2(1.0), 2.0);

	RealMapping square = boost::bind(mult, _1, _1);
	BOOST_CHECK_EQUAL(square(0.0), 0.0);
	BOOST_CHECK_EQUAL(square(1.0), 1.0);
	BOOST_CHECK_EQUAL(square(2.0), 4.0);

	RealMapping squarePlusOne = boost::bind(multAdd, _1, _1, 1.0);
	BOOST_CHECK_EQUAL(squarePlusOne(0.0), 1.0);
	BOOST_CHECK_EQUAL(squarePlusOne(1.0), 2.0);
	BOOST_CHECK_EQUAL(squarePlusOne(2.0), 5.0);
}

BOOST_AUTO_TEST_CASE(CastMapping){
	// casting introduced to fix Bug #1396
	Mapping<double,int> round = reinterpret<int>(arg<double>());
	BOOST_CHECK_EQUAL(round(1.5), 1);
	BOOST_CHECK_EQUAL(round(2.0), 2);
	
}


BOOST_AUTO_TEST_SUITE_END()

