
#define BOOST_TEST_MODULE MappingArithmetics
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Mapping.h"
#include "../src/nemo/Vector.h"
#include "../src/nemo/Matrix.h"

#include <sstream>

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MappingArithmetics)


typedef Mapping<double,double> RealMapping;
typedef Mapping<int,int> IntMapping;
typedef Mapping<unsigned int,unsigned int> UIntMapping;

BOOST_AUTO_TEST_CASE(MappingScalarOperations){
	RealMapping id = arg<double>();
	BOOST_CHECK_EQUAL( id(42), 42 );

	RealMapping f1 = arg<double>() + 5.0;
	RealMapping f2 = id + 5.0;
	BOOST_CHECK_EQUAL( f1(2), 7 );
	BOOST_CHECK_EQUAL( f2(2), 7 );

	RealMapping f2Neg = - f2;
	BOOST_CHECK_EQUAL( f2Neg(2), -7 );

	RealMapping f3 = 1 / ( f1*f2 + 1.0 );
	BOOST_CHECK_EQUAL( f3(0), 1.0/26.0 );
	BOOST_CHECK_EQUAL( f3(1), 1.0/37.0 );
	BOOST_CHECK_EQUAL( f3(2), 1.0/50.0 );

	IntMapping f4 = (17 % arg<int>()) - 1;
	BOOST_CHECK_EQUAL( f4(1), -1 );
	BOOST_CHECK_EQUAL( f4(2),  0 );
	BOOST_CHECK_EQUAL( f4(3),  1 );
	BOOST_CHECK_EQUAL( f4(4),  0 );
}

BOOST_AUTO_TEST_CASE(MappingConcatenation){
	RealMapping id = arg<double>();
	const RealMapping cid = arg<double>();

	BOOST_CHECK_EQUAL( id(cid)(42.0), 42.0 );
	BOOST_CHECK_EQUAL( cid(id)(42.0), 42.0 );

	const RealMapping p1 = arg<double>()+1;
	const RealMapping t2 = arg<double>()*2;
	BOOST_CHECK_EQUAL( p1(t2)(2.0), 5.0 );
}

typedef Mapping<double,bool> RealTest;
typedef Mapping<int,bool> IntTest;

BOOST_AUTO_TEST_CASE(MappingLogicOperations){
	RealTest t = true;
	RealTest f = false;
	BOOST_CHECK_EQUAL( t(1.0), true );
	BOOST_CHECK_EQUAL( f(1.0), false );

	RealTest t1 = arg<double>() < 10.0;
	BOOST_CHECK_EQUAL( t1(-10.0), true );
	BOOST_CHECK_EQUAL( t1(  0.0), true );
	BOOST_CHECK_EQUAL( t1(  9.0), true );
	BOOST_CHECK_EQUAL( t1( 10.0), false );
	BOOST_CHECK_EQUAL( t1( 20.0), false );

	RealTest t2 = -10.0 < arg<double>() && arg<double>() < 10.0;
	BOOST_CHECK_EQUAL( t2(-20.0), false );
	BOOST_CHECK_EQUAL( t2(-10.0), false );
	BOOST_CHECK_EQUAL( t2( -9.0), true );
	BOOST_CHECK_EQUAL( t2(  0.0), true );
	BOOST_CHECK_EQUAL( t2(  9.0), true );
	BOOST_CHECK_EQUAL( t2( 10.0), false );
	BOOST_CHECK_EQUAL( t2( 20.0), false );

	RealTest t3 = arg<double>() > 10.0 || -10.0 > arg<double>();
	BOOST_CHECK_EQUAL( t3(-20.0), true );
	BOOST_CHECK_EQUAL( t3(  0.0), false );
	BOOST_CHECK_EQUAL( t3( 20.0), true );

	RealTest t4 = !t2;
	BOOST_CHECK_EQUAL( t4(-20.0), true );
	BOOST_CHECK_EQUAL( t4(  0.0), false );
	BOOST_CHECK_EQUAL( t4( 20.0), true );
	
}

typedef Mapping<RealVector,RealVector> VectorMapping;

BOOST_AUTO_TEST_CASE(MappingVectorOperations){

	VectorMapping id = arg<RealVector>() + RealVector(dim(3), 0.0);
	BOOST_CHECK_EQUAL( id(RealVector(1.0, 2.0, 3.0)), RealVector(1.0, 2.0, 3.0) );

	VectorMapping timesTwo = 2.0 * arg<RealVector>();
	BOOST_CHECK_EQUAL( timesTwo(RealVector(1.0, 2.0)), RealVector(2.0, 4.0) );

	RealMatrix m(dim(2,2));
	m(0,0) = 1.0;
	m(1,0) = 0.0;
	m(0,1) = 0.0;
	m(1,1) = 1.0;
	VectorMapping id2 = m * arg<RealVector>();
	BOOST_CHECK_EQUAL( id2(RealVector(1.0, 2.0)), RealVector(1.0, 2.0) );
	VectorMapping timesTwo2 = (m*2) * arg<RealVector>();
	BOOST_CHECK_EQUAL( timesTwo2(RealVector(1.0, 2.0)), RealVector(2.0, 4.0) );
}



BOOST_AUTO_TEST_CASE(MappingVectorCWiseOperations){
	VectorMapping leftMult = RealVector(1.0, 2.0).cwise() * arg<RealVector>();
	BOOST_CHECK_EQUAL( leftMult(RealVector(1.0, 1.0)), RealVector(1.0, 2.0) );

	VectorMapping inv = RealVector(2.0, 2.0).cwise() / arg<RealVector>();
	BOOST_CHECK_EQUAL( inv(RealVector(1.0, 2.0)), RealVector(2.0, 1.0) );

	VectorMapping div2 = arg<RealVector>() / RealVector(2.0, 2.0).cwise();
	BOOST_CHECK_EQUAL( div2(RealVector(1.0, 2.0)), RealVector(0.5, 1.0) );
}


BOOST_AUTO_TEST_CASE(MappingShiftOperations){
	IntMapping timesFour = arg<int>() << 2;
	BOOST_CHECK_EQUAL( timesFour(17), 68 );

#ifndef _WIN32 // spooky one: MSVC is not willing to evaluate the type information for this operator ^^
	UIntMapping divFour = arg<unsigned int>() >> (unsigned int)2;
	BOOST_CHECK_EQUAL( divFour(17), 4 );
#endif

	// TODO
	//std::stringstream s;
	//Mapping<int,std::ostream&> foo = s << arg<int>();

}

BOOST_AUTO_TEST_CASE(MappingUpTypeOperation){
	// regression for Bug #1396
	Mapping<char,int> i1 = arg<char>()+1;
	Mapping<char,int> i2 = 1+arg<char>();
	BOOST_CHECK_EQUAL( i1((char)17), 18 );
	BOOST_CHECK_EQUAL( i2((char)17), 18 );

	Mapping<float,double> d1 = arg<float>()+(double)1;
	Mapping<float,double> d2 = (double)1+arg<float>();
	BOOST_CHECK_EQUAL( d1((float)2.5), 3.5 );
	BOOST_CHECK_EQUAL( d2((float)2.5), 3.5 );
}
	


BOOST_AUTO_TEST_SUITE_END()

