
#define BOOST_TEST_MODULE MappingInversion
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Mapping.h"
#include "../src/nemo/Vector.h"
#include "../src/nemo/Matrix.h"
#include <../src/nemo/MatrixMath.h>

#include <string>
#include <sstream>

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace std;
using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MappingInversion)


typedef Mapping<double,double> RealMapping;
typedef Mapping<int,int> IntMapping;
typedef Mapping<unsigned int,unsigned int> UIntMapping;

typedef Mapping<string,string> StringMapping;

typedef Mapping<RealVector,RealVector> VectorMapping;
typedef Mapping<RealMatrix,RealMatrix> MatrixMapping;


BOOST_AUTO_TEST_CASE(MappingInversionIdentity){
	Mapping<string,string> idS = Identity<string>::create();
	Mapping<string,string> idSi = idS.inverse();
	BOOST_CHECK_EQUAL( idSi("foo"), string("foo") );
	BOOST_CHECK_EQUAL( idSi(idS("foo")), string("foo") );
}

BOOST_AUTO_TEST_CASE(MappingInversionLeftAddScalar){
	RealMapping d = 2.0 + arg<double>();
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  0.0 );
	BOOST_CHECK_EQUAL( dInv(1.0), -1.0 );

	IntMapping i = 2 + arg<int>();
	IntMapping iInv = i.inverse();
	BOOST_CHECK_EQUAL( iInv(2),  0 );
	BOOST_CHECK_EQUAL( iInv(1), -1 );

#ifndef NEMO_ABORT_ON_ERROR
	UIntMapping ui = ((unsigned int)2) + arg<unsigned int>();
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionRightAddScalar){
	RealMapping d = arg<double>() + 2.0;
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  0.0 );
	BOOST_CHECK_EQUAL( dInv(1.0), -1.0 );

	IntMapping i = arg<int>() + 2;
	IntMapping iInv = i.inverse();
	BOOST_CHECK_EQUAL( iInv(2),  0 );
	BOOST_CHECK_EQUAL( iInv(1), -1 );

#ifndef NEMO_ABORT_ON_ERROR
	UIntMapping ui = arg<unsigned int>() + ((unsigned int)2);
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionLeftSubtractScalar){
	RealMapping d = 2.0 - arg<double>();
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  0.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  1.0 );

	IntMapping i = 2 - arg<int>();
	IntMapping iInv = i.inverse();
	BOOST_CHECK_EQUAL( iInv(2),  0 );
	BOOST_CHECK_EQUAL( iInv(1),  1 );

#ifndef NEMO_ABORT_ON_ERROR
	UIntMapping ui = ((unsigned int)2) - arg<unsigned int>();
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionRightSubtractScalar){
	RealMapping d = arg<double>() - 2.0;
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  4.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  3.0 );

	IntMapping i = arg<int>() - 2;
	IntMapping iInv = i.inverse();
	BOOST_CHECK_EQUAL( iInv(2),  4 );
	BOOST_CHECK_EQUAL( iInv(1),  3 );

#ifndef NEMO_ABORT_ON_ERROR
	UIntMapping ui = arg<unsigned int>() - ((unsigned int)2);
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionLeftMultScalar){
	RealMapping d = 2.0 * arg<double>();
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  1.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  0.5 );

#ifndef NEMO_ABORT_ON_ERROR
	IntMapping i = 2 * arg<int>();
	BOOST_CHECK_THROW( i.inverse(), std::logic_error );

	UIntMapping ui = ((unsigned int)2) * arg<unsigned int>();
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionRightMultScalar){
	RealMapping d = arg<double>() * 2.0;
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  1.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  0.5 );

#ifndef NEMO_ABORT_ON_ERROR
	IntMapping i = arg<int>() * 2;
	BOOST_CHECK_THROW( i.inverse(), std::logic_error );

	UIntMapping ui = arg<unsigned int>() * ((unsigned int)2);
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionLeftDivideScalar){
	RealMapping d = 2.0 / arg<double>();
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(4.0),  0.5 );
	BOOST_CHECK_EQUAL( dInv(2.0),  1.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  2.0 );

#ifndef NEMO_ABORT_ON_ERROR
	IntMapping i = 2 / arg<int>();
	BOOST_CHECK_THROW( i.inverse(), std::logic_error );

	UIntMapping ui = ((unsigned int)2) / arg<unsigned int>();
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionRightDivideScalar){
	RealMapping d = arg<double>() / 2.0;
	RealMapping dInv = d.inverse();
	BOOST_CHECK_EQUAL( dInv(2.0),  4.0 );
	BOOST_CHECK_EQUAL( dInv(1.0),  2.0 );

#ifndef NEMO_ABORT_ON_ERROR
	IntMapping i = arg<int>() / 2;
	BOOST_CHECK_THROW( i.inverse(), std::logic_error );

	UIntMapping ui = arg<unsigned int>() / ((unsigned int)2);
	BOOST_CHECK_THROW( ui.inverse(), std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionSequantialOperators){
	RealMapping d1 = arg<double>() * 2.0;
	RealMapping d2 = arg<double>() - 1.0;

	// chaining creates instances of SequentialMapping's
	RealMapping d3 = d1(d2); // y = (x-1)*2
	RealMapping d4 = d2(d1); // y = (x*2)-1

	RealMapping d3Inv = d3.inverse(); // x = y/2 + 1 ????
	BOOST_CHECK_EQUAL( d3Inv(0.0),  1.0 );
	BOOST_CHECK_EQUAL( d3Inv(1.0),  1.5 );

	RealMapping d4Inv = d4.inverse(); // x = (y+1) / 2 ????
	BOOST_CHECK_EQUAL( d4Inv(0.0),  0.5 );
	BOOST_CHECK_EQUAL( d4Inv(1.0),  1.0 );
}

BOOST_AUTO_TEST_CASE(MappingInversionSequantialOperatorsDirect){
	// same functions as in above test, but write them down directly
	// ==> NO SequentialMapping will be created, but they will be chained in place of the Operator Mappings
	RealMapping d3 = (arg<double>() - 1.0) * 2.0;
	RealMapping d4 = (arg<double>() * 2.0) - 1.0;

	RealMapping d3Inv = d3.inverse(); // x = y/2 + 1 ????
	BOOST_CHECK_EQUAL( d3Inv(0.0),  1.0 );
	BOOST_CHECK_EQUAL( d3Inv(1.0),  1.5 );

	RealMapping d4Inv = d4.inverse(); // x = (y+1) / 2 ????
	BOOST_CHECK_EQUAL( d4Inv(0.0),  0.5 );
	BOOST_CHECK_EQUAL( d4Inv(1.0),  1.0 );
}

BOOST_AUTO_TEST_CASE(MappingInversionStringAddition){
	StringMapping s = arg<string>() + string("-foo");

	BOOST_CHECK_EQUAL( s("test"),  string("test-foo") );

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( s.inverse(),  std::logic_error );
#endif
}

BOOST_AUTO_TEST_CASE(MappingInversionVectorOperators){
	VectorMapping v1 = arg<RealVector>() + RealVector(1.0, 2.0);
	VectorMapping v1Inv = v1.inverse();
	BOOST_CHECK_EQUAL( v1(RealVector(0.0, 0.0)), RealVector(1.0, 2.0) );
	BOOST_CHECK_EQUAL( v1Inv(RealVector(1.0, 2.0)), RealVector(0.0, 0.0) );

	VectorMapping v2 = RealVector(1.0, 2.0) - arg<RealVector>();
	VectorMapping v2Inv = v2.inverse();
	BOOST_CHECK_EQUAL( v2Inv(RealVector(1.0, 2.0)), RealVector(0.0, 0.0) );
}

/*BOOST_AUTO_TEST_CASE(MappingInversionVectorCWiseOperators){
	VectorMapping v1 = RealVector(1.0, 2.0).cwise() * arg<RealVector>();
	VectorMapping v1Inv = v1.inverse();
	BOOST_CHECK_EQUAL( v1Inv(RealVector(1.0, 2.0)), RealVector(1.0, 1.0) );
	BOOST_CHECK_EQUAL( v1Inv(RealVector(2.0, 2.0)), RealVector(2.0, 1.0) );

	VectorMapping v2 = RealVector(1.0, 2.0).cwise() / arg<RealVector>();
	VectorMapping v2Inv = v2.inverse();
	BOOST_CHECK_EQUAL( v2Inv(RealVector(1.0, 2.0)), RealVector(1.0, 1.0) );
	BOOST_CHECK_EQUAL( v2Inv(RealVector(2.0, 2.0)), RealVector(0.5, 1.0) );
}*/



BOOST_AUTO_TEST_CASE(MappingInversionMatrixAddition){
	MatrixMapping m1 = arg<RealMatrix>() + RealMatrix(dim(2,2), 2.0);
	MatrixMapping m1Inv = m1.inverse();
	BOOST_CHECK_EQUAL( m1(RealMatrix(dim(2,2), 1.0)), RealMatrix(dim(2,2), 3.0) );
	BOOST_CHECK_EQUAL( m1Inv(RealMatrix(dim(2,2), 3.0)), RealMatrix(dim(2,2), 1.0) );

	MatrixMapping m2 = arg<RealMatrix>() - RealMatrix(dim(2,2), 2.0);
	MatrixMapping m2Inv = m2.inverse();
	BOOST_CHECK_EQUAL( m2(RealMatrix(dim(2,2), 1.0)), RealMatrix(dim(2,2), -1.0) );
	BOOST_CHECK_EQUAL( m2Inv(RealMatrix(dim(2,2), -1.0)), RealMatrix(dim(2,2), 1.0) );

	MatrixMapping m3 = RealMatrix(dim(2,2), 2.0) - arg<RealMatrix>();
	MatrixMapping m3Inv = m3.inverse();
	BOOST_CHECK_EQUAL( m3(RealMatrix(dim(2,2), 0.5)), RealMatrix(dim(2,2), 1.5) );
	BOOST_CHECK_EQUAL( m3Inv(RealMatrix(dim(2,2), 1.5)), RealMatrix(dim(2,2), 0.5) );
}

BOOST_AUTO_TEST_CASE(MappingInversionMatrixMult){
	RealMatrix multMat(dim(2,2));
	multMat(0,0) = 0.0;
	multMat(0,1) = 1.0;
	multMat(1,0) = 1.0;
	multMat(1,1) = 0.0;
	MatrixMapping m1 = arg<RealMatrix>() * multMat;
	MatrixMapping m1Inv = m1.inverse();

	RealMatrix in(dim(1,2));
	in(0,0) = 3.0;
	in(0,1) = 5.0;
	RealMatrix out(dim(1,2));
	out(0,0) = 5.0;
	out(0,1) = 3.0;

	BOOST_CHECK_EQUAL( m1(in), out );
	BOOST_CHECK_EQUAL( m1Inv(out), in );
}

BOOST_AUTO_TEST_CASE(MappingInversionLinear){
	RealMatrix multMat(dim(2,2));
	multMat(0,0) = 0.0;
	multMat(0,1) = 1.0;
	multMat(1,0) = 1.0;
	multMat(1,1) = 0.0;

	VectorMapping v1 = multMat * arg<RealVector>() + RealVector(2.0, 3.0);
	VectorMapping v1Inv = v1.inverse();

	RealVector in1(5.0, 7.0);
	RealVector out1(9.0, 8.0);
	BOOST_CHECK_EQUAL( v1(in1), out1 );
	BOOST_CHECK_EQUAL( v1Inv(out1), in1 );

	RealVector in2(0.0, 0.0);
	RealVector out2(2.0, 3.0);
	BOOST_CHECK_EQUAL( v1(in2), out2 );
	BOOST_CHECK_EQUAL( v1Inv(out2), in2 );
}

BOOST_AUTO_TEST_CASE(MappingUpTypeOperationInversion){
	// regression for Bug #1396
	Mapping<char,int> i1 = arg<char>()+1;
	Mapping<char,int> i2 = 1+arg<char>();
	BOOST_CHECK_EQUAL( i1.inverse()((int)18), (char)17 );
	BOOST_CHECK_EQUAL( i2.inverse()((int)18), (char)17 );

	Mapping<float,double> d1 = arg<float>()+(double)1;
	Mapping<float,double> d2 = (double)1+arg<float>();
	BOOST_CHECK_EQUAL( d1.inverse()((double)3.5), (float)2.5 );
	BOOST_CHECK_EQUAL( d2.inverse()((double)3.5), (float)2.5 );
}


BOOST_AUTO_TEST_SUITE_END()



