
#define BOOST_TEST_MODULE MathVectorTestArithmetics
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Vector.h"

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MathVectorTestArithmetics)


BOOST_AUTO_TEST_CASE(mathVectorAdd){
	RealVector vDim0;
	RealVector vDim0Add = vDim0 + vDim0;
	BOOST_CHECK_EQUAL( vDim0Add.dimension(), 0 );
	BOOST_CHECK_EQUAL( (vDim0 + vDim0).dimension(), 0 );

	RealVector v1(1.0, 1.0);
	RealVector v1B(1.0, 1.0);
	RealVector v2(2.0, 2.0);
	RealVector v2B(2.0, 2.0);
	RealVector v3;

	// "copy" construction 
	RealVector v4 = v1 + v2;
	BOOST_CHECK_EQUAL( v4.dimension(), 2 );
	BOOST_CHECK_CLOSE( (double)v4[0], 3.0, 1E-6 );
	BOOST_CHECK_CLOSE( (double)v4[1], 3.0, 1E-6 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

	v3 = v1 + v2;
	BOOST_CHECK_EQUAL( v3.dimension(), 2 );
	BOOST_CHECK_CLOSE( (double)v3[0], 3.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)v3[1], 3.0, 1E-6 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

	v3 = v1 + v2 + v3;
	BOOST_CHECK_EQUAL( v3.dimension(), 2 );
	BOOST_CHECK_CLOSE( (double)v3[0], 6.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)v3[1], 6.0, 1E-6 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

	// check throw on dimension mismatch
	RealVector z = RealVector() + RealVector(); // no throw on zero dimension
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( z + RealVector(dim(3)), std::invalid_argument );
	BOOST_CHECK_THROW( v1 + RealVector(dim(1)), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorAddInplace){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0 += vDim0).dimension(), 0 );

	RealVector v1(1.0, 1.0);
	RealVector v1B(v1);
	RealVector v2(2.0, 2.0);

	v1 += v1;
	BOOST_CHECK_EQUAL( v1, v2 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v1B[1], 1.0 ); // unchanged?
	
	v2 += v1;
	BOOST_CHECK_EQUAL( v2, RealVector(dim(2), 4.0) );
	BOOST_CHECK_EQUAL( v1[0], 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v1[1], 2.0 ); // unchanged?


	// check throw on dimension mismatch
	RealVector z;
	z += RealVector(); // no throw on zero dimension
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( z += RealVector(dim(3)), std::invalid_argument );
	BOOST_CHECK_EQUAL( z.dimension(), 0 ); // unchanged?
	BOOST_CHECK_THROW( v1 += RealVector(dim(1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1.dimension(), 2 ); // unchanged?
	BOOST_CHECK_EQUAL( v1[0], 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v1[1], 2.0 ); // unchanged?
#endif
}


BOOST_AUTO_TEST_CASE(mathVectorDiff){
	RealVector vDim0;

	BOOST_CHECK_EQUAL( (vDim0 - vDim0).dimension(), 0 );

	RealVector v1(1.0, 1.0);
	RealVector v1B(1.0, 1.0);
	RealVector v2(2.0, 2.0);
	RealVector v2B(2.0, 2.0);
	RealVector v3(3.0, 3.0);
	RealVector v3B(3.0, 3.0);

	// "copy" construction 
	RealVector v4 = v3 - v1;
	BOOST_CHECK_EQUAL( v3.dimension(), 2 );
	BOOST_CHECK_CLOSE( (double)v4[0], 2.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)v4[1], 2.0, 1E-6 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v3, v3B ); // unchanged?

	// pure assignment
	v2 = v3 - v1;
	BOOST_CHECK_EQUAL( v2.dimension(), 2 );
	BOOST_CHECK_CLOSE( (double)v2[0], 2.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)v2[1], 2.0, 1E-6 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v3, v3B ); // unchanged?


	// check throw on dimension mismatch
	RealVector z = RealVector() - RealVector(); // no throw on zero dimension
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( z - RealVector(dim(3)), std::invalid_argument );
	BOOST_CHECK_THROW( v1 - RealVector(dim(1)), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorDiffInplace){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0 -= vDim0).dimension(), 0 );

	RealVector v1(2.0, 2.0);
	RealVector v1B(v1);
	RealVector v2(1.0, 1.0);
	RealVector v2B(v2);

	v1 -= v2;
	BOOST_CHECK_EQUAL( v1, RealVector(1.0, 1.0) );
	BOOST_CHECK_EQUAL( v1B[0], 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v1B[1], 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?
	
	v2 -= v1;
	BOOST_CHECK_EQUAL( v2, RealVector(dim(2), 0.0) );
	BOOST_CHECK_EQUAL( v1, RealVector(dim(2), 1.0) ); // unchanged?


	// check throw on dimension mismatch
	RealVector z;
	z -= RealVector(); // no throw on zero dimension
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( z -= RealVector(dim(3)), std::invalid_argument );
	BOOST_CHECK_EQUAL( z.dimension(), 0 ); // unchanged?
	BOOST_CHECK_THROW( v1 -= RealVector(dim(1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, RealVector(dim(2), 1.0) ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorMultScalar){
	RealVector v1(dim(2));
	v1[0] = 1.0; v1[1] = 2.0;
	RealVector v1B(v1);
	RealVector v2(dim(2));
	v2[0] = 2.0; v2[1] = 4.0;
	RealVector v2B(v2);

	// left/right mult
	RealVector v3 = v1 * 2.0;
	BOOST_CHECK_EQUAL( v3, v2 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	RealVector v4 = 2.0 * v1;
	BOOST_CHECK_EQUAL( v4, v2 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?

	// assign instead of copy-construct
	v3 = v1 * 2.0;
	BOOST_CHECK_EQUAL( v3, v2 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	v4 = 2.0 * v1;
	BOOST_CHECK_EQUAL( v4, v2 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?

	// chain mults
	RealVector v5 = 2.0 * v1 *2.0;
	RealVector v6 = v1 * 2.0 * 2.0;
	BOOST_CHECK_EQUAL( v5, v6 );
	BOOST_CHECK_EQUAL( v5, RealVector(4.0, 8.0) );
}

BOOST_AUTO_TEST_CASE(mathVectorMultScalarInplace){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0 *= 2.0).dimension(), 0 );

	RealVector v1(dim(2), 1.0);
	RealVector v1B(v1);
	RealVector v2(dim(2), 2.0);

	v1 *= 2.0;
	BOOST_CHECK_EQUAL( v1, v2 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 ); // unchanged?
	BOOST_CHECK_EQUAL( v1B[1], 1.0 ); // unchanged?
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseMult){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0.cwise() * vDim0).dimension(), 0 );

	RealVector v1(5.0, 7.0);
	RealVector v1B(5.0, 7.0);
	RealVector v2(3.0, 2.0);
	RealVector v2B(3.0, 2.0);

	RealVector v3 = v1.cwise() * v2;
	BOOST_CHECK_EQUAL( v3, RealVector(15.0, 14.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

	RealVector v4 = v1 * v2.cwise();
	BOOST_CHECK_EQUAL( v4, RealVector(15.0, 14.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1.cwise() * RealVector(dim(1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseMultInplace){
	RealVector vDim0;
	RealVector vDim02 = (vDim0 *= vDim0.cwise());
	BOOST_CHECK_EQUAL( vDim02.dimension(), 0 );

	RealVector v1(5.0, 7.0);
	RealVector v1B(5.0, 7.0);
	RealVector v2(3.0, 2.0);
	RealVector v2B(3.0, 2.0);

	v1 *= v2.cwise();
	BOOST_CHECK_EQUAL( v1, RealVector(15.0, 14.0) );
	BOOST_CHECK_EQUAL( v1B, RealVector(5.0, 7.0) ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1 *= RealVector(dim(1)).cwise(), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, RealVector(15.0, 14.0) ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseDiv){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0.cwise() / vDim0).dimension(), 0 );

	RealVector v1(6.0, 35.0);
	RealVector v1B(6.0, 35.0);
	RealVector v2(3.0, 5.0);
	RealVector v2B(3.0, 5.0);

	RealVector v3 = v1.cwise() / v2;
	BOOST_CHECK_EQUAL( v3, RealVector(2.0, 7.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

	RealVector v4 = v1 / v2.cwise();
	BOOST_CHECK_EQUAL( v4, RealVector(2.0, 7.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1.cwise() / RealVector(dim(1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseDivInplace){
	RealVector vDim0;
	RealVector vDim02 = (vDim0 *= vDim0.cwise());
	BOOST_CHECK_EQUAL( vDim02.dimension(), 0 );

	RealVector v1(6.0, 35.0);
	RealVector v1B(6.0, 35.0);
	RealVector v2(3.0, 5.0);
	RealVector v2B(3.0, 5.0);

	v1 /= v2.cwise();
	BOOST_CHECK_EQUAL( v1, RealVector(2.0, 7.0) );
	BOOST_CHECK_EQUAL( v1B, RealVector(6.0, 35.0) ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1 /= RealVector(dim(1)).cwise(), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, RealVector(2.0, 7.0) ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseMax){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0.cwise().max(vDim0)).dimension(), 0 );

	RealVector v1(1.0, 4.0);
	RealVector v1B(1.0, 4.0);
	RealVector v2(3.0, 2.0);
	RealVector v2B(3.0, 2.0);

	RealVector v3 = v1.cwise().max(v2);
	BOOST_CHECK_EQUAL( v3, RealVector(3.0, 4.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1.cwise().max(RealVector(dim(1))), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseMin){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( (vDim0.cwise().min(vDim0)).dimension(), 0 );

	RealVector v1(1.0, 4.0);
	RealVector v1B(1.0, 4.0);
	RealVector v2(3.0, 2.0);
	RealVector v2B(3.0, 2.0);

	RealVector v3 = v1.cwise().min(v2);
	BOOST_CHECK_EQUAL( v3, RealVector(1.0, 2.0) );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1.cwise().min(RealVector(dim(1))), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(mathVectorScalarProduct){
	RealVector vDim0;
	BOOST_CHECK_EQUAL( vDim0.transpose()*vDim0, 0.0 );

	RealVector v1(1.0, 4.0);
	RealVector v1B(1.0, 4.0);
	RealVector v2(3.0, 2.0);
	RealVector v2B(3.0, 2.0);

	double s = v1.transpose() * v2;
	BOOST_CHECK_EQUAL( s, 11.0 );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
	BOOST_CHECK_EQUAL( v2, v2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1.transpose()*RealVector(dim(1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( v1, v1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_SUITE_END()

