
#define BOOST_TEST_MODULE MathVectorCreateAccess
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Vector.h"

#include <cmath>

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MathVectorCreateAccess)


BOOST_AUTO_TEST_CASE(mathVectorCreate) {
	RealVector v1;
	BOOST_CHECK_EQUAL(v1.dimension(), 0);

	RealVector v2(dim(3));
	BOOST_CHECK_EQUAL(v2.dimension(), 3);
	BOOST_CHECK_EQUAL(v2[0], 0.0);

	RealVector v3(dim(3), 13.0);
	BOOST_CHECK_EQUAL(v3.dimension(), 3);
	BOOST_CHECK_EQUAL(v3[0], 13.0);

	RealVector v1d(13.0);
	BOOST_CHECK_EQUAL(v1d.dimension(), 1);
	BOOST_CHECK_EQUAL(v1d[0], 13.0);

	RealVector v2d(13.0, 14.0);
	BOOST_CHECK_EQUAL(v2d.dimension(), 2);
	BOOST_CHECK_EQUAL(v2d[0], 13.0);
	BOOST_CHECK_EQUAL(v2d[1], 14.0);

	RealVector v3d(13.0, 14.0, 15.0);
	BOOST_CHECK_EQUAL(v3d.dimension(), 3);
	BOOST_CHECK_EQUAL(v3d[0], 13.0);
	BOOST_CHECK_EQUAL(v3d[1], 14.0);
	BOOST_CHECK_EQUAL(v3d[2], 15.0);

#ifdef _NEMO_HAS_INIT_LISTS_
	RealVector v4 = {};
	BOOST_CHECK_EQUAL( v4.dimension(), 0);
	RealVector v5 = {1.0, 2.0, 3.0};
	BOOST_CHECK_EQUAL( v5.dimension(), 3);
	BOOST_CHECK_EQUAL( v5, RealVector(1.0, 2.0, 3.0));
#endif

	// create from array
	double vals[3];
	vals[0] = 1.0; vals[1] = 2.0; vals[2] = 3.0;
	BOOST_CHECK_EQUAL( RealVector::fromArray(dim(3),vals), RealVector(1.0, 2.0, 3.0));

	// create from other-type array
	int intVals[3];
	intVals[0] = 1; intVals[1] = 2; intVals[2] = 3;
	BOOST_CHECK_EQUAL( RealVector::fromArray(dim(3),intVals), RealVector(1.0, 2.0, 3.0));

	// create with generator
	BOOST_CHECK_EQUAL( RealVector(dim(3),uniform<double>(1.0, 1.0)), RealVector(dim(3), 1.0) );
	RealVector vDim3(dim(3));
	BOOST_CHECK_EQUAL( RealVector(vDim3.dimension(),uniform<double>(1.0, 1.0)), RealVector(dim(3), 1.0) );

	// create with other vector's dimension object (thanks to cemmeric for this test case)
	BOOST_CHECK_EQUAL(RealVector( vDim3.dimension(), 42.0), RealVector(dim(3), 42.0));
	RealVector vDim2(dim(2));
	// type confusion between <unsigned int> and VectorDimension's <unsigned int>-conversion???
	BOOST_CHECK_EQUAL(
		MathVector<unsigned int>(vDim2.dimension(), (unsigned int)3), 
		MathVector<unsigned int>(dim(2), (unsigned int)3)
	); 
}

BOOST_AUTO_TEST_CASE(mathVectorPass) {
	RealVector v1;
	RealVector v2(v1);
	BOOST_CHECK_EQUAL(v2.dimension(), 0);
	BOOST_CHECK_EQUAL(v1, v2);

	RealVector v3(1.0);
	RealVector v4(2.0);
	RealVector v5(2.0, 2.0);
//	std::cout << v5 << std::endl;
	BOOST_REQUIRE( v3 != v4);
	BOOST_REQUIRE( v3 != v5);
	v4 = v3;
	BOOST_CHECK_EQUAL(v3, v4);
	v5 = v3;
	BOOST_CHECK_EQUAL(v5.dimension(), 1);
	BOOST_CHECK_EQUAL(v3, v5);
}

BOOST_AUTO_TEST_CASE(mathVectorAccess1) {
	RealVector v0d;
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v0d[0], std::invalid_argument )
#endif

	RealVector v2d(dim(2));
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v2d[2], std::invalid_argument )
#endif

	// non-const operator[]
	RealVector v1(1.0, 1.0);
	RealVector v1B(v1);
	BOOST_CHECK_EQUAL( v1[0], 1.0 );
	BOOST_CHECK_EQUAL( v1[1], 1.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );
	BOOST_CHECK_EQUAL( v1B[1], 1.0 );

	v1[0] = 42.0;
	BOOST_CHECK_EQUAL( v1[0], 42.0 );
	BOOST_CHECK_EQUAL( v1[1], 1.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );
	BOOST_CHECK_EQUAL( v1B[1], 1.0 );

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( v1[2] = 42.0, std::invalid_argument );
#endif
	BOOST_CHECK_EQUAL( v1[0], 42.0 );
	BOOST_CHECK_EQUAL( v1[1], 1.0 );

	// const operator[]
	const RealVector v2(2.0, 2.0);
	const RealVector v2B(v2);
	BOOST_CHECK_EQUAL( v2[0], 2.0 );
	BOOST_CHECK_EQUAL( v2[1], 2.0 );
	BOOST_CHECK_EQUAL( v2B[0], 2.0 );
	BOOST_CHECK_EQUAL( v2B[1], 2.0 );
}

BOOST_AUTO_TEST_CASE(mathVectorAccess2) {
	// non-const operator[]
	RealVector v1(1.0, 1.0);
	RealVector v1B(v1);
	RealVector v2(2.0, 2.0);
	RealVector v2B(v2);

	v1[0] = v2[0];
	BOOST_CHECK_EQUAL( v1[0], v2[0] );
	BOOST_CHECK_EQUAL( v1[1], 1.0 );
	BOOST_CHECK_EQUAL( v2, v2B );

	// const operator[]
	const RealVector c1(3.0, 3.0);
	const RealVector c1B(c1);

	v1[0] = c1[0];
	BOOST_CHECK_EQUAL( v1[0], c1[0] );
	BOOST_CHECK_EQUAL( v1[1], 1.0 );
	BOOST_CHECK_EQUAL( c1, c1B );

	// use non-const[] inside something else
	RealVector z(dim(2), 0.0);
	double a = std::sin(z[0]);
	BOOST_CHECK_EQUAL( a, 0.0 );

	double b = 2*v1[1] + v2[1];
	BOOST_CHECK_EQUAL( b, 4.0 );
}


BOOST_AUTO_TEST_CASE(mathVectorAccessCompound){
	RealVector v1(1.0, 1.0);
	RealVector v1B(v1);
	RealVector v2(2.0, 2.0);
	RealVector v2B(v2);

	v1[0] += 1.0;
	BOOST_CHECK_EQUAL( v1[0], 2.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );

	v1[0] *= 2.0;
	BOOST_CHECK_EQUAL( v1[0], 4.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );

	v1[0] /= 4.0;
	BOOST_CHECK_EQUAL( v1[0], 1.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );

	v1[0] -= 1.0;
	BOOST_CHECK_EQUAL( v1[0], 0.0 );
	BOOST_CHECK_EQUAL( v1B[0], 1.0 );

	MathVector<int> i1(1, 1);
	MathVector<int> i1B(i1);
	
	/*BOOST_CHECK_EQUAL( i1[0]++, 1 );
	BOOST_CHECK_EQUAL( i1[0], 2 );
	BOOST_CHECK_EQUAL( i1B[0], 1 );
	
	BOOST_CHECK_EQUAL( ++i1[0], 3 );
	BOOST_CHECK_EQUAL( i1[0], 3 );
	BOOST_CHECK_EQUAL( i1B[0], 1 );
	
	BOOST_CHECK_EQUAL( i1[0]--, 3 );
	BOOST_CHECK_EQUAL( i1[0], 2 );
	BOOST_CHECK_EQUAL( i1B[0], 1 );
	
	BOOST_CHECK_EQUAL( --i1[0], 1 );
	BOOST_CHECK_EQUAL( i1[0], 1 );
	BOOST_CHECK_EQUAL( i1B[0], 1 );*/
}

BOOST_AUTO_TEST_CASE(mathVectorAccessSwap){
	RealVector v1(1.0, 2.0);
	RealVector v2 = v1;

	using std::swap;
	swap( v1[0], v1[1] );
	BOOST_CHECK_EQUAL( v1, RealVector(2.0, 1.0) );
	BOOST_CHECK_EQUAL( v2, RealVector(1.0, 2.0) );
}

BOOST_AUTO_TEST_CASE(mathVectorAccessMax){
	RealVector v(1.0, 2.0);

	using std::max;
	BOOST_CHECK_EQUAL( max(v[0], v[1]), 2.0 );
	BOOST_CHECK_EQUAL( max(v[0], 2.0), 2.0 );
	BOOST_CHECK_EQUAL( max(1.0, v[1]), 2.0 );
}
BOOST_AUTO_TEST_CASE(mathVectorAccessMin){
	RealVector v(1.0, 2.0);

	using std::min;
	BOOST_CHECK_EQUAL( min(v[0], v[1]), 1.0 );
	BOOST_CHECK_EQUAL( min(v[0], 2.0), 1.0 );
	BOOST_CHECK_EQUAL( min(1.0, v[1]), 1.0 );
}

BOOST_AUTO_TEST_CASE(mathVectorSetSubVector){
	RealVector v0;
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(v0.setSubVector(0,RealVector()), std::invalid_argument);
#endif

	RealVector v1(1.0, 2.0);
	RealVector v2 = v1;
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(v1.setSubVector(0, RealVector(dim(3))), std::invalid_argument);
	BOOST_CHECK_THROW(v1.setSubVector(2, RealVector(dim(1))), std::invalid_argument);
	BOOST_CHECK_THROW(v1.setSubVector(1, RealVector(dim(2))), std::invalid_argument);
#endif
	v1.setSubVector(0, RealVector(3.0));
	BOOST_CHECK_EQUAL( v1, RealVector(3.0, 2.0) );
	BOOST_CHECK_EQUAL( v2, RealVector(1.0, 2.0) );
	v1.setSubVector(1, RealVector(4.0));
	BOOST_CHECK_EQUAL( v1, RealVector(3.0, 4.0) );
	BOOST_CHECK_EQUAL( v2, RealVector(1.0, 2.0) );
}

BOOST_AUTO_TEST_CASE(mathVectorCWiseConstruct){
	RealVector::CWise cDim0;
	BOOST_CHECK_EQUAL( ((RealVector)cDim0).dimension(), 0 );
	RealVector dim0;
	BOOST_CHECK_NO_THROW(dim0 / cDim0);
	
	RealVector::CWise cDim2 = RealVector(1.0, 2.0).cwise();
	BOOST_CHECK_EQUAL( ((RealVector)cDim2).dimension(), 2 );

	RealVector dim3 = RealVector(1.0, 2.0, 3.0);
	RealVector::CWise cDim3 = dim3.cwise();
	dim3[2] = 4.0;
	BOOST_CHECK_EQUAL( ((RealVector)cDim3)[2], 3.0 );
}


BOOST_AUTO_TEST_SUITE_END()


