
#define BOOST_TEST_MODULE MathVectorIterator
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Vector.h"


// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MathVectorIterator)


BOOST_AUTO_TEST_CASE(mathVectorIterate) {
	RealVector v1;
	BOOST_CHECK(v1.begin() == v1.end());

	// iterate and read
	RealVector v2(1.0, 2.0, 3.0);
	RealVector v2B(v2);
	RealVector::Iterator it = v2.begin();
	BOOST_CHECK_EQUAL(*it, 1.0);
	++it;
	BOOST_CHECK_EQUAL(*it, 2.0);
	++it;
	BOOST_CHECK_EQUAL(*it, 3.0);
	++it;
	BOOST_CHECK(it == v2.end());

	// iterate and write
	for(RealVector::Iterator it2 = v2.begin(); it2 != v2.end(); ++it2){
		*it2 = (*it2)+1.0;
	}
	BOOST_CHECK_EQUAL(v2, RealVector(2.0,3.0,4.0));
	BOOST_CHECK_EQUAL(v2B, RealVector(1.0,2.0,3.0));

	// iterator assignment ('swap' operation)
	//auto tmp = *v2.begin(); // FIXME with ReadWriteRef, this doesn't work
	double tmp = *v2.begin();
	*v2.begin() = *(v2.begin()+1);
	BOOST_CHECK_EQUAL(v2, RealVector(3.0,3.0,4.0));
	*(v2.begin()+1) = tmp;
	BOOST_CHECK_EQUAL(v2, RealVector(3.0,2.0,4.0));
	

	// read const-iterator
	const RealVector c(1.0, 2.0, 3.0);
	const RealVector::Iterator itC = c.begin();
	BOOST_CHECK_EQUAL(*itC, 1.0);
	++itC;
	BOOST_CHECK_EQUAL(*itC, 2.0);
	++itC;
	BOOST_CHECK_EQUAL(*itC, 3.0);
	++itC;
	BOOST_CHECK(itC == c.end());
}

BOOST_AUTO_TEST_CASE(mathVectorPostfixIterate) {
	RealVector v2(1.0, 2.0, 3.0);
	
	// postfix increment
	RealVector::Iterator it = v2.begin();
	BOOST_CHECK_EQUAL( *(it++), 1.0 );
	BOOST_CHECK_EQUAL( *(it++), 2.0 );
	BOOST_CHECK_EQUAL( *(it++), 3.0 );
	BOOST_CHECK( it == v2.end() );

	// postfix decrement
	it = v2.end()-1;
	BOOST_CHECK_EQUAL( *(it--), 3.0 );
	BOOST_CHECK_EQUAL( *(it--), 2.0 );
	BOOST_CHECK_EQUAL( *(it--), 1.0 );
	BOOST_CHECK( it == v2.begin()-1 );
}

BOOST_AUTO_TEST_CASE(mathVectorStdSwap) {
	RealVector v(1.0, 2.0, 3.0);
	RealVector vB(v);
	// TODO short note: does not work with std::swap, swap needs to be found by ADL inside ReadWriteRef
	swap( *v.begin(), *(v.begin()+1) );
	BOOST_CHECK_EQUAL( v, RealVector(2.0, 1.0, 3.0) );
	BOOST_CHECK_EQUAL( vB, RealVector(1.0, 2.0, 3.0) );
}

BOOST_AUTO_TEST_CASE(mathVectorStlRead) {
	RealVector v(1.0, 2.0, 3.0);
	BOOST_CHECK_EQUAL( std::count(v.begin(), v.end(), 3.0), 1 );
	BOOST_CHECK_EQUAL( *std::min_element(v.begin(), v.end()), 1.0 );
	BOOST_CHECK_EQUAL( *std::max_element(v.begin(), v.end()), 3.0 );

	const RealVector c(1.0, 2.0, 3.0);
	BOOST_CHECK_EQUAL( std::count(c.begin(), c.end(), 3.0), 1 );
	BOOST_CHECK_EQUAL( *std::min_element(c.begin(), c.end()), 1.0 );
	BOOST_CHECK_EQUAL( *std::max_element(c.begin(), c.end()), 3.0 );
}

static double incrX = 0.0;
double incrementor(){
	double val = incrX;
	incrX += 1.0;
	return val;
}

BOOST_AUTO_TEST_CASE(mathVectorStlWrite) {
	RealVector v(1.0, 3.0, 2.0);

	std::reverse( v.begin()+1, v.end());
	BOOST_CHECK_EQUAL( v, RealVector(1.0, 2.0, 3.0) );

	std::reverse( v.begin(), v.end());
	BOOST_CHECK_EQUAL( v, RealVector(3.0, 2.0, 1.0) );
	std::sort( v.begin(), v.end());
	BOOST_CHECK_EQUAL( v, RealVector(1.0, 2.0, 3.0) );

	std::generate( v.begin(), v.end(), incrementor);
	BOOST_CHECK_EQUAL( v, RealVector(0.0, 1.0, 2.0) );
}



BOOST_AUTO_TEST_SUITE_END()


