
#define BOOST_TEST_MODULE MatrixArithmetics
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Matrix.h"

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MatrixArithmetics)


BOOST_AUTO_TEST_CASE(matrixAdd){
	RealMatrix mDim0;
	RealMatrix mDim0Add = mDim0 + mDim0;
	BOOST_CHECK_EQUAL( mDim0Add.dimension(), dim(0,0) );
	BOOST_CHECK_EQUAL( (mDim0 + mDim0).dimension(), dim(0,0) );

	RealMatrix mDim01(dim(0,1));
	RealMatrix mDim01Add = mDim01 - mDim01;
	BOOST_CHECK_EQUAL( mDim01Add.dimension(), dim(0,1) );

	RealMatrix mDim10(dim(1,0));
	RealMatrix mDim10Add = mDim10 - mDim10;
	BOOST_CHECK_EQUAL( mDim10Add.dimension(), dim(1,0) );

	RealMatrix m1(dim(1,2), 1.0);
	RealMatrix m1B(dim(1,2), 1.0);
	RealMatrix m2(dim(1,2), 2.0);
	RealMatrix m2B(dim(1,2), 2.0);
	RealMatrix m3;

	// "copy" construction 
	RealMatrix m4 = m1 + m2;
	BOOST_CHECK_EQUAL( m4.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m4(0,0), 3.0, 1E-6 );
	BOOST_CHECK_CLOSE( (double)m4(0,1), 3.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m3 = m1 + m2;
	BOOST_CHECK_EQUAL( m3.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m3(0,0), 3.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)m3(0,1), 3.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m3 = m1 + m2 + m3;
	BOOST_CHECK_EQUAL( m3.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m3(0,0), 6.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)m3(0,1), 6.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	// check throw on dimension mismatch
	RealMatrix z = RealMatrix() + RealMatrix(); // no throw on zero dimension
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( z + RealMatrix(dim(3,1)), std::invalid_argument );
	BOOST_CHECK_THROW( m1 + RealMatrix(dim(1,4)), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(matrixAddInplace){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0 += mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2), 1.0);
	RealMatrix m1B(m1);
	RealMatrix m2(dim(1,2), 2.0);

	m1 += m1;
	BOOST_CHECK_EQUAL( m1, m2 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1B(0,1), 1.0 ); // unchanged?
	
	m2 += m1;
	BOOST_CHECK_EQUAL( m2, RealMatrix(dim(1,2), 4.0) );
	BOOST_CHECK_EQUAL( m1(0,0), 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 2.0 ); // unchanged?


#ifndef NEMO_ABORT_ON_ERROR
	// check throw on dimension mismatch
	RealMatrix z;
	z += RealMatrix(); // no throw on zero dimension
	BOOST_CHECK_THROW( z += RealMatrix(dim(3,3)), std::invalid_argument );
	BOOST_CHECK_EQUAL( z.dimension(), dim(0,0) ); // unchanged?
	BOOST_CHECK_THROW( m1 += RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1.dimension(), dim(1,2) ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,0), 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 2.0 ); // unchanged?
#endif
}


BOOST_AUTO_TEST_CASE(matrixDiff){
	RealMatrix mDim0;
	RealMatrix mDim0Diff = mDim0 - mDim0;
	BOOST_CHECK_EQUAL( mDim0Diff.dimension(), dim(0,0) );

	RealMatrix mDim01(dim(0,1));
	RealMatrix mDim01Diff = mDim01 - mDim01;
	BOOST_CHECK_EQUAL( mDim01Diff.dimension(), dim(0,1) );

	RealMatrix mDim10(dim(1,0));
	RealMatrix mDim10Diff = mDim10 - mDim10;
	BOOST_CHECK_EQUAL( mDim10Diff.dimension(), dim(1,0) );
	
	// cppcheck-suppress duplicateExpression
	BOOST_CHECK_EQUAL( (mDim0 - mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2), 1.0);
	RealMatrix m1B(dim(1,2), 1.0);
	RealMatrix m2(dim(1,2), 2.0);
	RealMatrix m2B(dim(1,2), 2.0);
	RealMatrix m3;

	// "copy" construction 
	RealMatrix m4 = m1 - m2;
	BOOST_CHECK_EQUAL( m4.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m4(0,0), -1.0, 1E-6 );
	BOOST_CHECK_CLOSE( (double)m4(0,1), -1.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m3 = m1 - m2;
	BOOST_CHECK_EQUAL( m3.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m3(0,0), -1.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)m3(0,1), -1.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m3 = m1 - m2 - m3;
	BOOST_CHECK_EQUAL( m3.dimension(), dim(1,2) );
	BOOST_CHECK_CLOSE( (double)m3(0,0), 0.0, 1E-6);
	BOOST_CHECK_CLOSE( (double)m3(0,1), 0.0, 1E-6 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	// check throw on dimension mismatch
	RealMatrix z = RealMatrix() - RealMatrix(); // no throw on zero dimension
	BOOST_CHECK_THROW( z - RealMatrix(dim(3,1)), std::invalid_argument );
	BOOST_CHECK_THROW( m1 - RealMatrix(dim(1,4)), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(matrixDiffInplace){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0 -= mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2), 2.0);
	RealMatrix m1B(m1);
	RealMatrix m2(dim(1,2), 1.0);
	RealMatrix m3(dim(1,2), 4.0);

	m1 -= m1;
	BOOST_CHECK_EQUAL( m1, RealMatrix(dim(1,2), 0.0) );
	BOOST_CHECK_EQUAL( m1B(0,0), 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1B(0,1), 2.0 ); // unchanged?
	
	m2 -= m3;
	BOOST_CHECK_EQUAL( m2, RealMatrix(dim(1,2), -3.0) );
	BOOST_CHECK_EQUAL( m3(0,0), 4.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m3(0,1), 4.0 ); // unchanged?


#ifndef NEMO_ABORT_ON_ERROR
	// check throw on dimension mismatch
	RealMatrix z;
	z -= RealMatrix(); // no throw on zero dimension
	BOOST_CHECK_THROW( z -= RealMatrix(dim(3,3)), std::invalid_argument );
	BOOST_CHECK_EQUAL( z.dimension(), dim(0,0) ); // unchanged?
	BOOST_CHECK_THROW( m1 -= RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1.dimension(), dim(1,2) ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,0), 0.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 0.0 ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixMult){
	RealMatrix mDim0;
	RealMatrix mDim0Mult = mDim0 * mDim0;
	BOOST_CHECK_EQUAL( mDim0Mult.dimension(), dim(0,0) );
	BOOST_CHECK_EQUAL( (mDim0 * mDim0).dimension(), dim(0,0) );

	RealMatrix mDim01(dim(0,1));
	RealMatrix mDim01Mult = mDim01 * mDim01.transpose();
	BOOST_CHECK_EQUAL( mDim01Mult.dimension(), dim(0,0) );

	RealMatrix mDim10(dim(1,0));
	RealMatrix mDim10Mult = mDim10 * mDim10.transpose();
	BOOST_CHECK_EQUAL( mDim10Mult.dimension(), dim(1,1) );
	BOOST_CHECK_EQUAL( mDim10Mult(0,0), 0.0 );

	RealMatrix mDim01MultDim11 = mDim01 * RealMatrix(dim(1,1));
	BOOST_CHECK_EQUAL( mDim01MultDim11.dimension(), dim(0,1) );

	RealMatrix m1(dim(1,2));
	m1(0,0) = 1.0; m1(0,1) = 2.0;
	RealMatrix m1B(m1);
	RealMatrix m2(dim(2,1));
	m2(0,0) = 3.0; m2(1,0) = 5.0;
	RealMatrix m2B(m2);
	RealMatrix m3;

	RealMatrix expected1 = RealMatrix(dim(1,1), 13.0);
	RealMatrix expected2 = RealMatrix(dim(2,2));
	expected2(0,0) = 3.0;
	expected2(0,1) = 6.0;
	expected2(1,0) = 5.0;
	expected2(1,1) = 10.0;

	// "copy" construction 
	RealMatrix m4 = m1 * m2;
	BOOST_CHECK_EQUAL( m4, expected1 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?
	RealMatrix m5 = m2 * m1;
	BOOST_CHECK_EQUAL( m5, expected2 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m3 = m1 * m2;
	BOOST_CHECK_EQUAL( m3, expected1 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?
	m3 = m2 * m1;
	BOOST_CHECK_EQUAL( m3, expected2 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	// chain multiplication
	m3 = m2 * m1 * m3;
	RealMatrix expected3 = RealMatrix(dim(2,2));
	expected3(0,0) = 39.0;
	expected3(0,1) = 78.0;
	expected3(1,0) = 65.0;
	expected3(1,1) = 130.0;
	BOOST_CHECK_EQUAL( m3, expected3 );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	// self-potentiation
	RealMatrix m6 = RealMatrix(dim(2,2));
	m6(0,0) = 3.0;
	m6(0,1) = 6.0;
	m6(1,0) = 5.0;
	m6(1,1) = 10.0;

	m6 = m6 * m6;
	BOOST_CHECK_EQUAL( m6, expected3 );

#ifndef NEMO_ABORT_ON_ERROR
	// check throw on dimension mismatch
	BOOST_CHECK_THROW( RealMatrix() * RealMatrix(dim(3,1)), std::invalid_argument );
	BOOST_CHECK_THROW( m1 * RealMatrix(dim(1,4)), std::invalid_argument );
	BOOST_CHECK_THROW( m1 * RealMatrix(dim(3,4)), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(matrixMultInplace){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0 *= mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0) = 1.0; m1(0,1) = 2.0;
	RealMatrix m1B(m1);
	RealMatrix m2(dim(2,1));
	m2(0,0) = 3.0; m2(1,0) = 5.0;
	RealMatrix m2B(m2);
	RealMatrix m3;

	RealMatrix expected1 = RealMatrix(dim(1,1), 13.0);
	RealMatrix expected2 = RealMatrix(dim(2,2));
	expected2(0,0) = 3.0;
	expected2(0,1) = 6.0;
	expected2(1,0) = 5.0;
	expected2(1,1) = 10.0;

	m1 *= m2;
	BOOST_CHECK_EQUAL( m1.dimension(), dim(1,1) ); // dimension change!
	BOOST_CHECK_EQUAL( m1, expected1 );
	BOOST_REQUIRE_EQUAL( m1B(0,0), 1.0 ); // unchanged?
	BOOST_REQUIRE_EQUAL( m1B(0,1), 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

	m1 = m1B;
	
	m2 *= m1;
	BOOST_CHECK_EQUAL( m2.dimension(), dim(2,2) );
	BOOST_CHECK_EQUAL( m2, expected2 );
	BOOST_REQUIRE_EQUAL( m2B(0,0), 3.0 ); // unchanged?
	BOOST_REQUIRE_EQUAL( m2B(1,0), 5.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?


#ifndef NEMO_ABORT_ON_ERROR
	// check throw on dimension mismatch
	RealMatrix z;
	BOOST_CHECK_THROW( z *= RealMatrix(dim(3,3)), std::invalid_argument );
	BOOST_CHECK_EQUAL( z.dimension(), dim(0,0) ); // unchanged?
	BOOST_CHECK_THROW( m1 *= RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1.dimension(), dim(1,2) ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,0), 1.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 2.0 ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseMult){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0.cwise() * mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=5.0; m1(0,1) = 7.0;
	RealMatrix m1B(dim(1,2));
	m1B(0,0)=5.0; m1B(0,1) = 7.0;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 2.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 2.0;

	RealMatrix e(dim(1,2));
	e(0,0)=15.0; e(0,1) = 14.0;

	RealMatrix m3 = m1.cwise() * m2;
	BOOST_CHECK_EQUAL( m3, e );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise() * RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseMultInplace){
	RealMatrix mDim0;
	RealMatrix mDim02 = (mDim0.cwise() *= mDim0);
	BOOST_CHECK_EQUAL( mDim02.dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=5.0; m1(0,1) = 7.0;
	RealMatrix m1B = m1;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 2.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 2.0;

	RealMatrix e(dim(1,2));
	e(0,0)=15.0; e(0,1) = 14.0;

	m1.cwise() *= m2;
	BOOST_CHECK_EQUAL( m1, e );
	BOOST_CHECK_EQUAL( m1B(0,0), 5.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1B(0,1), 7.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise() *= RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1(0,0), 15.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 14.0 ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseDiv){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0.cwise() / mDim0).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=6.0; m1(0,1) = 35.0;
	RealMatrix m1B(dim(1,2));
	m1B(0,0)=6.0; m1B(0,1) = 35.0;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 5.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 5.0;

	RealMatrix e(dim(1,2));
	e(0,0)=2.0; e(0,1) = 7.0;

	RealMatrix m3 = m1.cwise() / m2;
	BOOST_CHECK_EQUAL( m3, e );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise() / RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseDivInplace){
	RealMatrix mDim0;
	RealMatrix mDim02 = (mDim0.cwise() /= mDim0);
	BOOST_CHECK_EQUAL( mDim02.dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=6.0; m1(0,1) = 35.0;
	RealMatrix m1B = m1;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 5.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 5.0;

	RealMatrix e(dim(1,2));
	e(0,0)=2.0; e(0,1) = 7.0;

	m1.cwise() /= m2;
	BOOST_CHECK_EQUAL( m1, e );
	BOOST_CHECK_EQUAL( m1B(0,0), 6.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1B(0,1), 35.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise() /= RealMatrix(dim(1,1)), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1(0,0), 2.0 ); // unchanged?
	BOOST_CHECK_EQUAL( m1(0,1), 7.0 ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseMax){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0.cwise().max(mDim0)).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=1.0; m1(0,1) = 4.0;
	RealMatrix m1B(dim(1,2));
	m1B(0,0)=1.0; m1B(0,1) = 4.0;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 2.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 2.0;

	RealMatrix e(dim(1,2));
	e(0,0)=3.0; e(0,1) = 4.0;

	RealMatrix m3 = m1.cwise().max(m2);
	BOOST_CHECK_EQUAL( m3, e );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise().max(RealMatrix(dim(1,1))), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
#endif
}

BOOST_AUTO_TEST_CASE(matrixCWiseMin){
	RealMatrix mDim0;
	BOOST_CHECK_EQUAL( (mDim0.cwise().min(mDim0)).dimension(), dim(0,0) );

	RealMatrix m1(dim(1,2));
	m1(0,0)=1.0; m1(0,1) = 4.0;
	RealMatrix m1B(dim(1,2));
	m1B(0,0)=1.0; m1B(0,1) = 4.0;

	RealMatrix m2(dim(1,2));
	m2(0,0)=3.0; m2(0,1) = 2.0;
	RealMatrix m2B(dim(1,2));
	m2B(0,0)=3.0; m2B(0,1) = 2.0;

	RealMatrix e(dim(1,2));
	e(0,0)=1.0; e(0,1) = 2.0;

	RealMatrix m3 = m1.cwise().min(m2);
	BOOST_CHECK_EQUAL( m3, e );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
	BOOST_CHECK_EQUAL( m2, m2B ); // unchanged?

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1.cwise().min(RealMatrix(dim(1,1))), std::invalid_argument );
	BOOST_CHECK_EQUAL( m1, m1B ); // unchanged?
#endif
}


BOOST_AUTO_TEST_CASE(matrixByVectorOuterProduct){
	const RealVector v0;
	const RealVector v2(1.0,2.0);
	const RealVector v3(1.0,2.0,3.0);

	RealMatrix m(dim(2,3));
	m.setCol(0, RealVector(1.0,2.0));
	m.setCol(1, RealVector(2.0,4.0));
	m.setCol(2, RealVector(3.0,6.0));

	BOOST_CHECK_EQUAL( v0*v0.transpose(), RealMatrix() );
	BOOST_CHECK_EQUAL( v0*v3.transpose(), RealMatrix(dim(0,3)) );
	BOOST_CHECK_EQUAL( v3*v0.transpose(), RealMatrix(dim(3,0)) );
	BOOST_CHECK_EQUAL( v2*v3.transpose(), m );
	BOOST_CHECK_EQUAL( v3*v2.transpose(), m.transpose() );
}

BOOST_AUTO_TEST_SUITE_END()

