
#define BOOST_TEST_MODULE MatrixCreateAccess
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Matrix.h"

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MatrixCreateAccess)


BOOST_AUTO_TEST_CASE(matrixCreate) {
	RealMatrix m1;
	BOOST_CHECK_EQUAL(m1.dimension(), dim(0,0));
	BOOST_CHECK_EQUAL(m1.rows(), 0);
	BOOST_CHECK_EQUAL(m1.cols(), 0);

	RealMatrix m2(dim(3,1));
	BOOST_CHECK_EQUAL(m2.dimension(), dim(3,1));
	BOOST_CHECK_EQUAL(m2.rows(), 3);
	BOOST_CHECK_EQUAL(m2.cols(), 1);
	for(unsigned int i=0; i<3; i++){
		BOOST_CHECK_EQUAL(m2(i,0), 0.0);
	}

	RealMatrix m3(dim(1,3), 13.0);
	BOOST_CHECK_EQUAL(m3.dimension(), dim(1,3));
	BOOST_CHECK_EQUAL(m3(0,0), 13.0);
	BOOST_CHECK_EQUAL(m3(0,1), 13.0);
	BOOST_CHECK_EQUAL(m3(0,2), 13.0);

#ifdef _NEMO_HAS_INIT_LISTS_
	RealMatrix m4z = {};
	BOOST_CHECK_EQUAL( m4z.dimension(), dim(0,0));
	RealMatrix m4 = {{}};
	BOOST_CHECK_EQUAL( m4.dimension(), dim(1,0));
	RealMatrix m5 = {{},{}};
	BOOST_CHECK_EQUAL( m5.dimension(), dim(2,0));
	RealMatrix m6 = {{1.0, 2.0},{3.0, 4.0}};
	BOOST_CHECK_EQUAL( m6.dimension(), dim(2,2));
	BOOST_CHECK_EQUAL( m6.col(0), RealVector(1.0, 3.0));
	BOOST_CHECK_EQUAL( m6.col(1), RealVector(2.0, 4.0));
	RealMatrix m7 = {{1.0, 2.0},{3.0}};
	BOOST_CHECK_EQUAL( m7.dimension(), dim(2,2));
	BOOST_CHECK_EQUAL( m7.col(0), RealVector(1.0, 3.0));
	BOOST_CHECK_EQUAL( m7.col(1), RealVector(2.0, 0.0));
#endif
}
BOOST_AUTO_TEST_CASE(matrixDiagonalScalarCreate) {
	RealMatrix m1 = RealMatrix::diagonal(dim(0,0), 1.0);
	BOOST_CHECK_EQUAL(m1.dimension(), dim(0,0));
	BOOST_CHECK_EQUAL(m1.rows(), 0);
	BOOST_CHECK_EQUAL(m1.cols(), 0);

	RealMatrix m2 = RealMatrix::diagonal(dim(3,1), 1.0);
	BOOST_CHECK_EQUAL(m2.dimension(), dim(3,1));
	BOOST_CHECK_EQUAL(m2.rows(), 3);
	BOOST_CHECK_EQUAL(m2.cols(), 1);
	BOOST_CHECK_EQUAL(m2(0,0), 1.0);
	BOOST_CHECK_EQUAL(m2(1,0), 0.0);
	BOOST_CHECK_EQUAL(m2(2,0), 0.0);

	RealMatrix m3 = RealMatrix::diagonal(dim(1,3), 2.0);
	BOOST_CHECK_EQUAL(m3.dimension(), dim(1,3));
	BOOST_CHECK_EQUAL(m3(0,0), 2.0);
	BOOST_CHECK_EQUAL(m3(0,1), 0.0);
	BOOST_CHECK_EQUAL(m3(0,2), 0.0);

	RealMatrix m4 = RealMatrix::diagonal(dim(2,2), 3.0);
	BOOST_CHECK_EQUAL(m4.dimension(), dim(2,2));
	BOOST_CHECK_EQUAL(m4(0,0), 3.0);
	BOOST_CHECK_EQUAL(m4(0,1), 0.0);
	BOOST_CHECK_EQUAL(m4(1,0), 0.0);
	BOOST_CHECK_EQUAL(m4(1,1), 3.0);
}
BOOST_AUTO_TEST_CASE(matrixDiagonalVectorCreate) {
	RealMatrix m1 = RealMatrix::diagonal(dim(0,0), RealVector());
	BOOST_CHECK_EQUAL(m1.dimension(), dim(0,0));
	BOOST_CHECK_EQUAL(m1.rows(), 0);
	BOOST_CHECK_EQUAL(m1.cols(), 0);

	RealMatrix m2 = RealMatrix::diagonal(dim(3,1), RealVector(1.0));
	BOOST_CHECK_EQUAL(m2.dimension(), dim(3,1));
	BOOST_CHECK_EQUAL(m2.rows(), 3);
	BOOST_CHECK_EQUAL(m2.cols(), 1);
	BOOST_CHECK_EQUAL(m2(0,0), 1.0);
	BOOST_CHECK_EQUAL(m2(1,0), 0.0);
	BOOST_CHECK_EQUAL(m2(2,0), 0.0);

	RealMatrix m3 = RealMatrix::diagonal(dim(1,3), RealVector(2.0));
	BOOST_CHECK_EQUAL(m3.dimension(), dim(1,3));
	BOOST_CHECK_EQUAL(m3(0,0), 2.0);
	BOOST_CHECK_EQUAL(m3(0,1), 0.0);
	BOOST_CHECK_EQUAL(m3(0,2), 0.0);

	RealMatrix m4 = RealMatrix::diagonal(dim(2,2), RealVector(3.0,4.0));
	BOOST_CHECK_EQUAL(m4.dimension(), dim(2,2));
	BOOST_CHECK_EQUAL(m4(0,0), 3.0);
	BOOST_CHECK_EQUAL(m4(0,1), 0.0);
	BOOST_CHECK_EQUAL(m4(1,0), 0.0);
	BOOST_CHECK_EQUAL(m4(1,1), 4.0);

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( RealMatrix::diagonal(dim(3,1), RealVector(dim(2))), std::invalid_argument );
	BOOST_CHECK_THROW( RealMatrix::diagonal(dim(1,3), RealVector(dim(2))), std::invalid_argument );
	BOOST_CHECK_THROW( RealMatrix::diagonal(dim(2,2), RealVector(dim(1))), std::invalid_argument );
#endif
}

BOOST_AUTO_TEST_CASE(matrixPass) {
	RealMatrix m1;
	RealMatrix m2(m1);
	BOOST_CHECK_EQUAL(m2.dimension(), dim(0,0));
	BOOST_CHECK_EQUAL(m1, m2);

	RealMatrix m3(dim(1,1), 1.0);
	RealMatrix m4(dim(1,1), 2.0);
	RealMatrix m5(dim(1,2), 2.0);
	BOOST_REQUIRE( m3 != m4);
	BOOST_REQUIRE( m3 != m5);
	m4 = m3;
	BOOST_CHECK_EQUAL(m3, m4);
	m5 = m3;
	BOOST_CHECK_EQUAL(m5.dimension(), dim(1,1));
	BOOST_CHECK_EQUAL(m3, m5);
}

BOOST_AUTO_TEST_CASE(matrixAccess1) {
	/// read, write for literals

#ifndef NEMO_ABORT_ON_ERROR
	RealMatrix m0d;
	BOOST_CHECK_THROW( m0d(0,0), std::invalid_argument );

	RealMatrix m2d(dim(2,1));
	BOOST_CHECK_THROW( m2d(2,0), std::invalid_argument );
	BOOST_CHECK_THROW( m2d(0,1), std::invalid_argument );
#endif

	// non-const operator()
	RealMatrix m1(dim(1,2), 1.0);
	RealMatrix m1B(m1);
	BOOST_CHECK_EQUAL( m1(0,0), 1.0 );
	BOOST_CHECK_EQUAL( m1(0,1), 1.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );
	BOOST_CHECK_EQUAL( m1B(0,1), 1.0 );

	m1(0,0) = 42.0;
	BOOST_CHECK_EQUAL( m1(0,0), 42.0 );
	BOOST_CHECK_EQUAL( m1(0,1), 1.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );
	BOOST_CHECK_EQUAL( m1B(0,1), 1.0 );

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m1(0,2) = 42.0, std::invalid_argument );
#endif
	BOOST_CHECK_EQUAL( m1(0,0), 42.0 );
	BOOST_CHECK_EQUAL( m1(0,1), 1.0 );

	// const operator()
	const RealMatrix m2(dim(1,2), 2.0);
	const RealMatrix m2B(m2);
	BOOST_CHECK_EQUAL( m2(0,0), 2.0 );
	BOOST_CHECK_EQUAL( m2(0,1), 2.0 );
	BOOST_CHECK_EQUAL( m2B(0,0), 2.0 );
	BOOST_CHECK_EQUAL( m2B(0,1), 2.0 );
}

BOOST_AUTO_TEST_CASE(matrixAccess2) {
	/// read, write for ReadWriteRef

	// non-const operator()
	RealMatrix m1(dim(1,2), 1.0);
	RealMatrix m1B(m1);
	RealMatrix m2(dim(1,2), 2.0);
	RealMatrix m2B(m2);

	m1(0,0) = m2(0,0);
	BOOST_CHECK_EQUAL( m1(0,0), m2(0,0) );
	BOOST_CHECK_EQUAL( m1(0,1), 1.0 );
	BOOST_CHECK_EQUAL( m2, m2B );

	// const operator()
	const RealMatrix c1(dim(1,2), 3.0);
	const RealMatrix c1B(c1);

	m1(0,0) = c1(0,0);
	BOOST_CHECK_EQUAL( m1(0,0), c1(0,0) );
	BOOST_CHECK_EQUAL( m1(0,1), 1.0 );
	BOOST_CHECK_EQUAL( c1, c1B );

	// use non-const() inside something else
	RealMatrix z(dim(2,1), 0.0);
	double a = std::sin( z(0,0) );
	BOOST_CHECK_EQUAL( a, 0.0 );

	double b = 2*m1(0,1) + m2(0,1);
	BOOST_CHECK_EQUAL( b, 4.0 );
}


BOOST_AUTO_TEST_CASE(matrixAccessCompound){
	RealMatrix m1(dim(2,1), 1.0);
	RealMatrix m1B(m1);
	RealMatrix m2(dim(2,1), 2.0);
	RealMatrix m2B(m2);

	m1(0,0) += 1.0;
	BOOST_CHECK_EQUAL( m1(0,0), 2.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );

	m1(0,0) *= 2.0;
	BOOST_CHECK_EQUAL( m1(0,0), 4.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );

	m1(0,0) /= 4.0;
	BOOST_CHECK_EQUAL( m1(0,0), 1.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );

	m1(0,0) -= 1.0;
	BOOST_CHECK_EQUAL( m1(0,0), 0.0 );
	BOOST_CHECK_EQUAL( m1B(0,0), 1.0 );

//	Matrix<int> i1(dim(2,1), 1);
//	Matrix<int> i1B(i1);
}

BOOST_AUTO_TEST_CASE(matrixAccessSwap){
	RealMatrix m1(dim(1,2));
	m1.setRow( 0, RealVector(1.0, 2.0).transpose() );
	RealMatrix m2 = m1;

	using std::swap;
	swap( m1(0,0), m1(0,1) );
	BOOST_CHECK_EQUAL( m1.row(0), RealVector(2.0, 1.0).transpose() );
	BOOST_CHECK_EQUAL( m2.row(0), RealVector(1.0, 2.0).transpose() );
}

BOOST_AUTO_TEST_CASE(mathVectorAccessMax){
	RealMatrix m(dim(1,2));
	m.setRow( 0, RealVector(1.0, 2.0).transpose() );

	using std::max;
	BOOST_CHECK_EQUAL( max(m(0,0), m(0,1)), 2.0 );
	BOOST_CHECK_EQUAL( max(m(0,0), 2.0), 2.0 );
	BOOST_CHECK_EQUAL( max(1.0, m(0,1)), 2.0 );
}
BOOST_AUTO_TEST_CASE(mathVectorAccessMin){
	RealMatrix m(dim(1,2));
	m.setRow( 0, RealVector(1.0, 2.0).transpose() );

	using std::min;
	BOOST_CHECK_EQUAL( min(m(0,0), m(0,1)), 1.0 );
	BOOST_CHECK_EQUAL( min(m(0,0), 2.0), 1.0 );
	BOOST_CHECK_EQUAL( min(1.0, m(0,0)), 1.0 );
}

BOOST_AUTO_TEST_CASE(matrixColAccess){
	RealMatrix m(dim(2,2));
	m(0,0) = 1.0;
	m(0,1) = 2.0;
	m(1,0) = 3.0;
	m(1,1) = 4.0;

	BOOST_CHECK_EQUAL( m.col(0), RealVector(1.0, 3.0) );
	BOOST_CHECK_EQUAL( m.col(1), RealVector(2.0, 4.0) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m.col(2), std::invalid_argument );
	BOOST_CHECK_THROW( m.setCol(2, RealVector(dim(2))), std::invalid_argument );
#endif
	m.setCol(0, RealVector(5.0, 6.0));
	BOOST_CHECK_EQUAL( m.col(0), RealVector(5.0, 6.0) );
	m.setCol(1, RealVector(7.0, 8.0));
	BOOST_CHECK_EQUAL( m.col(1), RealVector(7.0, 8.0) );
}

BOOST_AUTO_TEST_CASE(matrixRowAccess){
	RealMatrix m(dim(2,2));
	m(0,0) = 1.0;
	m(0,1) = 2.0;
	m(1,0) = 3.0;
	m(1,1) = 4.0;

	BOOST_CHECK_EQUAL( m.row(0).transpose(), RealVector(1.0, 2.0) );
	BOOST_CHECK_EQUAL( m.row(1).transpose(), RealVector(3.0, 4.0) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m.row(2), std::invalid_argument );
	BOOST_CHECK_THROW( m.setRow(2, RealVector(dim(2)).transpose()), std::invalid_argument );
#endif
	m.setRow(0, RealVector(5.0, 6.0).transpose());
	BOOST_CHECK_EQUAL( m.row(0).transpose(), RealVector(5.0, 6.0) );
	m.setRow(1, RealVector(7.0, 8.0).transpose());
	BOOST_CHECK_EQUAL( m.row(1).transpose(), RealVector(7.0, 8.0) );
}

BOOST_AUTO_TEST_CASE(matrixSubMatAccess){
	RealMatrix m(dim(2,2));
	m(0,0) = 1.0;
	m(0,1) = 2.0;
	m(1,0) = 3.0;
	m(1,1) = 4.0;
	RealMatrix mB(dim(2,2));
	mB(0,0) = 1.0;
	mB(0,1) = 2.0;
	mB(1,0) = 3.0;
	mB(1,1) = 4.0;

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( m.setSubMatrix(0,0,RealMatrix(dim(3,1))), std::invalid_argument );
	BOOST_CHECK_THROW( m.setSubMatrix(0,0,RealMatrix(dim(1,3))), std::invalid_argument );
	BOOST_CHECK_THROW( m.setSubMatrix(2,0,RealMatrix(dim(1,1))), std::invalid_argument );
	BOOST_CHECK_THROW( m.setSubMatrix(0,2,RealMatrix(dim(3,1))), std::invalid_argument );
#endif

	m.setSubMatrix(0,0,RealMatrix(dim(0,0)));
	BOOST_CHECK_EQUAL(m, mB);
	m.setSubMatrix(0,0,RealMatrix(dim(2,0)));
	BOOST_CHECK_EQUAL(m, mB);
	m.setSubMatrix(0,0,RealMatrix(dim(0,2)));
	BOOST_CHECK_EQUAL(m, mB);
	m.setSubMatrix(2,2,RealMatrix(dim(0,0)));
	BOOST_CHECK_EQUAL(m, mB);

	m.setSubMatrix(0, 0, RealMatrix(dim(2,1), 5.0));
	BOOST_CHECK_EQUAL( m.col(0), RealVector(5.0, 5.0) );
	m.setSubMatrix(0, 1, RealMatrix(dim(2,1), 6.0));
	BOOST_CHECK_EQUAL( m.col(1), RealVector(6.0, 6.0) );

	m.setSubMatrix(0, 0, RealMatrix(dim(2,2), 7.0));
	BOOST_CHECK_EQUAL( m, RealMatrix(dim(2,2), 7.0) );

}


BOOST_AUTO_TEST_SUITE_END()

