
#define BOOST_TEST_MODULE MatrixIterator
#include <boost/test/unit_test.hpp>

#include "../src/nemo/Matrix.h"


// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


BOOST_AUTO_TEST_SUITE(MatrixIterator)


BOOST_AUTO_TEST_CASE(matrixIterate) {
	RealMatrix m1;
	BOOST_CHECK(m1.begin() == m1.end());

	// iterate and read
	RealMatrix m2(dim(2,2));
	m2.setRow(0, RealVector(1.0, 2.0).transpose());
	m2.setRow(1, RealVector(3.0, 4.0).transpose());
	RealMatrix m2B(m2);
	RealMatrix::Iterator it = m2.begin();
	BOOST_CHECK_EQUAL(*it, 1.0);
	++it;
	BOOST_CHECK_EQUAL(*it, 2.0);
	++it;
	BOOST_CHECK_EQUAL(*it, 3.0);
	++it;
	BOOST_CHECK_EQUAL(*it, 4.0);
	++it;
	BOOST_CHECK(it == m2.end());

	// iterate and write
	for(RealMatrix::Iterator it2 = m2.begin(); it2 != m2.end(); ++it2){
		*it2 = (*it2)+1.0;
	}
	BOOST_CHECK_EQUAL(m2.row(0), RealVector(2.0,3.0).transpose());
	BOOST_CHECK_EQUAL(m2.row(1), RealVector(4.0,5.0).transpose());
	BOOST_CHECK_EQUAL(m2B.row(0), RealVector(1.0,2.0).transpose());
	BOOST_CHECK_EQUAL(m2B.row(1), RealVector(3.0,4.0).transpose());

	// iterator assignment ('swap' operation)
	double tmp = *m2.begin();
	*m2.begin() = *(m2.begin()+1);
	BOOST_CHECK_EQUAL(m2.row(0), RealVector(3.0,3.0).transpose());
	BOOST_CHECK_EQUAL(m2.row(1), RealVector(4.0,5.0).transpose());
	*(m2.begin()+1) = tmp;
	BOOST_CHECK_EQUAL(m2.row(0), RealVector(3.0,2.0).transpose());
	BOOST_CHECK_EQUAL(m2.row(1), RealVector(4.0,5.0).transpose());
	

	// read const-iterator
	RealMatrix cVal(dim(1,3));
	cVal.setRow(0, RealVector(1.0,2.0,3.0).transpose());
	const RealMatrix c = cVal;
	const RealMatrix::Iterator itC = c.begin();
	BOOST_CHECK_EQUAL(*itC, 1.0);
	++itC;
	BOOST_CHECK_EQUAL(*itC, 2.0);
	++itC;
	BOOST_CHECK_EQUAL(*itC, 3.0);
	++itC;
	BOOST_CHECK(itC == c.end());
}

BOOST_AUTO_TEST_CASE(matrixPostfixIterate) {
	RealMatrix m2(dim(2,2));
	m2.setRow(0, RealVector(1.0, 2.0).transpose());
	m2.setRow(1, RealVector(3.0, 4.0).transpose());
	
	// postfix increment
	RealMatrix::Iterator it = m2.begin();
	BOOST_CHECK_EQUAL( *(it++), 1.0 );
	BOOST_CHECK_EQUAL( *(it++), 2.0 );
	BOOST_CHECK_EQUAL( *(it++), 3.0 );
	BOOST_CHECK_EQUAL( *(it++), 4.0 );
	BOOST_CHECK( it == m2.end() );

	// postfix decrement
	it = m2.end()-1;
	BOOST_CHECK_EQUAL( *(it--), 4.0 );
	BOOST_CHECK_EQUAL( *(it--), 3.0 );
	BOOST_CHECK_EQUAL( *(it--), 2.0 );
	BOOST_CHECK_EQUAL( *(it--), 1.0 );
	BOOST_CHECK( it == m2.begin()-1 );
}

BOOST_AUTO_TEST_CASE(matrixStdSwap) {
	RealMatrix m(dim(1,3));
	m.setRow(0, RealVector(1.0,2.0,3.0).transpose());
	RealMatrix mB = m;
	swap( *m.begin(), *(m.begin()+1) );
	BOOST_CHECK_EQUAL( m.row(0), RealVector(2.0, 1.0, 3.0).transpose() );
	BOOST_CHECK_EQUAL( mB.row(0), RealVector(1.0, 2.0, 3.0).transpose() );
}

BOOST_AUTO_TEST_CASE(matrixStlRead) {
	RealMatrix m(dim(1,3));
	m.setRow(0, RealVector(1.0,2.0,3.0).transpose());
	BOOST_CHECK_EQUAL( std::count(m.begin(), m.end(), 3.0), 1 );
	BOOST_CHECK_EQUAL( *std::min_element(m.begin(), m.end()), 1.0 );
	BOOST_CHECK_EQUAL( *std::max_element(m.begin(), m.end()), 3.0 );

	const RealMatrix c = m;
	BOOST_CHECK_EQUAL( std::count(c.begin(), c.end(), 3.0), 1 );
	BOOST_CHECK_EQUAL( *std::min_element(c.begin(), c.end()), 1.0 );
	BOOST_CHECK_EQUAL( *std::max_element(c.begin(), c.end()), 3.0 );
}

static double incrX = 0.0;
double incrementor(){
	double val = incrX;
	incrX += 1.0;
	return val;
}

BOOST_AUTO_TEST_CASE(matrixStlWrite) {
	RealMatrix m(dim(1,3));
	m.setRow(0, RealVector(1.0,3.0,2.0).transpose());

	std::reverse( m.begin()+1, m.end());
	BOOST_CHECK_EQUAL( m.row(0), RealVector(1.0, 2.0, 3.0).transpose() );

	std::reverse( m.begin(), m.end());
	BOOST_CHECK_EQUAL( m.row(0), RealVector(3.0, 2.0, 1.0).transpose() );
	std::sort( m.begin(), m.end());
	BOOST_CHECK_EQUAL( m.row(0), RealVector(1.0, 2.0, 3.0).transpose() );

	std::generate( m.begin(), m.end(), incrementor);
	BOOST_CHECK_EQUAL( m.row(0), RealVector(0.0, 1.0, 2.0).transpose() );
}



BOOST_AUTO_TEST_SUITE_END()


