
#define BOOST_TEST_MODULE MatrixMath
#include <boost/test/unit_test.hpp>

#include "../src/nemo/MatrixMath.h"
#include "../src/nemo/Norms.h"

#include <iostream>
#include <limits>

using namespace nemo;
//using ::nemo::MatrixMath;
//using ::nemo::MatrixMath::pseudoInverse;
using namespace boost::unit_test;

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"


BOOST_AUTO_TEST_SUITE(MatrixMath)


BOOST_AUTO_TEST_CASE(matrixPseudoInverse){
	// zero dim matrices
	RealMatrix mDim0;
	RealMatrix mDim0I = pseudoInverse(mDim0);
	BOOST_CHECK_EQUAL(mDim0, mDim0I);

	RealMatrix mDim01(dim(0,1));
	RealMatrix mDim10(dim(1,0));
	BOOST_CHECK_EQUAL(pseudoInverse(mDim01), mDim10);
	BOOST_CHECK_EQUAL(mDim01, pseudoInverse(mDim10));

	RealMatrix mId = RealMatrix::diagonal(dim(3,3), 1.0);
	BOOST_CHECK_EQUAL(pseudoInverse(mId), mId);

	RealMatrix mZero(dim(2,2));
	BOOST_CHECK_EQUAL(pseudoInverse(mZero), mZero);

	RealMatrix m1(dim(2,2));
	m1(0,0) = 2;
	m1(0,1) = 1;
	m1(1,0) = 1;
	m1(1,1) = 0;
	RealMatrix m1I(dim(2,2));
	m1I(0,0) = 0;
	m1I(0,1) = 1;
	m1I(1,0) = 1;
	m1I(1,1) = -2;
	const RealMatrix m1Iest = pseudoInverse(m1);
	BOOST_CHECK_SMALL((double)m1Iest(0,0)-(double)m1I(0,0), 1E-9);
	BOOST_CHECK_SMALL((double)m1Iest(0,1)-(double)m1I(0,1), 1E-9);
	BOOST_CHECK_SMALL((double)m1Iest(1,0)-(double)m1I(1,0), 1E-9);
	BOOST_CHECK_SMALL((double)m1Iest(1,1)-(double)m1I(1,1), 1E-9);

	RealMatrix m2(dim(2,1));
	m2(0,0) = 2;
	m2(1,0) = 1;
	RealMatrix m2I(dim(1,2));
	m2I(0,0) = 0.4;
	m2I(0,1) = 0.2;
	const RealMatrix m2Iest = pseudoInverse(m2);
	BOOST_CHECK_SMALL((double)m2Iest(0,0)-(double)m2I(0,0), 1E-9);
	BOOST_CHECK_SMALL((double)m2Iest(0,1)-(double)m2I(0,1), 1E-9);

	
	RealMatrix m23(dim(2,3), 0.0);
	m23.setCol(0, RealVector(1.0, 4.0));
	m23.setCol(1, RealVector(2.0, 5.0));
	m23.setCol(2, RealVector(3.0, 6.0));
	RealMatrix m23inv = pseudoInverse(m23);
	Mapping<RealMatrix,double> norm = euclideanNorm<RealMatrix,double>();
	BOOST_CHECK_SMALL( norm(m23*m23inv - RealMatrix::identity(2)) , 1E-9);

#ifndef NEMO_ABORT_ON_ERROR
	RealMatrix m22nan( dim(2,2) );
	m22nan.setCol(0, RealVector(2.0, 0.2));
	m22nan.setCol(1, RealVector(std::numeric_limits<double>::quiet_NaN(), 0.5)); 
	
	BOOST_CHECK_THROW( pseudoInverse(m22nan), std::invalid_argument )
#endif
}

class incr_gen : public Generator<int> {
public:
	incr_gen() : Generator<int>(), num(0) {}

protected:
	int generate(){
		return num++;
	}
private:
	int num;
};

BOOST_AUTO_TEST_CASE(matrixMatMatT){
	RealMatrix mDim0;
	RealMatrix m1_3(dim(1,3));
	m1_3.setRow(0, RealVector(1.0, 2.0, 3.0).transpose());
	RealMatrix m3_1(dim(3,1));
	m3_1.setCol(0, RealVector(1.0, 2.0, 3.0));

	BOOST_CHECK_EQUAL( matMatT(mDim0), mDim0 );
	BOOST_CHECK_EQUAL( matMatT(m1_3), m1_3 * m1_3.transpose() );
	BOOST_CHECK_EQUAL( matMatT(m3_1), m3_1 * m3_1.transpose() );

	Mapping<RealMatrix,double> norm = euclideanNorm<RealMatrix,double>();
	for(unsigned int rows = 0; rows <= 8; rows++){
		for(unsigned int cols = 0; cols <= 8; cols++){
			// test matrix with successive integral values
			RealMatrix m(dim(rows,cols), incr_gen());
			BOOST_CHECK_SMALL( norm( matMatT(m) - m*m.transpose() ), 1e-6 );
		}
	}
}

BOOST_AUTO_TEST_CASE(matrixMatTMat){
	RealMatrix mDim0;
	RealMatrix m1_3(dim(1,3));
	m1_3.setRow(0, RealVector(1.0, 2.0, 3.0).transpose());
	RealMatrix m3_1(dim(3,1));
	m3_1.setCol(0, RealVector(1.0, 2.0, 3.0));

	BOOST_CHECK_EQUAL( matTMat(mDim0), mDim0 );
	BOOST_CHECK_EQUAL( matTMat(m1_3), m1_3.transpose() * m1_3 );
	BOOST_CHECK_EQUAL( matTMat(m3_1), m3_1.transpose() * m3_1 );

	Mapping<RealMatrix,double> norm = euclideanNorm<RealMatrix,double>();
	for(unsigned int rows = 0; rows <= 8; rows++){
		for(unsigned int cols = 0; cols <= 8; cols++){
			// test matrix with successive integral values
			RealMatrix m(dim(rows,cols), incr_gen());
			BOOST_CHECK_SMALL( norm( matTMat(m) - m.transpose()*m ), 1e-6 );
		}
	}
}

void checkSVD(const RealMatrix &mat, const SVD<double> &svd){
	Mapping<RealMatrix,double> norm = euclideanNorm<RealMatrix,double>();

//std::cout << "M:   " << (mat.rows()) << "x" << (mat.cols()) << std::endl << mat << std::endl;
//std::cout << "U:   " << (svd.matrixU().rows()) << "x" << (svd.matrixU().cols()) << std::endl << svd.matrixU() << std::endl;
//std::cout << "S:   " << (svd.matrixS().rows()) << "x" << (svd.matrixS().cols()) << std::endl << svd.matrixS() << std::endl;
//std::cout << "V^T: " << (svd.matrixV().cols()) << "x" << (svd.matrixV().rows()) << std::endl << svd.matrixV().transpose() << std::endl;

	unsigned int n = mat.rows();
	unsigned int m = mat.cols();
	unsigned int l = (m<n)?(m):(n);

	// dimensions
	BOOST_CHECK_EQUAL( svd.singularValues().dimension(), l );
	BOOST_CHECK_EQUAL( svd.matrixS().dimension(), dim(l,l) );
	BOOST_CHECK_EQUAL( svd.matrixU().dimension(), dim(n,l) );
	BOOST_CHECK_EQUAL( svd.matrixV().dimension(), dim(m,l) );

	// orthonormal U:
	if(n==l){
		BOOST_CHECK_SMALL( norm( svd.matrixU() * svd.matrixU().transpose() - RealMatrix::identity(n) ), 1e-6 );
	}
	BOOST_CHECK_SMALL( norm( svd.matrixU().transpose() * svd.matrixU() - RealMatrix::identity(l) ), 1e-6 );
	// orthonormal V:
	if(m==l){
		BOOST_CHECK_SMALL( norm( svd.matrixV() * svd.matrixV().transpose() - RealMatrix::identity(m) ), 1e-6 );
	}
	BOOST_CHECK_SMALL( norm( svd.matrixV().transpose() * svd.matrixV() - RealMatrix::identity(l) ), 1e-6 );
	
	// Reconstruction:
	BOOST_CHECK_SMALL( norm( svd.matrixU() * svd.matrixS() * svd.matrixV().transpose() - mat ), 1e-6 );
}

BOOST_AUTO_TEST_CASE(matrixSVD){

	Matrix<double> m00;
	SVD<double> svd00(m00);
	checkSVD(m00, svd00);

	Matrix<double> m50(dim(5,0));
	SVD<double> svd50(m50);
	checkSVD(m50, svd50);

	Matrix<double> m05(dim(0,5));
	SVD<double> svd05(m05);
	checkSVD(m05, svd05);

	Matrix<double> m11(dim(1,1), 2.0);
	SVD<double> svd11(m11);
	checkSVD(m11, svd11);

	Matrix<double> m13(dim(1,3), 0.0);
	m13(0,1) = 2;
	SVD<double> svd13(m13);
	checkSVD(m13, svd13);

	Matrix<double> m31(dim(3,1), 0.0);
	m31(0,0) = 2;
	SVD<double> svd31(m31);
	checkSVD(m31, svd31);

	Matrix<double> m23(dim(2,3), 0.0);
	m23.setCol(0, RealVector(1.0, 4.0));
	m23.setCol(1, RealVector(2.0, 5.0));
	m23.setCol(2, RealVector(3.0, 6.0));
	SVD<double> svd23(m23);
	checkSVD(m23, svd23);

	Matrix<double> m32 = m23.transpose();
	SVD<double> svd32(m32);
	checkSVD(m32, svd32);

#ifndef NEMO_ABORT_ON_ERROR
	RealMatrix m22inf( dim(2,2) );
	m22inf.setCol(0, RealVector(2.0, 0.2));
	m22inf.setCol(1, RealVector(std::numeric_limits<double>::infinity(), 0.5)); 
	
	BOOST_CHECK_THROW( (SVD<double>(m22inf)), std::invalid_argument )

	RealMatrix m22nan( dim(2,2) );
	m22nan.setCol(0, RealVector(2.0, 0.2));
	m22nan.setCol(1, RealVector(std::numeric_limits<double>::quiet_NaN(), 0.5)); 
	
	BOOST_CHECK_THROW( (SVD<double>(m22nan)), std::invalid_argument )
#endif
}

void checkSED(const RealMatrix &mat, const SymmetricEigenDecomposition<double> &sed){
	Mapping<RealMatrix,double> normM = euclideanNorm<RealMatrix,double>();
	Mapping<RealVector,double> normV = euclideanNorm<RealVector,double>();
	
	// dimensions
	BOOST_REQUIRE_EQUAL( mat.rows(), mat.cols() );
	const unsigned int N = mat.rows();
	BOOST_CHECK_EQUAL( sed.eigenValues().dimension(), N );
	BOOST_CHECK_EQUAL( sed.matrixD().dimension(), dim(N,N) );
	BOOST_CHECK_EQUAL( sed.matrixQ().dimension(), dim(N,N) );
	
	// orthonormality
	BOOST_CHECK_SMALL( normM( sed.matrixQ() * sed.matrixQ().transpose() - RealMatrix::identity(N) ), 1e-6 );
	BOOST_CHECK_SMALL( normM( sed.matrixQ().transpose() * sed.matrixQ() - RealMatrix::identity(N) ), 1e-6 );

	// Eigen value/vector relation
	for(unsigned int i=0; i<N; i++){
		RealVector left = mat * sed.matrixQ().col(i);
		RealVector right = sed.eigenValues()[i] * sed.matrixQ().col(i);
		BOOST_CHECK_SMALL( normV( left-right ) , 1e-6 );
	}

	// decomposition
	BOOST_CHECK_SMALL( normM( sed.matrixQ() * sed.matrixD() * sed.matrixQ().transpose() - mat ), 1e-6 );
}

BOOST_AUTO_TEST_CASE(matrixEigenDec){
	RealMatrix m00;
	SymmetricEigenDecomposition<double> sed00(m00);
	checkSED(m00, sed00);

	RealMatrix m11(dim(1,1), 2.0);
	SymmetricEigenDecomposition<double> sed11(m11);
	checkSED(m11, sed11);
	BOOST_CHECK_EQUAL( sed11.eigenValues()[0], 2 );

	RealMatrix m22a = RealMatrix::diagonal( dim(2,2), RealVector(2.0,1.0) );
	SymmetricEigenDecomposition<double> sed22a(m22a);
	checkSED(m22a, sed22a);
	BOOST_CHECK( sed22a.eigenValues()[0]==2 || sed22a.eigenValues()[1]==2 );
	BOOST_CHECK( sed22a.eigenValues()[0]==1 || sed22a.eigenValues()[1]==1 );

	RealMatrix m22b = RealMatrix::diagonal( dim(2,2), RealVector(1.0, 0.0));
	SymmetricEigenDecomposition<double> sed22b(m22b);
	checkSED(m22b, sed22b);
	BOOST_CHECK( sed22b.eigenValues()[0]==0 || sed22b.eigenValues()[1]==0 );
	BOOST_CHECK( sed22b.eigenValues()[0]==1 || sed22b.eigenValues()[1]==1 );

	RealMatrix m22c( dim(2,2) );
	m22c.setCol(0, RealVector(2.0, 0.2));
	m22c.setCol(1, RealVector(0.2, 0.5));
	SymmetricEigenDecomposition<double> sed22c(m22c);
	checkSED(m22c, sed22c);

#ifndef NEMO_ABORT_ON_ERROR
	RealMatrix m22d( dim(2,2) );
	m22d.setCol(0, RealVector(2.0, 0.2));
	m22d.setCol(1, RealVector(0.1, 0.5)); // non-symmetric
	
	BOOST_CHECK_THROW( (SymmetricEigenDecomposition<double>(m22d)), std::invalid_argument )
#endif
}


BOOST_AUTO_TEST_SUITE_END()

