
#define BOOST_TEST_MODULE TimeSeriesCollectionAPI
#include <boost/test/unit_test.hpp>

#include "../src/nemo/TimeSeries.h"
#include "../src/nemo/Vector.h"

#include <cmath>

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

using namespace nemo;
using namespace boost::unit_test;


typedef TimeSeries<int,double> RealSeries;
typedef TimeSeries<int,RealVector> VectorSeries;

BOOST_AUTO_TEST_SUITE(TimeSeriesCollectionAPI)



BOOST_AUTO_TEST_CASE(timestampValueAccess) {
	RealSeries s1;
	BOOST_CHECK_EQUAL(s1.timestamps.numElements(), 0);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s1.timestamps.getElement(0), std::invalid_argument);
	BOOST_CHECK_THROW(s1.timestamps.setElement(0, 1.0), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(s1.values.numElements(), 0);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s1.values.getElement(0), std::invalid_argument);
	BOOST_CHECK_THROW(s1.values.setElement(0, 1.0), std::invalid_argument);
#endif

	RealSeries s2;
	s2.append(0, 1.0);
	s2.append(1, 2.0);
	s2.append(2, 3.0);	
	BOOST_CHECK_EQUAL(s2.timestamps.numElements(), 3);
	BOOST_CHECK_EQUAL(s2.timestamps.getElement(0), 0);
	BOOST_CHECK_EQUAL(s2.timestamps.getElement(1), 1);
	BOOST_CHECK_EQUAL(s2.timestamps.getElement(2), 2);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s2.timestamps.getElement(3), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(s2.values.numElements(), 3);
	BOOST_CHECK_EQUAL(s2.values.getElement(0), 1.0);
	BOOST_CHECK_EQUAL(s2.values.getElement(1), 2.0);
	BOOST_CHECK_EQUAL(s2.values.getElement(2), 3.0);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s2.values.getElement(3), std::invalid_argument);
#endif

	BOOST_CHECK_NO_THROW(s2.values.setElement(0, 4.0));
	BOOST_CHECK_NO_THROW(s2.timestamps.setElement(0, -1));
	BOOST_CHECK_NO_THROW(s2.values.setElement(1, 5.0));
	BOOST_CHECK_NO_THROW(s2.timestamps.setElement(1, 0));

	BOOST_CHECK_EQUAL( s2.length(), 3 );
	BOOST_CHECK_EQUAL( s2[-1], 4.0 );
	BOOST_CHECK_EQUAL( s2[0], 5.0 );
	BOOST_CHECK_EQUAL( s2[2], 3.0 );
}

BOOST_AUTO_TEST_CASE(valueTransform) {
	RealSeries r1;
	r1.append(0, 1.0);
	r1.append(1, 2.0);
	r1.append(2, 3.0);
	Mapping<double,double> rMap = 2.0 * arg<double>();
	RealSeries r2;
	r2.append(0, 2.0);
	r2.append(1, 4.0);
	r2.append(2, 6.0);

	BOOST_CHECK_EQUAL( r1.values.map(rMap), r2 );

	r1.values.transform(rMap);
	BOOST_CHECK_EQUAL( r1, r2 );
}

BOOST_AUTO_TEST_CASE(timestampTransform) {
	RealSeries r1;
	r1.append(0, 1.0);
	r1.append(1, 2.0);
	r1.append(2, 3.0);

	Mapping<int,int> shift = arg<int>() + 1;
	RealSeries r1shift;
	r1shift.append(1, 1.0);
	r1shift.append(2, 2.0);
	r1shift.append(3, 3.0);

	Mapping<int,int> reverse = 2 - arg<int>();
	RealSeries r1reverse;
	r1reverse.insert(2, 1.0);
	r1reverse.insert(1, 2.0);
	r1reverse.insert(0, 3.0);

	BOOST_CHECK_EQUAL( r1.timestamps.map(shift), r1shift );

	BOOST_CHECK_EQUAL( r1.timestamps.map(reverse), r1reverse );

	r1.timestamps.transform(reverse);
	BOOST_CHECK_EQUAL( r1, r1reverse );
}

BOOST_AUTO_TEST_CASE(entryAccess) {
	typedef TimeSeriesEntry<double,double> RealEntry;

	TimeSeries<double,double> r1;
	r1.append(0.0, 1.0);
	r1.append(1.0, 2.0);
	r1.append(2.0, 3.0);

	BOOST_CHECK_EQUAL( r1.numElements(), 3 );
	BOOST_CHECK_EQUAL( r1.getElement(0), RealEntry(0.0, 1.0) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( r1.getElement(3), std::invalid_argument );
#endif

	BOOST_CHECK_NO_THROW( r1.setElement(2, RealEntry(2.1, 3.1)) );
	BOOST_CHECK_EQUAL( r1.getElement(2), RealEntry(2.1, 3.1) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( r1.setElement(3, RealEntry(0, 0)), std::invalid_argument );
#endif

	// order changing 'set'
	BOOST_CHECK_NO_THROW( r1.setElement(1, RealEntry(3.0, 2.1)) );

	BOOST_CHECK_EQUAL( r1.getElement(0), RealEntry(0.0, 1.0) );
	BOOST_CHECK_EQUAL( r1.getElement(1), RealEntry(2.1, 3.1) );
	BOOST_CHECK_EQUAL( r1.getElement(2), RealEntry(3.0, 2.1) );
}


TimeSeriesEntry<double,double> realEntryFunc(TimeSeriesEntry<double,double> entry, double factor){
	return TimeSeriesEntry<double,double>( factor*entry.timestamp-1.0, entry.value*2 );
}

BOOST_AUTO_TEST_CASE(entryTransformMap) {
	typedef TimeSeriesEntry<double,double> RealEntry;
	typedef Mapping<RealEntry,RealEntry> RealEntryMapping;

	TimeSeries<double,double> r1;
	r1.append(0.0, 1.0);
	r1.append(1.0, 2.0);
	r1.append(2.0, 3.0);
	TimeSeries<double,double> r1B = r1;
	TimeSeries<double,double> r2;
	r2.append(-1.0, 2.0);
	r2.append(0.0, 4.0);
	r2.append(1.0, 6.0);
	TimeSeries<double,double> r3;
	r3.append(-3.0, 6.0);
	r3.append(-2.0, 4.0);
	r3.append(-1.0, 2.0);

	r1.transform( RealEntryMapping(boost::bind( realEntryFunc, _1, 1.0 )) );
	BOOST_CHECK_EQUAL( r1, r2 );
	BOOST_CHECK_EQUAL( r1B.getElement(0), RealEntry(0.0, 1.0) );
	r1 = r1B;

	r1.transform( RealEntryMapping(boost::bind( realEntryFunc, _1, -1.0 )) );
	BOOST_CHECK_EQUAL( r1, r3 );
	BOOST_CHECK_EQUAL( r1B.getElement(0), RealEntry(0.0, 1.0) );
	r1 = r1B;

	TimeSeries<double,double> r4 = r1.map( RealEntryMapping(boost::bind( realEntryFunc, _1, 1.0 )) );
	BOOST_CHECK_EQUAL( r4, r2 );
	BOOST_CHECK_EQUAL( r1B.getElement(0), RealEntry(0.0, 1.0) );
}



BOOST_AUTO_TEST_SUITE_END()


