
#define BOOST_TEST_MODULE TimeSeriesCreateAccess
#include <boost/test/unit_test.hpp>

#include "../src/nemo/TimeSeries.h"

#include <iostream>

using namespace nemo;
using namespace boost::unit_test;

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"


BOOST_AUTO_TEST_SUITE(TimeSeriesCreateAccess)


BOOST_AUTO_TEST_CASE(Construct){
	TimeSeries<unsigned int,double> s;
	BOOST_CHECK_EQUAL(s.length(), 0);
	BOOST_CHECK(s.isEmpty());
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s.values[0], std::invalid_argument);
	BOOST_CHECK_THROW(s.timestamps[0], std::invalid_argument);
#endif

	// copy contruct
	TimeSeries<unsigned int,double> sC = s;
	BOOST_CHECK_EQUAL(sC.length(), 0);

	// assignment
	TimeSeries<unsigned int,double> s2;
	s2.append(0, 0.0);
	BOOST_REQUIRE( !s2.isEmpty() );
	s2 = s;
	BOOST_CHECK( s2.isEmpty() );
}

BOOST_AUTO_TEST_CASE(IndexAccess){
	TimeSeries<unsigned int,double> s;
	s.append(0, 2.0);
	s.append(3, 4.0);

	// size	
	BOOST_CHECK_EQUAL(s.length(), 2);
	BOOST_CHECK( !s.isEmpty() );

	// indexed get
	BOOST_CHECK_EQUAL(s.values[0], 2.0);
	BOOST_CHECK_EQUAL(s.values[1], 4.0);
	BOOST_CHECK_EQUAL(s.timestamps[0], 0);
	BOOST_CHECK_EQUAL(s.timestamps[1], 3);
	
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(s.values[2], std::invalid_argument);
	BOOST_CHECK_THROW(s.timestamps[2], std::invalid_argument);
#endif

	// modify (check copy-on-write)
	TimeSeries<unsigned int,double> sC = s;
	BOOST_CHECK_EQUAL(sC.length(), 2);

	// set value
	sC.values.set(0, 1.0);
	BOOST_CHECK_EQUAL(sC.values[0], 1.0);
	BOOST_CHECK_EQUAL(s.values[0], 2.0); // unchanged?

	// set timestamp
	sC.timestamps.set(1, 2);
	BOOST_CHECK_EQUAL(sC.timestamps[1], 2);
	BOOST_CHECK_EQUAL(s.timestamps[1], 3); // unchanged?

	// remove index
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( sC.timestamps.remove(2), std::invalid_argument );
#endif
	sC.timestamps.remove(0);
	BOOST_CHECK_EQUAL(sC.length(), 1);
	BOOST_CHECK_EQUAL(s.length(), 2); // unchanged?
	BOOST_CHECK_EQUAL(sC.values[0], 4.0);
	BOOST_CHECK_EQUAL(sC.timestamps[0], 2);

	sC.append(3, 5.0);
	sC.append(4, 6.0);
	BOOST_CHECK_EQUAL( sC.length(), 3 );

	// remove first N
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( sC.removeFirst(4), std::invalid_argument );
#endif

	// remove first
	sC.removeFirst(0);
	BOOST_CHECK_EQUAL( sC.length(), 3 );

	sC.removeFirst(2);
	BOOST_CHECK_EQUAL( sC.length(), 1 );
	BOOST_CHECK_EQUAL( sC.values[0], 6.0 );
	BOOST_CHECK_EQUAL( sC.timestamps[0], 4 );

	sC.removeFirst();
	BOOST_CHECK_EQUAL( sC.length(), 0 );

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( sC.removeFirst(), std::invalid_argument );
#endif

	BOOST_CHECK_EQUAL( sC.length(), 0 );
	BOOST_CHECK_EQUAL( s.length(), 2 ); // unchanged?
}


BOOST_AUTO_TEST_CASE(IndexAccessModifyTimestamp){
	TimeSeries<int,double> s;
	s.append(0, 2.0);
	s.append(3, 4.0);

	// manipulate first
	s.timestamps.set(0, -1);
	BOOST_CHECK_EQUAL(s.timestamps[0], -1);
	BOOST_CHECK_EQUAL(s.timestamps[1], 3);
	s.timestamps.set(0, 3); // same timestamp
	BOOST_CHECK_EQUAL(s.timestamps[0], 3);
	BOOST_CHECK_EQUAL(s.timestamps[1], 3);

	s.timestamps.set(0, 0); // restore

	// manipulate last
	s.timestamps.set(1, 0); // same timestamp
	BOOST_CHECK_EQUAL(s.timestamps[0], 0);
	BOOST_CHECK_EQUAL(s.timestamps[1], 0);
	s.timestamps.set(1, 3); // restore

	// manipulate in between
	s.insert(1, 3.0);
	s.timestamps.set(1, 2);
	BOOST_CHECK_EQUAL(s.timestamps[0], 0);
	BOOST_CHECK_EQUAL(s.timestamps[1], 2);
	BOOST_CHECK_EQUAL(s.timestamps[2], 3);

	TimeSeries<int,double,true> strict;
	strict.append(0, 1.0);
	strict.append(3, 2.0);
	BOOST_CHECK_EQUAL(strict.timestamps[0], 0);
	BOOST_CHECK_EQUAL(strict.timestamps[1], 3);
}

BOOST_AUTO_TEST_CASE(timestampReOrdering){
	TimeSeries<double,double> r1;
	r1.append(0.0, 1.0);
	r1.append(1.0, 2.0);
	r1.append(2.0, 3.0);
	TimeSeries<double,double> r1B = r1;

	// manipulate front
	r1.timestamps.set(0, -1.0);
	BOOST_CHECK_EQUAL( r1[-1.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 2.0], 3.0 );
	r1 = r1B;

	r1.timestamps.set(0, 1.0);
	BOOST_CHECK_EQUAL( r1.timestamps[0], 1.0 );
	BOOST_CHECK_EQUAL( r1.values[0], 1.0 );
	BOOST_CHECK_EQUAL( r1.timestamps[1], 1.0 );
	BOOST_CHECK_EQUAL( r1.values[1], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 2.0], 3.0 );
	r1 = r1B;

	r1.timestamps.set(0, 1.5);
	BOOST_CHECK_EQUAL( r1[ 1.5], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 2.0], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}
	r1 = r1B;

	r1.timestamps.set(0, 2.5);
	BOOST_CHECK_EQUAL( r1[ 2.5], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 2.0], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}
	r1 = r1B;

	// manipulate back
	r1.timestamps.set(2, 2.5);
	BOOST_CHECK_EQUAL( r1[ 0.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 2.5], 3.0 );
	r1 = r1B;

	r1.timestamps.set(2, 1.5);
	BOOST_CHECK_EQUAL( r1[ 0.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 1.5], 3.0 );
	r1 = r1B;

	r1.timestamps.set(2, 0.5);
	BOOST_CHECK_EQUAL( r1[ 0.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ 0.5], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}
	r1 = r1B;

	r1.timestamps.set(2, -0.5);
	BOOST_CHECK_EQUAL( r1[ 0.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[ 1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[ -0.5], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}
	r1 = r1B;
}

BOOST_AUTO_TEST_CASE(timestampReOrderingStrict){
	TimeSeries<double,double,true> r1;
	r1.append(0.0, 1.0);
	r1.append(1.0, 2.0);
	r1.append(2.0, 3.0);
	TimeSeries<double,double,true> r1B = r1;

	BOOST_CHECK_NO_THROW( r1.timestamps.set(0, 0.0) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( r1.timestamps.set(0, 1.0), std::invalid_argument );
	BOOST_CHECK_THROW( r1.timestamps.set(0, 2.0), std::invalid_argument );
#endif

	BOOST_CHECK_NO_THROW( r1.timestamps.set(0, 2.1) );
	BOOST_CHECK_EQUAL( r1[2.1], 1.0 );
	BOOST_CHECK_EQUAL( r1[1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[2.0], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}

	r1 = r1B;

	BOOST_CHECK_NO_THROW( r1.timestamps.set(2, 3.0) );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( r1.timestamps.set(2, 0.0), std::invalid_argument );
	BOOST_CHECK_THROW( r1.timestamps.set(2, 1.0), std::invalid_argument );
#endif

	BOOST_CHECK_NO_THROW( r1.timestamps.set(2, -0.1) );
	BOOST_CHECK_EQUAL( r1[0.0], 1.0 );
	BOOST_CHECK_EQUAL( r1[1.0], 2.0 );
	BOOST_CHECK_EQUAL( r1[-0.1], 3.0 );
	for(unsigned int i=0; i<r1.length()-1; i++){
		BOOST_CHECK( r1.timestamps[i] <= r1.timestamps[i+1] );
	}
	
}

BOOST_AUTO_TEST_CASE(TimestampAccess){
	TimeSeries<int,double> s;
	s.append(0, 1.0);
	s.append(3, 2.0);
	TimeSeries<int,double> sC = s;

	// get
	BOOST_CHECK_EQUAL( s[0], 1.0 );
	BOOST_CHECK_EQUAL( s[3], 2.0 );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( s[1], std::invalid_argument );
	BOOST_CHECK_THROW( s[4], std::invalid_argument );
#endif

	// set
	s.setValue(0, 2.0);
	s.setValue(3, 3.0);
	BOOST_CHECK_EQUAL( s[0], 2.0 );
	BOOST_CHECK_EQUAL( s[3], 3.0 );
	BOOST_CHECK_EQUAL( sC[0], 1.0 ); // unchanged?
	BOOST_CHECK_EQUAL( sC[3], 2.0 ); // unchanged?

	// insert (before, in between, after)
	s.insert(-1, 1.0);
	s.insert( 1, 2.5);
	s.insert( 4, 4.0);
	BOOST_CHECK_EQUAL( s.timestamps[0],-1 );
	BOOST_CHECK_EQUAL( s.values[0], 1.0 );
	BOOST_CHECK_EQUAL( s.timestamps[1], 0 );
	BOOST_CHECK_EQUAL( s.values[1], 2.0 );
	BOOST_CHECK_EQUAL( s.timestamps[2], 1 );
	BOOST_CHECK_EQUAL( s.values[2], 2.5 );
	BOOST_CHECK_EQUAL( s.timestamps[3], 3 );
	BOOST_CHECK_EQUAL( s.values[3], 3.0 );
	BOOST_CHECK_EQUAL( s.timestamps[4], 4 );
	BOOST_CHECK_EQUAL( s.values[4], 4.0 );
	BOOST_CHECK_EQUAL( s.length(), 5 );
	BOOST_CHECK_EQUAL( sC.length(), 2 ); // unchanged?

	// insert (same timestamp)
	BOOST_CHECK_NO_THROW( s.insert(-1, 1.1) );
	BOOST_CHECK_EQUAL( s.timestamps[0], -1 );
	BOOST_CHECK_EQUAL( s.values[0], 1.0 );
	BOOST_CHECK_EQUAL( s.timestamps[1], -1 );
	BOOST_CHECK_EQUAL( s.values[1], 1.1 );

	// insert (same timestamp, strict ordering)
	TimeSeries<int,double,true> strict;
	strict.append(0, 1.0);
	strict.append(3, 2.0);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( strict.insert(0, 1.1), std::invalid_argument );
#endif
	BOOST_CHECK_EQUAL( strict.values[0], 1.0 );
	
	// append
	BOOST_CHECK_EQUAL( s.length(), 6 );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( s.append(3, 0.0), std::invalid_argument );
#endif
	BOOST_CHECK_EQUAL( s.length(), 6 );
	s.append(4, 4.1); // same timestamp
	s.append(5, 5.0); // later timestamp
	BOOST_CHECK_EQUAL( s.length(), 8 );

#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( strict.append(2, 0.0), std::invalid_argument );
	BOOST_CHECK_THROW( strict.append(3, 0.0), std::invalid_argument ); // same timestamp
#endif
	BOOST_CHECK_EQUAL( strict.length(), 2 );
	strict.append(4, 4.0);
	BOOST_CHECK_EQUAL( strict.length(), 3 );

	// offset append
	TimeSeries<int,double> s2;
	s2.append(offset(2), 1.0); // Timestamp = int() + 2
	BOOST_CHECK_EQUAL( s2[2], 1.0 );
	s2.append(offset(1), 2.0); // Timestamp = 2 + 1
	BOOST_CHECK_EQUAL( s2[3], 2.0 );
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW( s2.append( offset(-1), 3.0 ), std::invalid_argument );
#endif
	BOOST_CHECK_EQUAL( s2.length(), 2 );

	TimeSeries<int,double> s3;
	s3.append(0, 3.0);
	s3.append(2, 4.0);

	s2.append(offset(3), s3);
	BOOST_CHECK_EQUAL( s2.length(), 4 );
	BOOST_CHECK_EQUAL( s2[2], 1.0 );
	BOOST_CHECK_EQUAL( s2[3], 2.0 );
	BOOST_CHECK_EQUAL( s2[6], 3.0 );
	BOOST_CHECK_EQUAL( s2[8], 4.0 );
	
}



BOOST_AUTO_TEST_SUITE_END()

