
#define BOOST_TEST_MODULE VectorMappings
#include <boost/test/unit_test.hpp>

#include "../src/nemo/VectorMappings.h"

#include <iostream>

using namespace nemo;
using namespace boost::unit_test;

// boost has non-virtual constructors ^^
#pragma GCC diagnostic ignored "-Weffc++"

typedef Mapping<RealVector,RealVector > VectorMapping;


BOOST_AUTO_TEST_SUITE(VectorMappings)


BOOST_AUTO_TEST_CASE(subspaceMapping){
	VectorMapping m1 = SubspaceMapping<>::create(0, 2);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m1(RealVector(1.0)), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(m1(RealVector(1.0, 2.0)), RealVector(1.0, 2.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0, 2.0, 3.0)), RealVector(1.0, 2.0));

	VectorMapping m2 = SubspaceMapping<>::create(1, 2);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m2(RealVector(1.0, 2.0)), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(m2(RealVector(1.0, 2.0, 3.0)), RealVector(2.0, 3.0));

	bool select3[2];
	select3[0] = false;
	select3[1] = true;
	VectorMapping m3 = SubspaceMapping<>::create(2, select3);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m3(RealVector(1.0)), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(m3(RealVector(1.0, 2.0)), RealVector(2.0));

	bool select4[3];
	select4[0] = true;
	select4[1] = true;
	select4[2] = false;
	VectorMapping m4 = SubspaceMapping<>::create(3, select4);
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m4(RealVector(1.0)), std::invalid_argument);
#endif
	BOOST_CHECK_EQUAL(m4(RealVector(1.0, 2.0)), RealVector(1.0, 2.0));
	BOOST_CHECK_EQUAL(m4(RealVector(1.0, 2.0, 3.0)), RealVector(1.0, 2.0));
}

BOOST_AUTO_TEST_CASE(polynomExpansion){
	VectorMapping m1 = PolynomExpansion<>::create(0);
	BOOST_CHECK_EQUAL(m1(RealVector()), RealVector(1.0));
	BOOST_CHECK_EQUAL(m1(RealVector(2.0)), RealVector(1.0));

	VectorMapping m2 = PolynomExpansion<>::create(1);
	BOOST_CHECK_EQUAL(m2(RealVector()), RealVector(1.0));
	BOOST_CHECK_EQUAL(m2(RealVector(2.0)), RealVector(1.0, 2.0));
	BOOST_CHECK_EQUAL(m2(RealVector(2.0, 3.0)), RealVector(1.0, 2.0, 3.0));

	VectorMapping m3 = PolynomExpansion<>::create(2);
	BOOST_CHECK_EQUAL(m3(RealVector()), RealVector(1.0));
	BOOST_CHECK_EQUAL(m3(RealVector(2.0)), RealVector(1.0, 2.0, 4.0));
	RealVector v3(dim(6));
	v3[0] = 1.0; v3[1] = 2.0; v3[2] = 3.0;
	v3[3] = 4.0; v3[4] = 6.0; v3[5] = 9.0;
	BOOST_CHECK_EQUAL(m3(RealVector(2.0, 3.0)), v3);

	VectorMapping m4 = PolynomExpansion<>::create(10);
	RealVector v4(dim(11), 1.0);
	for(unsigned int i=1; i<11; i++){
		v4[i] = v4[i-1]*2;
	}
	BOOST_CHECK_EQUAL(m4(RealVector(2.0)), v4);
}


BOOST_AUTO_TEST_CASE(clamping){
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(
		Clamping<>::create(RealVector(0.0, 0.0), RealVector(1.0)), 
		std::invalid_argument
	);
	BOOST_CHECK_THROW(
		Clamping<>::create(RealVector(0.0, 0.0), RealVector(1.0, -1.0)), 
		std::invalid_argument
	);
#endif

	VectorMapping m1 = Clamping<>::create(RealVector(0.0, 0.0), RealVector(1.0, 2.0));
#ifndef NEMO_ABORT_ON_ERROR
	BOOST_CHECK_THROW(m1(RealVector()), std::invalid_argument);
#endif

	BOOST_CHECK_EQUAL(m1(RealVector(-1.0, 0.5)), RealVector(0.0, 0.5));
	BOOST_CHECK_EQUAL(m1(RealVector(0.5, -1.0)), RealVector(0.5, 0.0));

	BOOST_CHECK_EQUAL(m1(RealVector(0.0, 0.0)), RealVector(0.0, 0.0));
	BOOST_CHECK_EQUAL(m1(RealVector(0.5, 1.0)), RealVector(0.5, 1.0));
	BOOST_CHECK_EQUAL(m1(RealVector(1.0, 2.0)), RealVector(1.0, 2.0));

	BOOST_CHECK_EQUAL(m1(RealVector(2.0, 0.5)), RealVector(1.0, 0.5));
	BOOST_CHECK_EQUAL(m1(RealVector(0.5, 3.0)), RealVector(0.5, 2.0));
}


BOOST_AUTO_TEST_CASE(cwiseMapping){
	VectorMapping m = RealVector::cwise( arg<double>() * 2.0 );

	BOOST_CHECK_EQUAL(m(RealVector(0.0, 1.0)), RealVector(0.0, 2.0));
	BOOST_CHECK_EQUAL(m(RealVector(2.0, 1.0)), RealVector(4.0, 2.0));
}

BOOST_AUTO_TEST_CASE(CartesianSpherical){
	VectorMapping m = Cartesian2Spherical<>::create();
	BOOST_CHECK_EQUAL(m(RealVector()), RealVector());

	// 1D
	BOOST_CHECK_EQUAL(m(RealVector(-1.0)), RealVector(-1.0));
	BOOST_CHECK_EQUAL(m(RealVector(0.0)), RealVector(0.0));
	BOOST_CHECK_EQUAL(m(RealVector(1.0)), RealVector(1.0));

	// 2D
	BOOST_CHECK_EQUAL(m(RealVector(0.0, 0.0)), RealVector(0.0, 0.0));

	

	BOOST_CHECK_EQUAL(m(RealVector(1.0, 0.0)), RealVector(1.0, 0.0/2.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(0.0, 1.0)), RealVector(1.0, 1.0/2.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(-1.0, 0.0)), RealVector(1.0, 2.0/2.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(0.0, -1.0)), RealVector(1.0, 3.0/2.0*M_PI));

	BOOST_CHECK_EQUAL(m(RealVector(1.0, 1.0)), RealVector(sqrt(2.0), 1.0/4.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(-1.0,1.0)), RealVector(sqrt(2.0), 3.0/4.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(-1.0,-1.0)), RealVector(sqrt(2.0), 5.0/4.0*M_PI));
	BOOST_CHECK_EQUAL(m(RealVector(1.0,-1.0)), RealVector(sqrt(2.0), 7.0/4.0*M_PI));

	// 3D
	BOOST_CHECK_EQUAL(m(RealVector(0.0, 0.0, 0.0)), RealVector(0.0, 0.0, 0.0));
	
	BOOST_CHECK_EQUAL(m(RealVector(1.0, 0.0, 0.0)), RealVector(1.0, 0.0/2.0*M_PI, 0.0));
	BOOST_CHECK_EQUAL(m(RealVector(0.0, 1.0, 0.0)), RealVector(1.0, 1.0/2.0*M_PI, 0.0));
	BOOST_CHECK_EQUAL(m(RealVector(-1.0, 0.0, 0.0)), RealVector(1.0, 2.0/2.0*M_PI, 0.0));
	BOOST_CHECK_EQUAL(m(RealVector(0.0, -1.0, 0.0)), RealVector(1.0, 3.0/2.0*M_PI, 0.0));

	BOOST_CHECK_EQUAL(m(RealVector(0.0, 0.0, 1.0)), RealVector(1.0, 0.0, M_PI/2.0));
	BOOST_CHECK_EQUAL(m(RealVector(0.0, 0.0, -1.0)), RealVector(1.0, 0.0, -M_PI/2.0));
}

double dist(RealVector x, RealVector y){
	RealVector d = x-y;
	double s = 0.0;
	for(unsigned int i=0; i<d.dimension(); i++){
		s += d[i]*d[i];
	}
	return sqrt( s );
}

BOOST_AUTO_TEST_CASE(SphericalCartesian){
	VectorMapping m = Spherical2Cartesian<>::create();
	BOOST_CHECK_EQUAL(m(RealVector()), RealVector());
	
	RealVector v;

	// 1D
	BOOST_CHECK_EQUAL(m(RealVector(-1.0)), RealVector(-1.0));
	BOOST_CHECK_EQUAL(m(RealVector(0.0)), RealVector(0.0));
	BOOST_CHECK_EQUAL(m(RealVector(1.0)), RealVector(1.0));

	// 2D
	v = m(RealVector(0.0, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 0.0) ), 1e-12 );

	v = m(RealVector(1.0, 0.0/2.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(1.0, 0.0) ), 1e-12 );
	v = m(RealVector(1.0, 1.0/2.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 1.0) ), 1e-12 );
	v = m(RealVector(1.0, 2.0/2.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(-1.0, 0.0) ), 1e-12 );
	v = m(RealVector(1.0, 3.0/2.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, -1.0) ), 1e-12 );

	v = m(RealVector(sqrt(2.0), 1.0/4.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(1.0, 1.0) ), 1e-12 );
	v = m(RealVector(sqrt(2.0), 3.0/4.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(-1.0,1.0) ), 1e-12 );
	v = m(RealVector(sqrt(2.0), 5.0/4.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(-1.0,-1.0) ), 1e-12 );
	v = m(RealVector(sqrt(2.0), 7.0/4.0*M_PI));
	BOOST_CHECK_SMALL( dist( v, RealVector(1.0,-1.0) ), 1e-12 );

	// 3D
	v = m(RealVector(0.0, 0.0, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 0.0, 0.0) ), 1e-12 );
	
	v = m(RealVector(1.0, 0.0/2.0*M_PI, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(1.0, 0.0, 0.0) ), 1e-12 );
	v = m(RealVector(1.0, 1.0/2.0*M_PI, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 1.0, 0.0) ), 1e-12 );
	v = m(RealVector(1.0, 2.0/2.0*M_PI, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(-1.0, 0.0, 0.0) ), 1e-12 );
	v = m(RealVector(1.0, 3.0/2.0*M_PI, 0.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, -1.0, 0.0) ), 1e-12 );

	v = m(RealVector(1.0, 0.0, M_PI/2.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 0.0, 1.0) ), 1e-12 );
	v = m(RealVector(1.0, 0.0, -M_PI/2.0));
	BOOST_CHECK_SMALL( dist( v, RealVector(0.0, 0.0, -1.0) ), 1e-12 );
}

BOOST_AUTO_TEST_CASE(SphericalCartesianInversion){
	VectorMapping c2s = Cartesian2Spherical<>::create();
	VectorMapping s2c = c2s.inverse();
	for(int x = -1; x <= 1; x++){
	for(int y = -1; y <= 1; y++){
	for(int z = -1; z <= 1; z++){
	for(int w = -1; w <= 1; w++){
		RealVector c(dim(4));
		c[0] = x;
		c[1] = y;
		c[2] = z;
		c[3] = w;
		RealVector cr = s2c(c2s(c));
		BOOST_CHECK_SMALL( dist( c, cr ), 1e-12 );
	}
	}
	}
	}
}


BOOST_AUTO_TEST_SUITE_END()

